<?php

/*
   All First Quarter Moons
   For the 600-year range from 1600 to 2200

*/


   ob_start();


// Set GLOBAL Delta T mode to 'ON' state.
// Needed for Times Table.
   $DTOnOff = 'ON';

// Define the program cookie name and set it to expire in 30 days.
// This cookie can be shared by other programs in the same common
// home folder using this interface in common.
   $CookieName = 'First_Quarter_Moons_Interface_Settings';
   $SetToExpireIn30Days = time() + 30*86400;

// Set minimum and maximum year span to allow.   Any year outside this
// range will be considered an invalid year.   For the DE405 ephemeris,
// the valid year range is from 1600 to 2200 on the Gregorian calendar.
   $YMin = 1600;
   $YMax = 2200;

   $cYear = gmdate('Y');

// Define main TextArea text and background colors.
   $TxColor = 'black';
   $BgColor = 'white';

// Define script path and filename.
   $_AUTHOR_          = 'Jay Tanner';
   $_PROGRAM_VERSION_ = "v1.0"; $at = "&#97;&#116;"; $UTC = "&#85;&#84;&#67";
   $_SCRIPT_PATH_     = filter_input(INPUT_SERVER, 'SCRIPT_FILENAME');
   $_RUN_THIS_SCRIPT_ = filter_input(INPUT_POST, 'SCRIPT_NAME');

// Define internal document page title, HTML page heading text and revision date
   $_INTERNAL_TITLE_  = "All First Quarter Moons From 1600 to 2200";
   $_INTERFACE_TITLE_ = "<b style='font-size:16px;'>All First Quarter Moons</b><br><span style='font-size:12px;'>In the 600-Year Range From&nbsp; $YMin to $YMax</span>";
   $_REVISION_DATE_   = $_PROGRAM_VERSION_ .' - Revised: '. gmdate("l, F d, Y $at h:i A   $UTC", FileMTime($_SCRIPT_PATH_));

   $Match_Pattern = filter_input(INPUT_POST, 'MatchPattern');
   $Match_Pattern = Str_Replace("'",'', $Match_Pattern);
   $Match_Pattern = Str_Replace('"','', $Match_Pattern);
   if (trim($Match_Pattern) == ''){$Match_Pattern = " $cYear ";}
   if (trim($Match_Pattern) == '.'){$Match_Pattern = "";}

// Do this only if [COMPUTE] button was clicked.
   $w = filter_input(INPUT_POST, 'ComputeButton');
   if (!isset($w))

// ----------------------------------------------------
// If [COMPUTE] button was clicked and an active cookie
// exists, then restore the interface settings from it.
     {
      $w = filter_input(INPUT_COOKIE, $CookieName);
   if (isset($w))
      {
       $CookieDataString = filter_input(INPUT_COOKIE, $CookieName);
        list
       (
        $Y,
        $mn,
        $dn,
        $hhmmssLT,
        $hhmmTZ,
        $LatLon,
        $elev
       ) = preg_split("[\|]", $CookieDataString);

        $elev = floatval(trim($elev));

        $m = intval($mn);
        $d = intval($dn);
        $LatLon = Check_Lat_Lon($LatLon);
        list($lat, $lon) = preg_split("[,]", $LatLon);
        $lat  = trim($lat);
        $lon  = trim($lon);
        $LatLon = "$lat  ,  $lon";
        $elevFt = ($elev * 0.3048);
        $elev = sprintf("%+1.1f", $elev);
      }

   else

// Set the default interface startup values.

 {
  $pattern = '';

  $Y = date('Y');
  $m = date('n'); $mn  = sprintf("%02d", $m);
  $d = date('j'); $dn  = sprintf("%02d", $d);

  $hhmmssLT = date('H:i:s');

  $hhmmTZ = '+00:00'; // substr(Hours_to_HMS(date('Z')/3600, 0, '+', ':'),0,6);

// Home location coordinates for Ovid, New York, USA.
  $LatLon = '+42 40 35.3  ,  -76 49 22.8';
  $LatLon = Check_Lat_Lon($LatLon);
  list($lat, $lon) = preg_split("[,]", $LatLon);
  $lat    = trim($lat);
  $lon    = trim($lon);
  $LatLon = "$lat  ,  $lon";

  $elev   = '+0.0 m';
  $elevFt = '+0.0 ft';

// Store interface settings in cookie.
   $CookieDataString = "$Y|$mn|$dn|$hhmmssLT|$hhmmTZ|$LatLon|$elev";
   setcookie ($CookieName, $CookieDataString, $SetToExpireIn30Days);
  }
      } // End of  if (!isset(_POST['ComputeButton']))


// Read values of all interface settings.
   $w = filter_input(INPUT_POST, 'ComputeButton');
   if (isset($w))
{
   $Y  = filter_input(INPUT_POST, 'Y');
   $mn = filter_input(INPUT_POST, 'mn'); $m  = intval($mn);
   $dn = filter_input(INPUT_POST, 'dn'); $d  = intval($dn);

   $hhmmssLT = filter_input(INPUT_POST, 'hhmmssLT');
   $hhmmssLT = Num_to_HMS ($hhmmssLT);

   $hhmmTZ = trim(filter_input(INPUT_POST, 'hhmmTZ'));
   $hhmmTZ = substr(Num_to_HMS($hhmmTZ), 0,6);

   $LatLon = trim(filter_input(INPUT_POST, 'LatLon'));
   $elev = trim(filter_input(INPUT_POST, 'elev'));

// If elevation was given in feet, then convert
// into equivalent meters internally.
   if(substr_count(strtolower($elev), 'f') > 0)
     {$elev = floatval($elev) * 0.3048;}
   else
     {$elev = floatval($elev);}


// Do some interface formatting.
   $mn = sprintf("%02d", $m);
   $dn = sprintf("%02d", $d);
   $hhmmssLT = Hours_to_HMS(HMS_to_Hours($hhmmssLT), 0, '', ':');

   $hhmmTZ = substr(Hours_to_HMS(HMS_to_Hours($hhmmTZ), 0, '+', ':'),0,6);

   $LatLon = str_replace(',',' , ', $LatLon);
   $LatLon = Check_Lat_Lon($LatLon);
   list($lat, $lon) = preg_split("[,]", $LatLon);
   $lat = trim($lat);
   $lon = trim($lon);
   $LatLon = "$lat  ,  $lon";

//   $elevFt = sprintf("%+1.1f", (floatval($elev) * 0.3048)).' ft';
//   $elev = sprintf("%+1.1f", $elev);

// -----------------------------------
// Store interface settings in cookie.

   $CookieDataString = "$Y|$mn|$dn|$hhmmssLT|$hhmmTZ|$LatLon|$elev";
   setcookie ($CookieName, $CookieDataString, $SetToExpireIn30Days);
}









// ----------------------------------------
// Filter for date error. TRUE = Date is OK
// or FALSE = Bad date error.
   $DateFlag = checkdate($m, $d, $Y);
   $YearFlag = ($Y < $YMin or $Y > $YMax)? FALSE:TRUE;

if ($DateFlag === FALSE or $YearFlag === FALSE)
{
   $TextArea2Text = '';

   $w = 'DATE ERROR:';

      $TxColor = 'white';
      $BgColor = '#AA0000';

   $hhmmssDT = '-';

   if ($YearFlag === FALSE)
  {
   $TextArea1Text =
"$w

The given year ($Y) is outside the valid range from $YMin to $YMax.";
  }
   else
{
   $TextArea2Text = '';

   $TextArea1Text =
"$w

The given date elements (Year Month Day) = ($Y $mn $dn) do not
represent any real date on the standard Gregorian calendar.

-------------------
VALID DATE ELEMENTS

Year  =  1600 to 2200 on the modern Gregorian calendar.
Month =  1 to 12 - For January to December.
Day   =  1 to 31 - Depending on the month and if it is a leap year.
";
}

}

   else

// Do this if no date error detected.

{
// Compute local Julian Date for 00:00:00 of Date.
   $JDNumLT = GregorianToJD($m, $d, $Y);
   $JD00LT  = $JDNumLT - 0.5;
   $DateStrLT = trim(Inv_JD_Num($JDNumLT, 'G', 4));

// Compute NASA Delta T estimate for local date.
   $hhmmssDT = NASA_Delta_T_Poly($JD00LT, 'G', 3);

   $elevFt = sprintf("%+1.1f", floatval(trim($elev)) / 0.3048);
   $elev = sprintf("%+1.1f", $elev).' m';

// Compute fractions of a day corresponding to LT, TZ and DT.
   $FracDayLT = HMS_to_Hours($hhmmssLT)/24;
   $FracDayLT = 0.0; // OVERRIDE LT VALUE WITH ZERO.
   $FracDayTZ = HMS_to_Hours($hhmmTZ)/24;
   $FracDayDT = HMS_to_Hours($hhmmssDT)/24;

// Compute Julian Dates for LT, UT and TT scales.
   $JDLT = $JD00LT + $FracDayLT;
   $JDUT = $JDLT - $FracDayTZ;
   $JDTT = $JDUT + $FracDayDT;

// Compute JDNumLT, JDNumUT and JDNumTT.
   $JDNumLT = floor($JDLT + 0.5);
   $JDNumUT = floor($JDUT + 0.5);
   $JDNumTT = floor($JDTT + 0.5);

   $DateStrLT = trim(Inv_JD_Num($JDNumLT, 'G', 4));
   $DateStrUT = trim(Inv_JD_Num($JDNumUT, 'G', 4));
   $DateStrTT = trim(Inv_JD_Num($JDNumTT, 'G', 4));

// Do some formatting.
   $JDLT = sprintf("%7.8f", $JDLT);
   $JDUT = sprintf("%7.8f", $JDUT);
   $JDTT = sprintf("%7.8f", $JDTT);

   $LatDMS = Deg_to_DMS($lat, 3, '+', '');
   $LonDMS = Deg_to_DMS($lon, 3, '+', '');
   $LonHMS = Hours_to_HMS($lon/15, 3, '+', 'h');

// Set (Lat, Lon) symbols for N|S and E|W
   $SN  = (floatval($lat) < 0)? "S":"N"; if (floatval($lat) == 0) {$SN = '';}
   $WE  = (floatval($lon) < 0)? "W":"E"; if (floatval($lon) == 0) {$WE = '';}

// Compute equivalent times table for LT, UT and TT scales.
   $TimesTableText = JDTT_to_Times_Table($JDTT, $hhmmTZ);
   $TimesTableText = ''; // OVERRIDE TIMES TABLE WITH EMPTY STRING.

// Compute topocentric elements for WGS-84 ellipsoid.
   $ii  = 6356752.314245 / 6378137;
   $jj  = floatval(trim($elev)) / 6378137;
   $kk = atan($ii*tan($lat));

   $RhoxSinPhiPrime = $ii*sin($kk) + $jj*sin($lat);
   $RhoxCosPhiPrime = cos($kk) + $jj*cos($lat);

   $RhoxSinPhiPrime = sprintf('%+1.10f', $RhoxSinPhiPrime);
   $RhoxCosPhiPrime = sprintf('%+1.10f', $RhoxCosPhiPrime);

// Get number of days in the given month.
   $mDays = cal_days_in_month(CAL_GREGORIAN, $m, $Y);

// Get 3-letter abbreviation of month name.
   $MmmStr = substr("JanFebMarAprMayJunJulAugSepOctNovDec", 3*($m-1), 3);

// Construct location text block.
   $LocationText =
"-------------------------------------------------
Geographic Location      WGS-84 (GPS) Coordinates

                 &deg;   &#39;   &#34;
Latitude    $SN  $LatDMS  =  $lat&deg;
Longitude   $WE  $LonDMS  =  $lon&deg;  =  $LonHMS

Elevation      $elev  =  $elevFt ft";

$LocationText = ''; // OVERRIDE LOCATION BLOCK WITH EMPTY STRING.

// #############################################################
// START OF MAIN COMPUTATIONS BLOCK USING JDXX AND OTHER VALUES.
// Y m mn d dn JDLT JDUT JDTT hhmmssLT hhmmTZ hhmmssDT ... etc.

   $UT_or_LT = ($FracDayTZ == 0)? 'UT':'LT';

   $Output_Table = trim(Match_Pattern ($Match_Pattern, $hhmmTZ));

   $Total_Matches = 1 + substr_count($Output_Table, "\n");
   if (strpos($Output_Table, '???') !== FALSE) {$Total_Matches = 0;}

   if(trim($Match_Pattern) == '') {$Total_Matches = 0;}

// Replace single quote marks in search pattern.
   $Match_Pattern = "'$Match_Pattern'";




// END OF MAIN COMPUTATIONS BLOCK
// #############################################################















// --------------------------------------
// Construct output text block 1 (upper).
// This shows the interface settings values.

   $TextArea1Text =
" All First Quarter Moons Within the 600-Year
 Range From 1600 to 2200

 Local Time Zone Offset  = UT$hhmmTZ
 Optional Search Pattern = $Match_Pattern
                 Matches = $Total_Matches

=========================================
 Julian_Date_TT   Calendar_Date  Time_$UT_or_LT
================ =============== ========
$Output_Table
================ =============== ========
 Julian_Date_TT   Calendar_Date  Time_$UT_or_LT
";


// Construct output text block 2 (lower).
// This shows the computed results and tables.

   $TextArea2Text =
"                       GENERAL PROGRAM INFORMATION

This program lists all first quarter moons in the 600-year range from
1600 to 2200 with a simple pattern search option.  The search pattern
string is both SPACE and CASE sensitive so that it can also contain
leading and trailing spaces in the search.

The default startup search pattern is the current year value ' $cYear '.

To list all first quarter moons in a specific year, use the year value
as the search pattern padded by a single leading space, e.g., ' 1992'

To list ALL first quarter moons in the 600-year span, simply use a single
dot '.' as the search pattern string.

The program defaults to UT, but the local time zone can be set for any
general location and standard local time (LT) as needed.

To account for Daylight Saving/Summer Time, add 1 hour to the computed
standard times taking care to watch for a change of date in doing so.

The date on which an event occurs, depends on your local time zone.
For example, an event may happen on New Year's eve in New York City
while at the same moment taking place on New Year's Day in Paris.

";

}




// SPECIAL AFTER-THE-FACT PATCH TO ADD INFO BUTTON.
   $InfoButton  = filter_input(INPUT_POST, 'InfoButton');

   if (isset($InfoButton))
      {
$TextArea2Text = '';
      }





// Determine the number of text rows to use in the output TextAreas.
   $TextArea1Rows = 2 + substr_count($TextArea1Text, "\n");
   $TextArea2Rows = 2 + substr_count($TextArea2Text, "\n");


// Generate HTML web page to display computed results.
   print <<< _HTML

<!DOCTYPE HTML>

<HTML>

<head>

<title>All First Quarter Moons From 1600 to 2200</title>
<meta name='viewport' content='width=device-width, initial-scale=1'>
<meta http-equiv='content-type' content='text/html; charset=UTF-8'>
<meta name='description' content='All First Quarter Moons Table From 1600 to 2200'>
<meta name='keywords' content='full moon,moon,lunar,first quarter,phase'>
<meta name='author' content='Jay Tanner - PHPScienceLabs.com'>
<meta http-equiv='pragma'  content='no-cache'>
<meta http-equiv='expires' content='-1'>
<meta name='robots'    content='index,follow'>
<meta name='googlebot' content='index,follow'>
<link rel="canonical" href="https://www.phpsciencelabs.com/lunar-first-quarter-moons-table/">

<style type='text/css'>
BODY
{color:silver; background-color:black;  font-family:Verdana; font-size:15px}

TABLE
{font-size:12px; border: 1px solid black;}

TD
{color:black; background-color:white;  line-height:200%; font-size:12px; padding:6px; text-align:center;}

DIV
{color:black; background-color:white;  font-family:Verdana; text-align:justify; padding:4px; font-size:16px;}

PRE
{color:black; background-color:white;  font-family:monospace; font-size:14px; font-weight:bold; text-align:left; padding:4px;}

CODE
{font-family:monospace; font-size:15px; font-weight:normal;}

TEXTAREA
{color:black; background-color:white; font-family:monospace; font-size:15px; font-weight:bold; box-shadow:2px 3px 4px #888888; border-radius: 8px; border:1px solid black; padding:4px; white-space:pre;}

INPUT[type="text"]::-ms-clear {width:0; height:0;}

INPUT[type="text"]
{color:black; background-color:white; font-size:18px; font-family:monospace; font-weight:bold; text-align:center; box-shadow:2px 2px 3px #666666; border:2px solid black;}

INPUT[type="text"]:focus
{background-color:white; box-shadow:2px 2px 3px #666666; font-family:monospace; font-size:18px; border:2px solid red; text-align:center; font-weight:bold;}

INPUT[type="submit"]
{font-weight:bold;}

INPUT[type="submit"]:hover
{box-shadow:2px 2px 5px #555555; font-weight:bold;}

INPUT[type="checkbox"]:checked
{box-shadow:2px 2px 3px #555555;}

HR {background-color:black; height:2px; border:0px;}

A:link,  A:visited {background-color:transparent; color:blue; font-family:Verdana; font-size:12px; font-weight:bold; text-decoration:none; line-height:150%; padding:2px;}
A:hover, A:active  {background-color:transparent; color:red;  font-family:Verdana; font-size:12px; font-weight:bold; text-decoration:none; line-height:150%; padding:2px;}

::selection      {background-color:yellow; color:black;}  /* Only affects IE9+ and Safari     */
::-moz-selection {background-color:yellow; color:black;}  /* Only affects FireFox and Mozilla */
</style>


</head>

<body>

<form name="form1" method="post" action="$_RUN_THIS_SCRIPT_">

<div ID='SmoothScroll' style='background:black;'>
<table width='620' border="0" cellspacing="1" cellpadding="3" align='center'>

<tr><td colspan="2" style='color:white; background-color:#000066; border:2px solid white; border-radius:8px 8px 0px 0px;'><b>$_INTERFACE_TITLE_</b></td></tr>

<td width="350"><b>Optional&nbsp;Search&nbsp;Pattern</b><br>Search Pattern is SPACE and CASE Sensitive<br>
<input name='Y'  type='hidden' value="$Y"  size='5' maxlength='4'>
<input name='MatchPattern'  type='text' value="$Match_Pattern"  size='17' maxlength='16'>

<input name='mn' type='hidden' value="$mn" size='3' maxlength='2'>
<input name='dn' type='hidden' value="$dn" size='3' maxlength='2'>
</td>



<td width="350"><b>Local&nbsp;Time&nbsp;Zone&nbsp;Offset</b><br>Neg&nbsp;|&nbsp;Pos&nbsp;&nbsp;=&nbsp;&nbsp;W&nbsp;|&nbsp;E<br>
<b style='font-size:15px;'>UT </b><input name="hhmmTZ" type="text" value="$hhmmTZ" size="7" maxlength="6">

<!-- These settings are hidden, but internally active for sharing cookie data. --->
<input name="hhmmssLT" type="hidden" value="$hhmmssLT" size="9" maxlength="8">
<input name="LatLon" type="hidden" value="$LatLon" size="39" maxlength="38">
<input name="elev" type="hidden" value="$elev" size="15" maxlength="14">
<input name="hhmmssDT" style='color:white; background-color:black;' type="hidden" value="$hhmmssDT" size="10" maxlength="9" readonly>
</td>


</td>
</tr>





<!--
<tr>
<td colspan="4"><b>Latitude</b>&nbsp;(Neg | Pos&nbsp;&nbsp;=&nbsp;&nbsp;S | N)&nbsp;&nbsp;&nbsp;and&nbsp;&nbsp;&nbsp;<b>Longitude</b>&nbsp;(Neg | Pos&nbsp;&nbsp;=&nbsp;&nbsp;W | E)<br>
±Deg.ddddd&nbsp; <b>or</b> &nbsp; ±Deg min sec<br>
<input name="LatLon" type="text" value="$LatLon" size="39" maxlength="38">
</td>

<td><b>Sea&nbsp;Level&nbsp;Elevation</b><br>
(<b>m</b>) meters or feet (<b>ft</b>)
<input name="elev" type="text" value="$elev" size="15" maxlength="14">
</td>

<td colspan="6">&nbsp;</td>
</tr>
-->









<tr><td colspan="7" style='background:black;'>
<input name="ComputeButton" type="submit" value=" C O M P U T E ">
</td></tr>

<tr>
<td colspan="7" style='background:black;'>
<textarea style='font-size:18px; color:$TxColor; background-color:$BgColor; padding:6px; box-shadow:none; border:2px solid white;' name="OutputText1" cols="46" rows="$TextArea1Rows" readonly>
$TextArea1Text
</textarea></td>
</tr>

</table>


<table align='center'>
<tr>
<td colspan="7" style='background:black;'>
<textarea style='font-size:18px; color:white; background-color:#002200; padding:6px; border:2px solid lime; border-radius:8px;' name="OutputText2" cols="76" rows="$TextArea2Rows" readonly>
$TextArea2Text
</textarea></td>
</tr>



<tr><td colspan="7" style='color:gray; background-color:black;'>Program by $_AUTHOR_<br>$_REVISION_DATE_</td></tr>

</table>
</div>

</form>

<script src='Smooth_Scroll.js'></script>
<script>SmoothScroll.init({cancelOnTouch: true});</script>

</body>

</html>

_HTML;



// ********
// ********


/*
   This function returns a table of astronomical events
   containing a target search string for finding things
   such as all
*/

   function Match_Pattern ($SearchString, $TZhhmm='+00:00')
{
// Case sensitive search string.
// Do not trim, leave as -is.
   $S = $SearchString; if (trim($S) == ''){$S = ' ';}

   $hhmmTZ = trim($TZhhmm);

   $File_Name = 'All_First_Quarter_Moons_1600_to_2200.data';
   $File_Path = "".$File_Name;

// Load events data file into work array.
   $wArray = file($File_Path);
   $wCount = count($wArray);

   $wTable = '';

   for($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (substr($Curr_Line,0,1) == '2')
      {
       list($JDTT) = preg_split("[ ]", $Curr_Line);

       $JDLT = JDTT_to_JDLT($JDTT, $hhmmTZ);
       $JDTT = sprintf("%7.8f", $JDTT);
       $JDLT = sprintf("%7.8f", $JDLT);
       $Date_Time_LT = Inv_JD($JDLT, 'AMPM');

       $w = "$JDTT $Date_Time_LT\n";

       if (strpos($w, $S) !== FALSE)
          {
           $wTable .= $w;
          }
      }

  }
   if (trim($wTable) == '')
      {$wTable = "???\n    No matches found for search pattern\n    '$S'\n???";}

   return $wTable;

}





/*
   PHPSL TRANSPORT API - FOR INDEPENDENT PROGRAM TRANSPORTABILITY.
   NOTE:
   This is a machine-generated file. Any manual edits
   will be lost when updated and recompiled.

*/

// Version Date: 2019 Mar 27 Wed 08:39:31 UTC


/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   =====================
   CONSTANTS DEFINITIONS

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode
*/


// ___________________________________________
// Speed of light in meters per second. Stolen
// from NASA/JPL DE430 constants.

// Due to the way the length of a second and the
// length of the meter are EXACTLY defined, this
// speed of light is also an EXACT definition.
// But it still doesn't REALLY tell us the TRUE
// speed of light, just what we decide it is.

   define ('CLIGHT', '299792458');

// -----------------------------------------------
// Standard Earth gravity (1 G) acceleration (SI).

   define ('G_STD', '9.80665');

// --------------------------------
// Universal gravity constant (SI).

   define ('G_UNI', '0.0000000000667408');

// -------------------------------------------------------------------
// Define some EXACT standard metric length/distance interconversion
// factors in terms of meters per unit (M_PER_XXX).  This is the
// factor by which to multiply the units to obtain equivalent meters.

   define ('M_PER_A'   , '0.0000000001');    // m per angstrom
   define ('M_PER_NM'  , '0.000000001');     // m per nanometer
   define ('M_PER_UM'  , '0.000001');        // m per micrometer (micron)
   define ('M_PER_MM'  , '0.001');           // m per millimeter
   define ('M_PER_CM'  , '0.01');            // m per centimeter
   define ('M_PER_IN'  , '0.0254');          // m per inch
   define ('M_PER_FT'  , '0.3048');          // m per foot
   define ('M_PER_YD'  , '0.9144');          // m per yard
   define ('M_PER_M'   , '1');               // m per meter
   define ('M_PER_KM'  , '1000');            // m per kilometer
   define ('M_PER_MI'  , '1609.344');        // m per mile (statute)
   define ('M_PER_NMI' , '1852');            // m per nautical mile
   define ('M_PER_LS'  , '299792458');       // m per Light Second
   define ('M_PER_LM'  , '17987547480');     // Light Minute
   define ('M_PER_LH'  , '1079252848800');   // Light Hour
   define ('M_PER_LD'  , '25902068371200');  // Light Day

   define ('KM_PER_NMI', '1.852');           // km per nautical mile
   define ('KM_PER_MI' , '1.609344');        // kilometers per statute mile

// -----------------------------
// Define the Astronomical Unit.

   define ('M_PER_AU' , '149597870700');    // NASA/IAU/DE430 as of 2012
   define ('KM_PER_AU', '149597870.700');
   define ('MI_PER_AU', '92955807.273');

// -----------------------------------------------------
// Derived factors based on the above adopted standards.
// The AU used is the IAU value and the year value used
// is the standard Julian year of exactly 365.25 days.

   define ('M_PER_LY'  , '9460730472580800'); // EXACT
   define ('M_PER_PC'  , '30856775814913672.7891393795779647');
   define ('AU_PER_LY' , '63241.0770880709446742');
   define ('AU_PER_PC' , '206264.8062470963551565');
   define ('LT_1AU'    , '0.00577551830441213');

// ----------------------------------
// Define some weight/mass standards.

   define ('KG_PER_OZ' , '0.028349523125'); // EXACT
   define ('KG_PER_LB' , '0.45959237');     // EXACT
   define ('G_PER_OZ'  , '28.349523125');   // EXACT

   define ('N_PER_LB'  , '4.4482216152605'); // DERIVED

// ___________________________________________________
// Length of standard Julian calendar year in seconds.
// exactly 365.25 days.

   define ('S_PER_JYR', '31557600');

// __________________________________
// DEFINE NUMBER OF SECONDS IN 1 DAY.

   define ('S_PER_DAY' , '86400');

// _____________________________________________
// Length of Gregorian calendar year in seconds.
// exactly 365.2425 days.

   define ('S_PER_GYR', '31556952');

// ---------------------------
// Define Planck constant (h).
// 6.62606896E-34

   define ('PLANCK_h', '0.000000000000000000000000000000000662606896'); //  J/s

// -----------------------------------------
// Define Stefan-Boltzmann constant (sigma).
//
// Radiated energy = sigma * K**4 Watts per sq. meter per second
// 5.670373E-8
// Where K = temperature in kelvins

   define ('STEF_BOLT_sigma', '0.00000005670373'); // Watts/sq. meter/k^4

// ------------------------------
// Define Boltzmann constant (k).
//
// k = Joules / deg Kelvin
   define ('BOLTZ_k', '0.000000000000000000000013806504'); //  J/s/K = Watts/K

// ==================================================================
// Define special PI-related constants to 110 decimals (not rounded).
// These are for when higher than standard accuracy is needed.

   define ('PI',           '3.141592653589793238462643383279502884197169399375105820974944592307816406286208998628034825342117067982148');
   define ('PI_OVER_2',    '1.57079632679489661923132169163975144209858469968755291048747229615390820314310449931401741267105853399107404325');
   define ('TWO_PI',       '6.28318530717958647692528676655900576839433879875021164194988918461563281257241799725606965068423413596429617302');
   define ('PI_SQR',       '9.86960440108935861883449099987615113531369940724079062641334937622004482241920524300177340371855223182402591377');
   define ('PI_CUB',       '31.00627668029982017547631506710139520222528856588510769414453810380639491746570603756670103260288619303012196157');
   define ('RAD_PER_DEG',  '0.01745329251994329576923690768488612713442871888541725456097191440171009114603449443682241569634509482212304492');
   define ('DEG_PER_RAD', '57.29577951308232087679815481410517033240547246656432154916024386120284714832155263244096899585111094418622338163');
   define ('SQRT_PI',      '1.77245385090551602729816748334114518279754945612238712821380778985291128459103218137495065673854466541622682362');
   define ('LN_10',        '2.30258509299404568401799145468436420760110148862877297603332790096757260967735248023599720508959829834196778403');

/* ======================================================================
   Define the available solar system body names. These definitions can be
   made GLOBAL to all function modules.  All the names are right-padded
   with white-space to a uniform 10-character field width.
   ======================================================================
*/
   define ('BODY_NAMES', 'Sun       Moon      Mercury   Venus     Mars      Jupiter   Saturn    Uranus    Neptune   Pluto');
   define ('ASTEROID_NAMES', 'Ceres     Cybele    Davida    Eunomia   Europa    Flora     Hebe      Hygiea    InteramniaIris      Juno      Metis     Pallas    Psyche    Vesta');

// ===================================================================
// Define the names of months and days of the week.  All the names are
// right-padded with white-space to a uniform 10-character field width.

   define ('MONTH_NAMES', 'January   February  March     April     May       June      July      August    September October   November  December');
   define ('DOW_NAMES', 'Sunday    Monday    Tuesday   Wednesday Thursday  Friday    Saturday');

/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/





/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ===================================
   STRING AND ARRAY HANDLING FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 STRING/ARRAY-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 SYSTEM MODULES AND RELATED DATA AND
   GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/


/*
   This function removes all external and internal
   white-space from within a text string.
*/
   function Strip_WSpace ($TextString)
{
   return PReg_Replace("/\s+/", " ", trim($TextString));
}





/*
   =========================================
   This function simply swaps (in-place) the
   contents of two simple variable arguments.
*/
   function Swap (&$var1, &$var2)
{
   $w=$var1;  $var1=$var2;  $var2=$w;
}













/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/





/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ================================
   CALENDAR DATE AND TIME FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 CALENDAR DATE AND TIME-RELATED
   FUNCTIONS FOR USE WITH THE PHPSL G2 SYSTEM MODULES AND RELATED
   DATA AND GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/




/*
   This function returns the JDTT value corresponding to the
   given local date/time/zone elements.

   Default time and zone = 00:00:00 UT  and  TZ +00:00
*/

   function YMD_TZ_to_JDTT ($Year, $MonthNum, $DayNum, $LThhmmss, $TZhhmm)
{
   $Y           = IntVal(trim($Year));
   $m           = IntVal(trim($MonthNum));
   $d           = IntVal(trim($DayNum));
   $hhmmss_LT   = trim($LThhmmss);
   $hhmm_TZ     = trim($TZhhmm);
   $Frac_Day_LT = HMS_to_Hours($hhmmss_LT) / 24;
   $Frac_Day_TZ = HMS_to_Hours($hhmm_TZ) / 24;
   $JD_Num_LT   = GregorianToJD($m, $d, $Y);
   $JD_LT       = $JD_Num_LT + $Frac_Day_LT - 0.5;
   $JD_TT       = JDLT_to_JDTT ($JD_LT, $hhmm_TZ);

   return SPrintF("%7.8f", $JD_TT);
}





/*
   This is the inverse Julian Date function.   Given any general
   Gregorian Julian Date, it returns the corresponding full date
   and time string. The fractional part of the Julian Date value
   is used to determine the time of day.

   The time

   This function is NOT the same as the Inv_JD_Num() function,
   which only returns the calendar date string and nothing more.

   The JD Number is ALWAYS an integer value.

   The Julian Date can have a fractional part.  The fractional
   part of the Julian Date value is used to determine the time
   of day.

   EXAMPLES:
   print Inv_JD (2433056.63211234); // = 1949 May 20 Fri 06:31:51
   print Inv_JD (2433056.43211234, 'A'); // = 1949 May 20 Fri 06:32 AM

*/

   function Inv_JD ($JD, $AMPMor24='24')
{
   $J  = FloatVal(trim($JD));  $JDNum = floor($J + 0.5);

   $hhmmssXT = Hours_to_HMS(24*(($J - 0.5) - floor($J - 0.5)), 0, '', ':');

// Adjust returned time according to AM/PM or default 24h mode.
// If not 24, then must be AM/PM mode instead.
   $Ret_Mode = IntVal(trim($AMPMor24));
   if ($Ret_Mode == 0)
      {
       $hhmmssXT = HMS_24_to_12 ($hhmmssXT, FALSE);
      }

   list($m, $d, $Y) = PReg_Split("[\/]", JDtoGregorian($JDNum));

// Do some output formatting.
   $Y   = SPrintF("%04d", $Y);  $m = IntVal($m);  $d = IntVal($d);
   $Mmm = substr("JanFebMarAprMayJunJulAugSepOctNovDec", 3*($m - 1), 3);
   $dn  = SPrintF("%02d", $d);
   $DoW = substr("SunMonTueWedThuFriSat", 3*(($JDNum + 1) % 7), 3);

// Done.
   return "$Y $Mmm $dn $DoW $hhmmssXT";
}






/*
   This function return the calendar date elements (Y m d) corre-
   sponding to the given calendar mode and Julian Day number, or
   JD number, not to be confused with the Julian Date (JD).

   The JD Number is the serial number of the date on the calendar
   and is always an integer value.

   Mathematically, the Julian Day number is numerically equivalent
   to the Julian Date value  at 12:00:00 noon on the corresponding
   calendar date.

   JDNum  = Julian Day Number
   JGMode = 'J' or 'G' = Julian or Gregorian calendar mode

   ReturnMode = 1 to 5
            1 = '127 1 3'
            2 = '+0127 01 03'
            3 = '+0127 Jan 03 Fri'
            4 = ' 1901 Feb 09 Sat'
*/

   function Inv_JD_Num ($JDNum, $JGMode='G', $ReturnMode14=1)
{
   $NumRetModes = 4; // Define number of return modes.

// ERROR if JDNum is NOT numeric or less than zero.
// Otherwise, truncate JDNum value without rounding.
   $J = trim($JDNum);  if (!Is_Numeric($J) or $J < 0) {return FALSE;}
   $J = floor($J);

   $JG = (StrToUpper(substr(trim($JGMode), 0,1)) == 'J')? 'J':'G';

   $r = abs(IntVal(trim($ReturnMode14)));
   if ($r > $NumRetModes) {$RMode = $NumRetModes;} else {$RMode = $r;}

   $M = (StrToUpper(substr($JG,0,1)) == 'J')? 0:1;
   $e = $M*floor(($J - 1867216.25)/36524.25);
   $f = $J + $M*($e - floor($e/4) + 1);
   $g = $f + 1524;
   $h = floor(($g - 122.1)/365.25);
   $i = floor(365.25 * $h);
   $j = floor(($g - $i)/30.6001);
   $d = $g - $i - floor(30.6001 * $j);
   $m = $j - 12*floor($j/14) - 1;
   $y = $h + floor((14-$m)/12) - 4716; // Mathematical year value
   $Y = ($y > 0)? $y : $y-1;           // Calendar year value


// Format returned date string according to return mode value.
   $w   = '';
   $Mmm = substr('JanFebMarAprMayJunJulAugSepOctNovDec', 3*($m-1), 3);
   $DoW = substr('SunMonTueWedThuFriSat', 3*(($J+1) % 7), 3);
   $Y   = substr('     '.SPrintF("%+05d", $Y), -5);

   switch ($RMode)
  {
   case  1: $w = IntVal($Y)." $m $d"; break;
   case  2: $w = SPrintF("%+05d", $Y).' '.SPrintF("%02d", $m).' '.SPrintF("%02d", $d); break;
   case  3: $w = "$Y $Mmm ".SPrintF("%02d", $d)." $DoW"; break;

   case  4: $w = "$Y $Mmm ".SPrintF("%02d", $d)." $DoW";
            $w = Str_Replace('+', '', $w);
            break;

   default: $w = IntVal($Y)." $m $d"; break;
  }
   return $w;

} // End of Inv_JD_Num (JDNum, ...)




/*
   Given a JDTT value and local time zone offset from UT,
   this function returns a table of  equivalent times in
   LT, UT and TT and the value of Delta T. This could be
   the Julian Date of some astronomical event.

   This function uses the NASA polynomial expressions for
   Delta T.
*/

   function JDTT_to_Times_Table ($JDTTArg, $TZhhmm='+00:00')
{
   $JDTT = FloatVal(trim($JDTTArg));

// Compute fraction of day corresponding to
// the Time Zone value.
   $FracDayTZ = HMS_to_Hours(trim($TZhhmm))/24;

// Determine longitudinal or TZ West|East marker
   $WE = ($FracDayTZ < 0)?'(West)':'(East)';
   if ($FracDayTZ == 0) {$WE = '      ';}

// Get local Time Zone offset from UT as "hh:mm" string.
   $hhmmTZ = substr(Hours_to_HMS(24*$FracDayTZ, 0, '+', ':'),0,6);

// Compute fraction of day corresponding to
// the Delta T value.
   $FracDayDT = NASA_Delta_T_Poly($JDTT)/86400;

// Get Delta T as HMS format string.
   $hhmmssDT = Hours_to_HMS(24*$FracDayDT, 0, '+', ':');

// Compute Julian Dates for UT and TT.
   $JDUT = $JDTT - $FracDayDT;
   $JDLT = $JDUT + $FracDayTZ;

// Compute Date/Time strings corresponding
// to Julian Dates LT, UT and TT.
   $DateTimeLT = Inv_JD($JDLT);
   $DateTimeUT = Inv_JD($JDUT);
   $DateTimeTT = Inv_JD($JDTT);

// Compute Julian Day Numbers for LT, UT and TT.
   $JDNumLT = floor($JDLT + 0.5);
   $JDNumUT = floor($JDUT + 0.5);
   $JDNumTT = floor($JDTT + 0.5);

// Do some formatting.
   $FracDayDT = SPrintF("%+1.8f", $FracDayDT);
   $FracDayTZ = SPrintF("%+1.8f", $FracDayTZ);

   $JDLT = SPrintF("%7.8f", $JDLT);
   $JDUT = SPrintF("%7.8f", $JDUT);
   $JDTT = SPrintF("%7.8f", $JDTT);

// Done. Return times table text.
   return
"------------------------------------------------------------------------------
    Julian_Date         Time_Scale      Calendar_Date      Time     JD_Number
-------------------  ----------------  --------------- ------------ ---------
$JDLT LT  LOCAL DATE/TIME   $DateTimeLT LT   $JDNumLT
     $FracDayTZ TZ  Time Zone $WE                 $hhmmTZ:00 TZ
$JDUT UT  Universal Time    $DateTimeUT UT   $JDNumUT
     $FracDayDT DT  NASA Delta T                     $hhmmssDT DT
$JDTT TT  Terrestrial Time  $DateTimeTT TT   $JDNumTT";

} // end of  JDTT_to_Times_Table (...)





/*
   This function translates an assumed time string into standard form.
   Applies ONLY to purely numerical time strings.


   Time String          Translates to
   1                    01:00:00
   12                   12:00:00
   123                  01:23:00
   1234                 12:34:00
   12345                01:23:45
   123456               12:34:56

Numeric strings are read in 2-digit groups from R to L

Odd length numbers are converted to even length and then
the even length numbers are converted to 6 digit values.

THIS FUMCTION DOES NOT CHECK FOR INVALID TIMES.  IT ONLY
CASTS THE GIVEN NUMERIC TIME STRING IN NORMAL HH:MM:SS
FORMAT AS DEFINED ABOVE.

*/

   function Num_to_HMS ($NumHMSVal)
{
   $w = trim($NumHMSVal);

   if ($w == '') {return '00:00:00';}

// Return input unchanged if non-numeric argument
// rather than return an error message.
   if (!Is_Numeric($w)) {return $NumHMSVal;}

   $FirstChar = substr($w,0,1);

   $NumSign = (IntVal($w) < 0)? '-':'+';

// Remove any numerical (±) sign
// and work with absolute value.
   $w = Str_Replace($NumSign, '', $w);

// Construct (w) into a 6-digit integer.
   if (StrLen($w) == 1) {$w = "0$w";}
   if (StrLen($w) == 2) {$w = $w."0000";}
   if (StrLen($w) == 3) {$w = "0$w";}
   if (StrLen($w) == 4) {$w = $w."00";}
   if (StrLen($w) == 5) {$w = "0$w";}

// Only keep (+) sign if it was in the original argument.
// Otherwise, a positive value will NOT be returned with
// a (+) sign. This does not effect the sign of negative
// values.
   if ($FirstChar <> '+' and $NumSign == '+') {$NumSign = '';}

// Done.  Return time equivalent with colons.
   return $NumSign.substr($w,0,2).':'.substr($w,2,2).':'.substr($w,4,2);

} // END OF  Num_to_HMS (...)





/*
  This function converts 24-hour time to 12-houre AM/PM time

  EXAMPLES:  17:25:45 converts to 05:25:45 PM
             00:01:23 converts to 12:01:23 AM
             00:15:45 converts to 12:16 AM

*/

   function HMS_24_to_12 ($hhmmss24, $SecondsFlag=FALSE)
{
   $hhmmss = trim($hhmmss24);

   $SecFlag = ($SecondsFlag !== TRUE)? FALSE:TRUE;

// Convert 24-hour time to decimal hours.
   $Hours24 = HMS_to_Hours($hhmmss);

   $hhmmss12 = $AMPM = '';

   if ($Hours24 < 1)
   {
    $hhmmss12 = Hours_to_HMS($Hours24+12, 0, '', ':');
    $AMPM = 'AM';
   }

   if ($Hours24 >= 1 and $Hours24 < 12)
   {
    $hhmmss12 = Hours_to_HMS($Hours24, 0, '', ':');
    $AMPM = 'AM';
   }

   if ($Hours24 == 12) // check
   {
    $hhmmss12 = Hours_to_HMS($Hours24, 0, '', ':');
    $AMPM = 'PM';
   }

   if ($Hours24 > 12)
   {
    $hhmmss12 = Hours_to_HMS($Hours24-12, 0, '', ':');
    $AMPM = 'PM';
   }

// Return long 24h hh:mm:ss format to nearest second.
   if ($SecFlag) {return "$hhmmss12 $AMPM";}

// Return short 12h hh:mm AM/PM format to nearest minute.
   $Hours12 = HMS_to_Hours($hhmmss12);
   $hh = floor($Hours12);
   $min = round(60*($Hours12 - $hh), 0);
   if ($min == 60){$Hours12 += 1;  $min=0;}
   $hhmmss12 = "$hh:$min:00";
   $Hours12 = HMS_to_Hours($hhmmss12);
   $hhmmss12 = Hours_to_HMS($Hours12, 0, '', ':');
   $hhmmss12 = substr($hhmmss12, 0,5);

   if (substr($hhmmss12,0,3) == '00:')
      {$hhmmss12 = '12:'.substr($hhmmss12, 3, StrLen($hhmmss12));}

   return "$hhmmss12 $AMPM";

} // end of  HMS_24_to_12 (hhmmss24, SecondsFlag=FALSE)








/*
   This function returns the equivalent H:M:S time string with
   several formatting options.

  hours         = ± Decimal hours value
  ssDecimals    = Number of decimals in seconds part

  PosSign = Symbol to use for positive values ('' or '+')
      ''  = Empty = Return numbers only, no symbols.
      '+' = Attach '+' sign to positive values.
            Has no effect on negative values.

  SymbolsMode   = If or not to attach time symbols (h m s) or (:)
                  'h' = '01h 02m 03s' (Default)
                  ':' = '01:02:03'
                  ''  = '01 02 03'
*/

   function Hours_to_HMS ($hours, $ssDecimals=0, $PosSignSymbol='', $SymbolsMode='h')
{
// Initialize symbol carriers.
   $_h_ = $_m_ = $_s_ = '';

   if (trim($PosSignSymbol) == '')  {$PosSignSymbol = FALSE;}

// Remember original numerical sign and work with absolute value.
   $sign = ($hours < 0)? '-' : '';    $hours = abs($hours);

// Remember numerical sign to be restored on exit, if any.
   if (($PosSignSymbol === TRUE or $PosSignSymbol == '+') and $sign == '')
       {$sign = '+';}

// Compute time elements from absolute hours argument.
   $hh = floor($hours);
   $minutes = 60*($hours - $hh);    $mm = floor($minutes);
   $seconds = 60*($minutes - $mm);  $ss = SPrintF("%1.3f",  $seconds);

// Format the time elements.
   $hh = SPrintF("%02d",  $hh);
   $mm = SPrintF("%02d",  $mm);
   $ss = SPrintF("%1.$ssDecimals" . "f", $ss);

// Patch for that blasted 60s glitch.
   if ($ss == 60) {$ss = 0; $mm++;}
   if ($mm == 60) {$mm = 0; $hh++;}

   $hh = SPrintF("%02d",  $hh);
   $mm = SPrintF("%02d",  $mm);

   $ss = SPrintF("%1.$ssDecimals" . "f", $ss);
   if ($ss < 10) {$ss = "0$ss";}

// Attach optional (h m s :) symbols as indicated.
// Default = 'h'
   if ($SymbolsMode == 0 or $SymbolsMode == '')
      {$_h_  =  $_m_  =  $_s_  = ' ';}

   if ($SymbolsMode == 1 or $SymbolsMode == ':')
      {$_h_  =  ':';  $_m_  =  ':';  $_s_  = '';}

   if ($SymbolsMode == 2 or StrToLower($SymbolsMode) == 'h')
      {$_h_  =  'h ';  $_m_  =  'm ';  $_s_  =  's';}

   $w = "$sign$hh$_h_$mm$_m_$ss$_s_";

// Done.
   return $w;

} // End of  Hours_to_HMS(...)








   function HMS_to_Hours ($hhmmss)
{
// Read hour angle string argument and force to lower case in
// the event any (h m s :) symbols are used.
   $hms = StrToLower(trim($hhmmss));

// Convert any (h m s :) symbols into spaces.
   $hms = Str_Replace('h', ' ', $hms);
   $hms = Str_Replace('m', ' ', $hms);
   $hms = Str_Replace('s', ' ', $hms);
   $hms = Str_Replace(':', ' ', $hms);

// Normalize the spacing and then split and extract the
// individual time string elements (hh, mm,ss).
   $hms = PReg_Replace("/\s+/", " ", trim($hms));
   $whms = PReg_Split("[ ]", $hms);
   $whmsCount = count($whms);
   $hh = ($whmsCount >= 1)? BCAdd($whms[0],"0", 20) : "0";
   $mm = ($whmsCount >= 2)? BCAdd($whms[1],"0", 20) : "0";
   $ss = ($whmsCount >= 3)? BCAdd($whms[2],"0", 20) : "0";

// Remember and then remove any numerical (±) sign.
   $NumSign = (substr($hhmmss,0,1) == '-')? '-' : '';
   $hh = Str_Replace('-', '', $hh);
   $hh = Str_Replace('+', '', $hh);

// If original angle argument began with a + sign, then preserve
// it so that all returned positive results will have a + sign.
// Otherwise, positive results will NOT have a + sign.
   if (substr($hhmmss,0,1) == '+') {$NumSign = '+';}

// Compute decimal hours value equivalent
// to the combined hh,mm,ss values.
   $w2 = BCAdd(BCAdd(BCMul($hh,'3600',20),BCMul($mm,'60',20),20),$ss,20);

// If result equates to zero, then suppress any numerical sign.
   if (BCComp($w2, '0', 20) == 0) {$NumSign = '';}

// Round off result to 16 decimals, recalling numerical sign.
   $w =  BCAdd(BCDiv($w2,3600, 20), '0.00000000000000005', 16);

// Trim off any redundant zeros/decimal point.
   return $NumSign . RTrim(RTrim($w, '0'), '.');

} // End of  HMS_to_Hours(...)




/*
   This function converts TT into LT for the given TZ.
   It is the inverse of the JDLT_to_JDTT() function.

   Given (TT, DT)
   UT = TT - (DT)
   LT = UT + (TZ)
*/
   function JDTT_to_JDLT ($JDTT, $TZhhmm)
{
   $JD_TT = FloatVal(trim($JDTT));
   $hhmm_TZ = trim($TZhhmm);

   $Frac_Day_TZ = HMS_to_Hours($hhmm_TZ) / 24;
   $Frac_Day_DT = NASA_Delta_T_Poly($JD_TT) / 86400;

   $JD_UT = $JD_TT - $Frac_Day_DT;
   $JD_LT = $JD_UT + $Frac_Day_TZ;

   return SPrintF("%7.8f", $JD_LT);
}







/*
   This function converts (LT, TZ) into corresponding TT.
   It is the inverse of the JDTT_to_JDLT() function.

   Given (LT, DT)
   UT = LT - (TZ)
   TT = UT + (DT)
*/
   function JDLT_to_JDTT ($JDLT, $TZhhmm)
{
   $JD_LT = FloatVal(trim($JDLT));
   $hhmm_TZ = trim($TZhhmm);

   $Frac_Day_TZ = HMS_to_Hours($hhmm_TZ) / 24;
   $Frac_Day_DT = NASA_Delta_T_Poly($JD_LT) / 86400;

   $JD_UT = $JD_LT - $Frac_Day_TZ;
   $JD_TT = $JD_UT + $Frac_Day_DT;

   return SPrintF("%7.8f", $JD_TT);
}










/*
   This function finds the Nth DoW
*/

   function Nth_DoW_of_mY ($n, $DoW, $m, $y)
{
   $JDNum1 = GregorianToJD ($m, 1, $y);
   $DoW1   = ($JDNum1 + 1) % 7;
   $dDoW   = $DoW - $DoW1;
   $JDNumX = $JDNum1 + (($dDoW < 0)? $dDoW + 7 : $dDoW) + 7*($n-1);

   list ($m,$d,$y) = PReg_Split("[\/]", JDtoGregorian($JDNumX));

   $yyyymndn = SPrintF('%04d', $y).SPrintF('%02d', $m).SPrintF('%02d', $d);

   return $yyyymndn; // Return integer-encoded date.

} // End of  Nth_DoW_of_Ym(...)


/*
   Thisa function finds the last (DoW) of any month of any
   given year, where DoW refers to the day of the week.

   mn = Month number (1 to 12
   DoW values:
   0=Sun, 1=Mon, 2=Tue, 3=Wed, 4=Thu, 5-Fri and 6=Sat

*/

   function Last_DoW_of_mY ($DoW, $mn, $yyyy)
{
// --------------------------
// READ MONTH/YEAR ARGUMENTS.

   $m = IntVal($mn);
   $y = IntVal($yyyy);

// ---------------------------------------------
// FIND JD NUMBER OF 1ST DoW OF FOLLOWING MONTH.
// IF DECEMBER, WE FIND THE JD NUMBER OF THE 1ST
// DOW OF JANUARY OF THE FOLLOWING YEAR AND THEN
// SUBTRACT 7.

   $m++;

   if ($m == 12) {$m = 1;   $y++;}

   $JDNum1 = GregorianToJD ($m, 1, $y);
   $DoW1   = ($JDNum1 + 1) % 7;
   $dDoW   = $DoW - $DoW1;

// ---------------------------------------------
// COMPUTE JD NUMBER OF LAST DOW OF GIVEN MONTH.

   $JDNumX = $JDNum1 + (($dDoW < 0)? $dDoW + 7 : $dDoW) - 7;

// ------------------------------------------------
// CONSTRUCT AND RETURN INTEGER-ENCODED DATE VALUE.

   list
  (
   $m,
   $d,
   $y
  ) = PReg_Split("[\/]", JDtoGregorian($JDNumX));

   return SPrintF('%04d', $y).SPrintF('%02d', $m).SPrintF('%02d', $d);

} // End of  Last_DoW_of_Ym(...)








/*
   This function translates an events JD TT table to
   local time (LT) table for any given TZ offset.

*/

   function Translate_JDTT_Table_to_JDLT ($JDTTTable, $TZhhmm)
{
   $JD_TT_Table = trim($JDTTTable);
   $hhmm_TZ = trim($TZhhmm);

   $wArray = PReg_Split("[\n]", $JD_TT_Table);
   $wCount = count($wArray);
// 2458485.57218620 Saturn AG
   $wTable = '';
   for($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (substr($Curr_Line, 0,1) == '2')
  {
   $JD_TT = FloatVal($Curr_Line);
   $Body_Name_and_Mark = substr($Curr_Line, 17, StrLen($Curr_Line));
   $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);
   $JD_TT = SPrintF ("%7.8f", $JD_TT);
   $JD_LT = SPrintF ("%7.8f", $JD_LT);
   $Date_Time_LT = trim(Inv_JD($JD_LT));
   $wTable .= "$JD_LT $Date_Time_LT $Body_Name_and_Mark\n";
  }
  }
   $wTable = trim($wTable);

   return $wTable;
} // end of  Translate_JDTT_Table_to_JDLT (JDTTTable, TZhhmm)


















// ***********************************************
// Volatile alternate special translation utility.

/*
  This function translates JDTT table to the
  corresponding JDLT table.

  Replaces Date/Time TT with Date/Time LT for given TZ.

2458486.72269915 2019 Jan 03 Thu 05:20:41  0.983301164 AU - Earth Perihelion
*/

   function JDTT_Table_to_JDLT ($JDTTTableText, $TZhhmm='+00:00')
{
   $JDTT_Table = $JDTTTableText; // DO NOT TRIM, USE AS-IS

   $TT = StrToUpper(trim($TZhhmm));
   if ($TT == 'TT')
      {
       $TT = 0;
       $TZhhmm = '0';
//       $JDTT_Table = Str_Replace('UT', 'TT', $JDTT_Table);
      }
   else
      {$TT = 1;}

   $Hours_TZ = HMS_to_Hours(trim($TZhhmm));
   $hhmm_TZ = Hours_to_HMS($Hours_TZ);
   $Frac_Day_TZ = $Hours_TZ / 24;

   $wArray = PReg_Split("[\n]", trim($JDTTTableText));
   $wCount = count($wArray);

   $wTable = '';
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

// Process the current line ONLY if
// the first char is the digit '2'.
   if (substr($Curr_Line,0,1) == '2')
      {
       $JD_TT = substr($Curr_Line, 0, 16);
       $JD_Num_TT = floor($JD_TT + 0.5);
       $Event_Str = trim(substr($Curr_Line, 41, StrLen($Curr_Line)));

       $Frac_Day_DT = NASA_Delta_T_Poly($JD_TT) / 86400;

       $JD_UT = $JD_TT - $TT * $Frac_Day_DT;
       $JD_LT = $JD_UT + $TT * $Frac_Day_TZ;

       $JD_LT = SPrintF("%7.8f", $JD_LT);

       if ($TT == 0)
          {$Date_Time_LT = Inv_JD($JD_LT);}
       else
          {$Date_Time_LT = Inv_JD($JD_LT, 'AMPM');}

       $Date_LT_Part = substr($Date_Time_LT, 0, 15);
       $Time_LT_Part = trim(substr($Date_Time_LT, 15, StrLen($Date_Time_LT)));

       if (substr($Time_LT_Part, 0,1) == '0')
          {$Time_LT_Part = ' '.trim(substr($Time_LT_Part, 1, StrLen($Time_LT_Part)));}
       $DT_LT = "$Date_LT_Part $Time_LT_Part";

       $wTable .= "$JD_LT $DT_LT $Event_Str\n";
      }
  }

  return ($wTable);

} // end of  JDTT_Table_to_JDLT (JDTTTableText, TZhhmm)






/*
   =========================================
   This function converts decimal days into
   a days, hours, minutes and seconds string
   in the format 'd hh mm ss'
   Example:
   367.10314836 returns '367d 02h 28m 32s'
*/
   function Days_to_DHMS ($Days, $ssDecimals=0)
{
   $days = trim($Days);
   $sign = ($days < 0)? '-':'';
   $days = Str_Replace('+','', Str_Replace('-','', $days));

   $d = 'd';

   $decimals = abs(IntVal(trim($ssDecimals)));
   if ($decimals > 28) {$decimals = 28;}

   $dd = BCAdd('0',$days);
//   $hours = 24*($days - $dd);
   $hours = BCMul('24', BCSub($days,$dd,$decimals),$decimals);
   $hh = BCAdd('0',$hours);
   $minutes = BCMul('60', BCSub($hours,$hh,$decimals),$decimals);
   $mm = BCAdd('0',$minutes);
   $seconds = BCMul('60', BCSub($minutes,$mm,$decimals),$decimals);

   $hh = Sprintf("%02d", $hh).'h';
   $mm = Sprintf("%02d", $mm).'m';
   $ss = BCAdd('0',$seconds, $decimals).'s';
   if ($ss < 10) {$ss = "0$ss";}

   return "$sign$dd$d $hh $mm $ss";
}

/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

   function Seconds_to_WDHMS ($Seconds, $Decimals=0)
{
   $sec     = trim($Seconds);
   $dec     = trim($Decimals);
   $Q       = $dec + 32;

   $weeks   = BCDiv($sec,'604800',$Q);
   $wk      = BCAdd('0',$weeks);

   $days    = BCMul('7',BCSub($weeks,$wk,$Q),$Q);
   $dd      = BCAdd('0',$days);

   $hours   = BCMul('24',BCSub($days,$dd,$Q),$Q);
   $hh      = BCAdd('0',$hours);

   $minutes = BCMul('60',BCSub($hours,$hh,$Q),$Q);
   $mm      = BCAdd('0',$minutes);

   $seconds = BCMul('60',BCSub($minutes,$mm,$Q),$Q);
   $ss      = BCAdd(('0.'.Str_Repeat('0',$dec).'5'),$seconds,$dec);

// Correct 60s glitch
   if (BCComp('60',$ss,$dec) == 0)
      {$ss = BCAdd('0','0',$dec); $mm += 1;}
   if (BCComp('60',$mm,$dec) == 0)
      {$mm = 0; $hh += 1;}
   if (BCComp('24',$hh,$dec) == 0)
      {$hh = 0; $dd += 1;}

// Do some formatting
   $dd = $dd.'d';
   $hh = SPrintF("%02d", $hh).'h';
   $mm = SPrintF("%02d", $mm).'m';
   if ($ss < 10){$ss = "0$ss";}
   $ss = $ss.'s';
   $wk = $wk.'wk';

   return "$wk $dd $hh $mm $ss";

}








/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ==========================================
   METRICS AND TEMPERATURE HANDLING FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 METRICS-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 DE405 SOLAR SYSTEM MODULE AND ANY
   RELATED DATA AND GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode
*/


/*
 ===========================================================================
 This function interconverts between any of eighteen length or distance
 units.  The length/distance units symbols are NOT case sensitive.

 This is a standard double-precision program.

 ----------
 ARGUMENTS:
 $LengthAndUnits = Numerical value and units symbol of length to convert.

                   IMPORTANT NOTE: There MUST be at least one space between
                   the number value and the units symbol in the string.

 $ToUnitsStr     = The units symbol corresponding to the computed output.

 -------
 ERRORS:
 On error FALSE is returned.
 An error occurs if the length value is non-numeric or either
 of the input or output length units symbols are unrecognized.

 ------------------------------------------------------------
 THE 18 LENGTH UNITS SYMBOLS RECOGNIZED BY THIS FUNCTION ARE:

 SYMBOL      UNITS

   A         Angstroms
   nm        Nanometers
 um or u     Micrometers (or Microns)
   mm        Millimeters
   cm        Centimeters
   in        Inches
   ft        Feet
   yd        Yards
   m         Meters
   km        Kilometers
   mi        Miles (statute)
   nmi       Nautical miles
   LS        Light Seconds
   LM        Light Minutes
   LH        Light Hours
   AU        Astronomical Units
   LY        Light Years (Based on Julian year of 365.25 days)
   pc        Parsecs

   ===========================================================================
*/

   function Leng_to_Leng ($LengthAndUnits, $ToUnitsStr)
{
// Define conversion constants for internal use only.
// These constants could also be defined externally
// for global use by any programs that need them.
//
// ______________________________________________
// This function was designed to be transportable
// and fully self-contained, constants and all.

   $M_PER_A   = '0.0000000001';
   $M_PER_NM  = '0.000000001';
   $M_PER_UM  = '0.000001';
   $M_PER_MM  = '0.001';
   $M_PER_CM  = '0.01';
   $M_PER_IN  = '0.0254';
   $M_PER_FT  = '0.3048';
   $M_PER_YD  = '0.9144';
   $M_PER_M   =  1;
   $M_PER_KM  = '1000';
   $M_PER_MI  = '1609.344';
   $M_PER_NMI = '1852';
   $CLIGHT    = '299792458';        // Light Second
   $M_PER_LM  = '17987547480';      // Light Minute
   $M_PER_AU  = '149597870700';
   $M_PER_LH  = '1079252848800';    // Light Hour
   $M_PER_LY  = '9460730472580800'; // Light Year
   $M_PER_PC  = '30856775876793114.6633';

/*
   ___________________________________________________________________
   Parse input value and separate the numerical value from the symbol.

   Read numeric value and then delete it from the string, leaving only
   the units symbol behind, if the number value was valid.

   NOTE: Scientific notation is NOT allowed, only regular numbers.

   Revision.  Strip white space.  Read string backwards to find point
   where it becomes a numeric string.

   start '12.304nmi' = non-numeric
   next  '12.304nm'  = non-numeric
   next  '12.304n'   = non-numeric
   next  '12.304'    = numeric  ;  units = 'nmi'

   start '12.304  km' = non-numeric
   next  '12.304  k'  = non-numeric
   next  '12.304  '   = non-numeric
   next  '12.304'     = numeric  ;  units = 'km'





*/


   $v = FloatVal(trim($LengthAndUnits));
   $u = trim(Str_Replace($v, '', $LengthAndUnits));

   $w = "$v $u";

   if (StrPos($w, ' ') === FALSE){return FALSE;}

   list($x, $FromUnits) = PReg_Split("[ ]", PReg_Replace("/\s+/", ' ', $w));

   if (!Is_Numeric($x)) {return FALSE;}

// _____________________________________________________________
// Get input units symbol. Units symbols are NOT case sensitive.

   $FromUnitsSymbol = StrToUpper(trim($FromUnits));

// ________________________________________
// Determine which conversion factor to use
// according to the units symbol argument.

   switch ($FromUnitsSymbol)
  {
   case 'A'  : $f = $M_PER_A;   break;
   case 'NM' : $f = $M_PER_NM;  break;
   case 'UM' : $f = $M_PER_UM;  break;
   case 'U'  : $f = $M_PER_UM;  break;
   case 'MM' : $f = $M_PER_MM;  break;
   case 'CM' : $f = $M_PER_CM;  break;
   case 'IN' : $f = $M_PER_IN;  break;
   case 'FT' : $f = $M_PER_FT;  break;
   case 'YD' : $f = $M_PER_YD;  break;
   case 'M'  : $f = $M_PER_M;   break;
   case 'KM' : $f = $M_PER_KM;  break;
   case 'MI' : $f = $M_PER_MI;  break;
   case 'NMI': $f = $M_PER_NMI; break;
   case 'LS' : $f = $CLIGHT;    break;
   case 'LM' : $f = $M_PER_LM;  break;
   case 'LH' : $f = $M_PER_LH;  break;
   case 'AU' : $f = $M_PER_AU;  break;
   case 'LY' : $f = $M_PER_LY;  break;
   case 'PC' : $f = $M_PER_PC;  break;
   default   : $f = FALSE;      break;
  }

// __________________________________________
// Error if conversion factor is FALSE, which
// means that the units symbol was not found
// in the table of valid symbols.

   if ($f === FALSE) {return $f;}

// ______________________________________________
// Otherwise, compute the equivalent meters value
// by multiplying the value of the argument by
// the conversion factor (f).

   $m = $x * $f;

// *********************************************************
// AT THIS POINT INPUT VALUE HAS BEEN CONVERTED INTO METERS.
// NEXT STEP IS TO CONVERT THE METERS INTO THE OUTPUT UNITS.
// *********************************************************

// ________________________________________________________
// Read the output units symbol argument.  The unit symbols
// are NOT case sensitive.

   $ToUnitsSymbol = StrToUpper(trim($ToUnitsStr));

// _____________________________________________________
// Determine the conversion factor to use from the units
// symbol argument.

   switch ($ToUnitsSymbol)
  {
   case 'A'  : $f = $M_PER_A;   break;
   case 'NM' : $f = $M_PER_NM;  break;
   case 'UM' : $f = $M_PER_UM;  break;
   case 'U'  : $f = $M_PER_UM;  break;
   case 'MM' : $f = $M_PER_MM;  break;
   case 'CM' : $f = $M_PER_CM;  break;
   case 'IN' : $f = $M_PER_IN;  break;
   case 'FT' : $f = $M_PER_FT;  break;
   case 'YD' : $f = $M_PER_YD;  break;
   case 'M'  : $f = $M_PER_M;   break;
   case 'KM' : $f = $M_PER_KM;  break;
   case 'MI' : $f = $M_PER_MI;  break;
   case 'NMI': $f = $M_PER_NMI; break;
   case 'LS' : $f = $CLIGHT;    break;
   case 'AU' : $f = $M_PER_AU;  break;
   case 'LY' : $f = $M_PER_LY;  break;
   case 'PC' : $f = $M_PER_PC;  break;
   default   : $f = FALSE;      break;
  }

// __________________________________________
// Error if conversion factor is FALSE, which
// means that the units symbol was not found
// in the valid symbols table.

   if ($f == FALSE) {return FALSE;}

// _____________________________________________
// Otherwise, return the equivalent of the given
// meters value by dividing the value of the
// argument by the conversion factor (f).

   return $m / $f;

} // END OF  Leng_To_Len(...)






/*
____________________________________________________________
TEMPERATURE SCALE INTERCONVERSION FUNCTION.

This function converts from any given temperature
scale into the equivalent temperature on any one
of four recognized (F, C, K, R) scales.

The four recognized scale symbols are:
'F' = Fahrenheit
'C' = Celsius
'K' = Kelvin  = Degrees Celsius above absolute zero.
'R' = Rankine = Degrees Fahrenheit above absolute zero.

The scale symbols are NOT case sensitive.

The conversion is performed by first converting the
input temperature value into its equivalent on the K
scale, then that K scale value is converted into the
specified output scale.


EXAMPLE:
To convert from 123.45 C into its equivalent
on the R scale:

print Temp_to_Temp('-123.45 C', 'R'); // Result = 269.46

ERRORS:
An error returns FALSE if a scale symbol is not recognized.
___________________________________________________________
*/

   function Temp_to_Temp ($FromTempValueAndScaleStr, $ToScaleStr)
{
// _______________________________________________
// Read temperature argument and conversion scale.

   $w = trim($FromTempValueAndScaleStr);   if ($w == '') {$w = '0';}
   $T = FloatVal($w);

   $FromScale = StrToUpper(trim(substr($w, -1)));
   $ToScale   = StrToUpper(trim($ToScaleStr));

// __________________________________________________
// Convert input (FromScale) into K-scale equivalent.

   switch ($FromScale)
  {
   case 'F' : $K = (5 * ($T - 32) / 9) + 273.15; break;
   case 'C' : $K = $T + 273.15; break;
   case 'K' : $K = $T; break;
   case 'R' : $K = (5 * ($T - 491.67) / 9) + 273.15; break;

   default  : $K = FALSE; break; // = ERROR.
  }

// ____________________________
// ERROR if invalid (FromScale)

   if ($K === FALSE) {return $K;}

// _________________________________________________
// Convert K value into output (ToScale) equivalent.

   switch (StrToUpper(trim($ToScale)))
  {
   case 'F' : $T = (9 * ($K - 273.15) / 5) + 32; break;
   case 'C' : $T = $K - 273.15; break;
   case 'K' : $T = $K; break;
   case 'R' : $T = (9 * ($K - 273.15) / 5) + 491.67; break;

   default  : $T = FALSE; break; // = ERROR.
  }

// Done.
   return $T;

} // end of  Temp_To_Temp(...)




/*
   This function simply converts from one
   unit of mass/weight force into another.

   Recognized units

   ug      Micrograms
   mg      Milligrams
   g       grams
   oz      ounces (avdp)
   lb      pounds (avdp)
   kg      kilograms
   N       Newtons
   ton     tons

*/

   function Mass_to_Mass ($MassAndUnits, $ToUnitsStr)
{
   $r = '0.0000000000000000000000005';

// Define conversion constants for internal use only.
// These constants could also be defined externally
// for global use by any programs that need them.
//
// ______________________________________________
// This function was designed to be transportable
// and fully self-contained, constants and all.

   $KG_PER_UG  = '0.000000001';
   $KG_PER_MG  = '0.000001';
   $KG_PER_G   = '0.001';
   $KG_PER_OZ  = '0.028349523125';
   $KG_PER_LB  = '0.45359237';
   $KG_PER_KG  = '1.0';
   $KG_PER_TON = '907.18474';
   $KG_PER_N   = BCDiv('1', '9.80665', 28);

   list($v, $u)
   = PReg_Split("[ ]", PReg_Replace("/\s+/", " ", trim($MassAndUnits)));

   $w = "$v $u";

   list($x, $FromUnits) = PReg_Split("[ ]", PReg_Replace("/\s+/", ' ', $w));

   if (!Is_Numeric($x)) {return FALSE;}

// ______________________________________________
// Get input units symbol. Force to lower case so
// that the units symbols are NOT case sensitive.

   $FromUnitsSymbol = StrToLower(trim($FromUnits));

// ________________________________________
// Determine which conversion factor to use
// according to the units symbol argument.

   switch ($FromUnitsSymbol)
  {
   case 'ug'  : $f = $KG_PER_UG;  break;
   case 'mg'  : $f = $KG_PER_MG;  break;
   case 'g'   : $f = $KG_PER_G;   break;
   case 'oz'  : $f = $KG_PER_OZ;  break;
   case 'lb'  : $f = $KG_PER_LB;  break;
   case 'kg'  : $f = $KG_PER_KG;  break;
   case 'n'   : $f = $KG_PER_N;   break;
   case 'ton' : $f = $KG_PER_TON; break;
   default    : $f = FALSE;       break;
  }
// __________________________________________
// Error if conversion factor is FALSE, which
// means that the units symbol was not found
// in the table of valid symbols.

   if ($f === FALSE) {return $f;}

// _________________________________________________
// Otherwise, compute the equivalent kilograms value
// by multiplying the value of the argument by the
// conversion factor (f).

   $kg = BCMul($x, $f,  28);


// ************************************************************
// AT THIS POINT INPUT VALUE HAS BEEN CONVERTED INTO KILOGRAMS.
// NEXT STEP IS TO CONVERT THE KILOGRAMS INTO THE OUTPUT UNITS.
// ************************************************************

// ________________________________________________________
// Read the output units symbol argument.  The unit symbols
// are NOT case sensitive.

   $ToUnitsSymbol = StrToLower(trim($ToUnitsStr));

// _____________________________________________________
// Determine the conversion factor to use from the units
// symbol argument.

   switch ($ToUnitsSymbol)
  {
   case 'ug'  : $f = $KG_PER_UG;  break;
   case 'mg'  : $f = $KG_PER_MG;  break;
   case 'g'   : $f = $KG_PER_G;   break;
   case 'oz'  : $f = $KG_PER_OZ;  break;
   case 'lb'  : $f = $KG_PER_LB;  break;
   case 'kg'  : $f = $KG_PER_KG;  break;
   case 'n'   : $f = $KG_PER_N;   break;
   case 'ton' : $f = $KG_PER_TON; break;
   default    : $f = FALSE;       break;
  }

// ==========================================
// Error if conversion factor is FALSE, which
// means that the units symbol was not found
// in the valid symbols table.

   if ($f == FALSE) {return FALSE;}

// =============================================
// Otherwise, return the equivalent of the given
// kilograms value by dividing the value of the
// argument by the conversion factor (f).

   $w = BCAdd($r, BCDiv($kg, $f, 28), 24);

   return RTrim(RTrim($w, '0'), '.');

} // end of  Mass_to_Mass (MassAndUnits, ToUnitsStr)
















// *****************
// Length to length for 64-decimal arbitrary precision.

/*
 ===========================================================================
 This function interconverts between any of eighteen length or distance
 units.  The length/distance units symbols are NOT case sensitive.

 This is a 64-decimal arbitrary precision program.

 ----------
 ARGUMENTS:
 $LengthAndUnits = Numerical value and units symbol of length to convert.

                   IMPORTANT NOTE: There MUST be at least one space between
                   the number value and the units symbol in the string.

 $ToUnitsStr     = The units symbol corresponding to the computed output.

 -------
 ERRORS:
 On error FALSE is returned.
 An error occurs if the length value is non-numeric or either
 of the input or output length units symbols are unrecognized.

 ------------------------------------------------------------
 THE 18 LENGTH UNITS SYMBOLS RECOGNIZED BY THIS FUNCTION ARE:

 SYMBOL      UNITS

   A         Angstroms
   nm        Nanometers
 um or u     Micrometers (or Microns)
   mm        Millimeters
   cm        Centimeters
   in        Inches
   ft        Feet
   yd        Yards
   m         Meters
   km        Kilometers
   mi        Miles (statute)
   nmi       Nautical miles
   LS        Light Seconds
   LM        Light Minutes
   LH        Light Hours
   AU        Astronomical Units
   LY        Light Years (Based on Julian year of 365.25 days)
   pc        Parsecs

   ===========================================================================
*/

   function BCLen_To_Len ($LengthAndUnits, $ToUnitsStr, $Decimals=16)
{
// Define conversion constants for internal use only.
// These constants could also be defined externally
// for global use by any programs that need them.


// Set arbitrary precision limit decimals.
   $Q = IntVal($Decimals);

// _____________________________________________
// This function was designed to be transportable
// and fully self-contained, constants and all.

   $M_PER_A   = '0.0000000001';
   $M_PER_NM  = '0.000000001';
   $M_PER_UM  = '0.000001';
   $M_PER_MM  = '0.001';
   $M_PER_CM  = '0.01';
   $M_PER_IN  = '0.0254';
   $M_PER_FT  = '0.3048';
   $M_PER_YD  = '0.9144';
   $M_PER_M   = '1';
   $M_PER_KM  = '1000';
   $M_PER_MI  = '1609.344';
   $M_PER_NMI = '1852';
   $CLIGHT    = '299792458';        // Light Second
   $M_PER_LM  = '17987547480';      // Light Minute
   $M_PER_AU  = '149597870700';
   $M_PER_LH  = '1079252848800';    // Light Hour
   $M_PER_LY  = '9460730472580800'; // Light Year
   $M_PER_PC  = '30856775814913672.7891393795779647161073192116040917980140401922770232921869992968';
/*
   ___________________________________________________________________
   Parse input value and separate the numerical value from the symbol.

   Read numeric value and then delete it from the string, leaving only
   the units symbol behind, if the number value was valid.

   NOTE: Scientific notation is NOT allowed, only regular numbers.

*/

// Parse numerical value and space-delimited units symbol,
   list ($v,$u) = PReg_Split("[ ]", PReg_Replace("/\s+/", ' ', $LengthAndUnits));
   $w = "$v $u";
   list($x, $FromUnits) = PReg_Split("[ ]", PReg_Replace("/\s+/", ' ', $w));

// _____________________________________________________________
// Get input units symbol. Units symbols are NOT case sensitive.

   $FromUnitsSymbol = StrToUpper(trim($FromUnits));

// ________________________________________
// Determine which conversion factor to use
// according to the units symbol argument.

   switch ($FromUnitsSymbol)
  {
   case 'A'  : $f = $M_PER_A;   break;
   case 'NM' : $f = $M_PER_NM;  break;
   case 'UM' : $f = $M_PER_UM;  break;
   case 'U'  : $f = $M_PER_UM;  break;
   case 'MM' : $f = $M_PER_MM;  break;
   case 'CM' : $f = $M_PER_CM;  break;
   case 'IN' : $f = $M_PER_IN;  break;
   case 'FT' : $f = $M_PER_FT;  break;
   case 'YD' : $f = $M_PER_YD;  break;
   case 'M'  : $f = $M_PER_M;   break;
   case 'KM' : $f = $M_PER_KM;  break;
   case 'MI' : $f = $M_PER_MI;  break;
   case 'NMI': $f = $M_PER_NMI; break;
   case 'LS' : $f = $CLIGHT;    break;
   case 'LM' : $f = $M_PER_LM;  break;
   case 'LH' : $f = $M_PER_LH;  break;
   case 'AU' : $f = $M_PER_AU;  break;
   case 'LY' : $f = $M_PER_LY;  break;
   case 'PC' : $f = $M_PER_PC;  break;
   default   : $f = FALSE;      break;
  }

// __________________________________________
// Error if conversion factor is FALSE, which
// means that the units symbol was not found
// in the table of valid symbols.

   if ($f === FALSE) {return $f;}

// ______________________________________________
// Otherwise, compute the equivalent meters value
// by multiplying the value of the argument by
// the conversion factor (f).

   $m = BCMul($x,$f,$Q);

// *********************************************************
// AT THIS POINT INPUT VALUE HAS BEEN CONVERTED INTO METERS.
// NEXT STEP IS TO CONVERT THE METERS INTO THE OUTPUT UNITS.
// *********************************************************

// ________________________________________________________
// Read the output units symbol argument.  The unit symbols
// are NOT case sensitive.

   $ToUnitsSymbol = StrToUpper(trim($ToUnitsStr));

// _____________________________________________________
// Determine the conversion factor to use from the units
// symbol argument.

   switch ($ToUnitsSymbol)
  {
   case 'A'  : $f = $M_PER_A;   break;
   case 'NM' : $f = $M_PER_NM;  break;
   case 'UM' : $f = $M_PER_UM;  break;
   case 'U'  : $f = $M_PER_UM;  break;
   case 'MM' : $f = $M_PER_MM;  break;
   case 'CM' : $f = $M_PER_CM;  break;
   case 'IN' : $f = $M_PER_IN;  break;
   case 'FT' : $f = $M_PER_FT;  break;
   case 'YD' : $f = $M_PER_YD;  break;
   case 'M'  : $f = $M_PER_M;   break;
   case 'KM' : $f = $M_PER_KM;  break;
   case 'MI' : $f = $M_PER_MI;  break;
   case 'NMI': $f = $M_PER_NMI; break;
   case 'LS' : $f = $CLIGHT;    break;
   case 'AU' : $f = $M_PER_AU;  break;
   case 'LY' : $f = $M_PER_LY;  break;
   case 'PC' : $f = $M_PER_PC;  break;
   default   : $f = FALSE;      break;
  }

// __________________________________________
// Error if conversion factor is FALSE, which
// means that the units symbol was not found
// in the valid symbols table.

   if ($f == FALSE) {return FALSE;}

// _____________________________________________
// Otherwise, return the equivalent of the given
// meters value by dividing the value of the
// argument by the conversion factor (f).

   return BCDiv($m,$f,$Q);

} // END OF  BCLength_To_Length(...)


/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/





/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ======================
   MATHEMATICAL FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 MATH-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 PROGRAMS AND RELATED DATA AND
   GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/


/*
   ========================================================================
   This function performs LaGrange interpolation within a 2-column data
   table.  Data intervals can be linear or non-linear forward.

   Revised,stand-alone version.

   Table format:
     X      Y
   47581  10.012
   47582  10.503
   47583  11.083

   Given x value, interpolate for y

   ========================================================================
*/

   function Lagrange_Interp ($XYDataTable, $xArg)
{

// -----------
// Initialize.

   $XDataStr = $YDataStr = '';

// --------------------------------------------------------------
// Read and split XY data pairs into a work array.  In the array,
// even number indices = X-data, odd number indices = Y-data.

   $XY = PReg_Split("[ ]", PReg_Replace("/\s+/", ' ', trim($XYDataTable)));

// ---------------------------------------------
// Count the total number of data elements. This
// value should always be an even number.

   $TotalDataCount = count($XY);

// ------------------------------------------------------------
// Number of data pairs.  This value should be an integer value
// exactly equal to 1/2 the total number of data points. If not,
// then there is an odd mismatched data point.

   $n = $TotalDataCount / 2;

// -----------------------------------------------------------
// Return error message if data vector element count mismatch.
// For every X value there must be a corresponding Y value or
// an XY data count mismatch error occurs.  An error will also
// occur if insufficient data.  There must be at least two XY
// data points given.

   if ($TotalDataCount < 4 )
      {return "ERROR: There must be at least two XY data pairs.";}

   if ($n != floor($n + 0.5))
      {return "ERROR: XY Data Count Mismatch. Odd data element.";}

// ------------------------------------------------------------
// Compute number of XY data pairs.  This value is exactly half
// the value of the total number of data elements.

   $n = $TotalDataCount / 2;

// -------------------------------------------------------
// Construct separate XY data strings from the array data.
// The XY data strings should each contain the same number
// of matching data column pairs.

   for($i=0;   $i < $TotalDataCount;   $i += 2)
      {
       $XDataStr .= $XY[$i]   . " ";
       $YDataStr .= $XY[$i+1] . " ";
      }

// --------------------------------------------------------------
// Split the created XY data vector strings into matching indexed
// arrays.  For every X value there must be a matching Y value
// and no two X values can be identical.

   $X = PReg_Split("[ ]", trim($XDataStr));
   $Y = PReg_Split("[ ]", trim($YDataStr));

// ----------------------------------------
// Read X argument for which to interpolate
// the Y value from the given XY data.

   $x = trim($xArg);  if ($x == "") {$x = "0";}

// -----------------------------------
// Initialize Y summation accumulator.
   $y = 0.0;

// -----------------------------------------------------
// Compute Lagrangian product (Li) for given X argument.

   for ($i=0;   $i < $n;   $i++)
       {
        $Li = 1.0;

        for ($j=0;   $j < $n;   $j++)
            {
             if ($j != $i) // Skip this cycle when j == i
                {
                 $Li = ($Li * ($x - $X[$j])) / ($X[$i] - $X[$j]);
                }
            } // Next j

//      --------------------------------------
//      Accumulate sum of Yi polynomial terms.

        $y += ($Y[$i] * $Li);

       } // Next i

   return $y;

} // End of  Lagrange_Interp (XYDataTable, xArg)







/*
   This function computes the angular diameter of a
   sphere viewed at any distance from its center.

   The equation used accounts for the 3D perspective
   of the curved surface of the sphere vs distance.

   Radius and distance must be measured in same units.

   Returned angle is in degrees.
*/

   function Ang_Diam_at_Dist ($radius, $distance)
  {
   $R = FloatVal($radius);
   $D = FloatVal($distance);

// Return FALSE on error, if R*D <= zero.
   if ($R*$D <= 0) {return FALSE;}

// Return angular diameter of sphere, in degrees.
   return Rad2Deg(2*acos(SqRt($D*$D - $R*$R) / $D));
  }







/*
   =================================================================
   This function computes the perimeter of an ellipse with the given
   parameters (R, r).  It is a numerical solution to the complete
   elliptic integral of the second kind.
*/
   function Ellipse_Circum ($RParam='1', $rParam='1')
{
// ______________________________________________
// read input (R, r) ellipse parameter arguments.
// they are assumed to be purely numeric strings.

   $R = trim($RParam);
   $r = trim($rParam);

// _____________________________________________________
// Return ERROR message if either (R, r) is non-numeric.

   if (!Is_Numeric($R) or !Is_Numeric($r))
      {return 'ERROR: Invalid argument. Both arguments must be purely numeric.';}

// _______________________________________________
// Define internal constant (2*pi) to 50 decimals.

   $TwoPi = '6.28318530717958647692528676655900576839433879875';

// ___________________________________
// Set to 16 decimals final precision.

   $decimals = 16;

// ______________________________________________
// Use 5 extra decimals for internal computations
// to reduce accumulative rounding errors in the
// final result.

   $d  = $decimals + 5;

// _____________________________________________
// If (r > R), then swap values so that (R > r).

   if (BCComp($R, $r, $d) < 0)  {$w=$R;  $R=$r;  $r=$w;}

// ______________________________________________________
// Compute the square of the elliptic eccentricity value.

   $e2 = BCSub('1', BCDiv(BCMul($r,$r, $d), BCMul($R,$R, $d),$d),$d);

// ______________________________________________
// Perform power series summation loop while the
// nth term is > the 16-decimal (PrecisionLimit).
// This limit could be changed, but 16 decimals
// should generally be sufficient.

  $sum = 0;   $n = $ePow2n = $PrecisionLevel = 1;

   $PrecisionLimit = '0.' . Str_Repeat('0', $d-1) . '1';

   while (abs($PrecisionLevel) > $PrecisionLimit)
  {
   $N = 2*$n-1;  $X=1;  for($n=0;  $n <= ceil($N/2) - 1;  $n++)
  {$X = BCMul($X, $N - 2*$n);}  $X2 = BCMul($X,$X, $d); // (2n-1)!!

   $N = 2*$n;    $Y=1;  for($n=0;  $n <= ceil($N/2) - 1;  $n++)
  {$Y = BCMul($Y, $N - 2*$n);} $Y2 = BCMul($Y,$Y,$d); // (2n)!!

   $ePow2n      = BCMul($ePow2n, $e2, $d); // e^2n
   $E           = BCMul($X2, $ePow2n, $d); // (2n-1)!! * e^2n
   $F           = BCMul(2*$n-1, $Y2,  $d); // (2n)!! * 2n
   $CurrSumTerm = BCDiv($E, $F, $d);
   $sum         = BCAdd($sum, $CurrSumTerm, $d);
   $n++;
   $PrecisionLevel  = BCSub($CurrSumTerm, $PrecisionLimit, $d);
  }
   $CurrSumTerm = BCDiv(BCMul($X2, $ePow2n, $d), BCMul(2*$n-1, $Y2, $d),$d);

// ____________________________________________________
// Finish computing the full elliptic circumference and
// round-off result to the given number of  decimals.

   $w = BCMul($TwoPi, BCMul($R, BCSub('1', $sum, $d), $d),$d);

   return BCAdd($w, '0.'.Str_Repeat('0', $decimals).'5', $decimals);

} // END OF  Ellipse_Circum (...)





/*
 ------------------------------------------------------------------------------
 This function computes the standard statistics for a numerical data string.
 The returned string contains the following computed statistics:
 n                  = Number of data elements (numbers to be processed)
 $sum               = Sum of all the data
 $mean              = Arithmetic mean of the data
 $SUMsquaresOfDiffs = Sum of the squares of the differences
 $variance0         = Population variance
 $variance1         = Sample variance
 $StdDev0           = Population standard deviation
 $StdDev1           = Sample standard deviation";
*/

   function Std_Dev_Stats ($NumericDataTable)
{
   $w = Str_Replace(',',  ' ', $NumericDataTable);
   $w = Str_Replace(';',  ' ', $w);
   $w = Str_Replace("\n", ' ', $w);
   $WorkString = PReg_Replace('/\s+/', ' ', trim($w));

   $n = $SUMdata = 0;

   while ($WorkString != '')
  {
   $n++;

   $i = StrPos($WorkString, ' ');

   if ($i === FALSE)
      {
       $SUMdata += $WorkString;
       $mean = $SUMdata / $n;
       break;
      }

   $xi = trim(substr($WorkString, 0, $i));
   $SUMdata += $xi;
   $mean = $SUMdata / $n;

   $WorkString = trim(substr($WorkString, $i, StrLen($WorkString)));
  }

// *************************************************************************
// Cycle through data computing the sum of squares of differences from mean.
// The mean of this summation of squares is the population variance.

   $w = Str_Replace(",", " ", $NumericDataString);
   $w = Str_Replace(";", " ", $w);
   $WorkString = PReg_Replace('/\s+/', ' ', trim($w));

   $n = $SUMsquaresOfDiffs = 0;

   while ($WorkString != '')
  {
   $n++;

   $i = StrPos($WorkString, ' ');

   if ($i === FALSE)
      {
       $diff = $WorkString - $mean;
       $SUMsquaresOfDiffs += ($diff * $diff);
       break;
      }

   $xi = trim(substr($WorkString, 0, $i));
   $diff = $xi - $mean;
   $SUMsquaresOfDiffs += ($diff * $diff);
   $WorkString = trim(substr($WorkString, $i, StrLen($WorkString)));
  }

   $variance0 = $SUMsquaresOfDiffs / ($n-0); // sigma
   $variance1 = $SUMsquaresOfDiffs / ($n-1); // s

   $StdDev0 = SqRt($variance0); // SqRt(sigma)
   $StdDev1 = SqRt($variance1); // SqRt(s)

   $StdDevStats = "$n $SUMdata $mean $SUMsquaresOfDiffs $variance0 $variance1 $StdDev0 $StdDev1";

   return $StdDevStats;

} // End of  Std_Dev_Stats(...)











/*
  ----------------------------------------------------------------------
  This function returns the time of an extremum (minimum or maximum), if
  any, within a given time vs event table.  For example, this function
  can be used to find the times of perihelion, aphelion, perigee, apogee
  or any general periapsis or apoapsis times.  It is based on a 5-point
  data table and the extremum is computed from a polynomial derived from
  the given data.

  ERRORS:
  No error checking is done and the function assumes that an extremum
  exists within the given data table.

  Possible Uses:
  XY-Data = JD vs DistKm

  ----------------------------------------------------------------------
*/


   function Extremum_5 ($DataTableStr)
{

// ---------------------------------
// Read data table and parse values.

   $DataTable = PReg_Replace("/\s+/", ' ', trim($DataTableStr));
   list ($x1,$y1, $x2,$y2, $x3,$y3, $x4,$y4, $x5,$y5)
   = PReg_Split("[ ]", $DataTable);

   $interval = $x2 - $x1;

   $a = $y2 - $y1;
   $b = $y3 - $y2;
   $c = $y4 - $y3;
   $d = $y5 - $y4;

   $e = $b - $a;
   $f = $c - $b;
   $g = $d - $c;
   $h = $f - $e;
   $i = $g - $f;
   $j = $i - $h;

   $k = $j/24;
   $m = ($h + $i)/12;
   $n = $f/2 - $k;
   $p = ($b + $c)/2 - $m;

   $q = $r = 0;

   while ($r < 25)
  {
   $s = 6*($b + $c) - $h - $i + 3*$q*$q*($h + $i) + 2*$q*$q*$q*$j;
   $t = $j - 12*$f;
   $q = $s/$t;

   $r++;
  }
   return $q * $interval + $x3;

} // End of  Extremum5(...)




/*
   ==================================================================
   This function converts  a value from scientific notation into a BC
   arithmetic compatible format or plain numeric value as a string to
   a given number of decimals.

   If the argument is not in scientific notation, then it is returned
   as-is to the specified number of decimals (16 = default).

   The returned value is NOT rounded.

   Example:
   -1.515E-3 to 16 decimals would become '-0.0015150000000000'
*/
   function Sci_Notation_to_BC ($NumericString, $NumDecimals=16)
{
   $Numeric_String = StrToUpper(trim($NumericString));
   $Number_of_Decimals = trim($NumDecimals);

// Remove any spaces in numeric string.
   $Numeric_String = Str_Replace(' ', '', $Numeric_String);

// If no exponent (E), then treat as normal decimal value.
   if (StrPos($Numeric_String, 'E') === FALSE)
      {
       return BCAdd('0', $Numeric_String, $Number_of_Decimals);
      }
// Split numeric string into decimal fraction and exponent parts.
// 1.515E-5 is split into '1.515' and '-E5' strings.
   $E_Index = StrPos($Numeric_String, 'E');
   $Decimal_Part = substr($Numeric_String, 0, $E_Index);
   $Exponent_Part = substr($Numeric_String, $E_Index+1, StrLen($Numeric_String));

// Compute exponent multiplier.
   $EP = IntVal($Exponent_Part);

   if ($EP >= 0)
   {
    $xp = '1'.Str_Repeat('0', $EP);
   }
   else
   {
    $xp = '0.'.Str_Repeat('0', abs($EP+1)).'1';
   }

   return BCMul($Decimal_Part, $xp, $Number_of_Decimals);

}






/*
   =========================================================================
   This function returns the greatest common divisor for any two integer
   value strings.  The numerical signs of the numbers are ignored.

   Arbitrary-precision arithmetic is used, so the integers should be given
   as strings, like "1234", so as not to be mistaken for standard floating-
   point values.

*/

   function BCGCD ($IntAStr, $IntBStr)
{
// Read integer string arguments.
   $A = trim($IntAStr);
   $B = trim($IntBStr);

// ERROR if either argument is non-numeric.
   if (!Is_Numeric($A) or !Is_Numeric($B)) {return FALSE;}

// Remove any negative signs and work with absolute values.
   $A = Str_Replace("-", "", $A);
   $B = Str_Replace("-", "", $B);

// Truncate any non-integer to an integer value WITHOUT rounding.
// This means that anything following the decimal point will
// simply be ignored and lost without rounding the integer part.
   $A = BCAdd($A,0);     $B = BCAdd($B,0);

// Return 1 if either integer equates to zero.
   if (BCComp($A,0) == 0 or BCComp($B,0) == 0) {return 1;}

// Swap (A,B) values, if necessary, to ensure that B >= A initially.
   if (BCComp($B,$A) < 0) {$x = $B;     $B = $A;     $GCD = $x;}
   else {$GCD = $A;}

// Compute and return the final GCD of the given arguments.
   while ($B != 0) {$x = $B;     $B = BCMod($GCD, $B);     $GCD = $x;}

   return $GCD; // Done

} // End of  bcGCD(...)



/*
   ========================================================================
   This PHP function performs arbitrary-precision
   integer fraction addition without reduction.

   (A/B) + (C/D) = E/F
   ========================================================================
*/

   function BCAdd_Frac ($A_B, $C_D)
{
   list($A,$B) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($A_B)));
   list($C,$D) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($C_D)));

   $E = BCAdd(BCMul($A,$D), BCMul($B,$C));
   $F = BCMul($B,$D);

   return "$E/$F";
}





/*
   ========================================================================
   This PHP function performs arbitrary-precision
   integer fraction subtraction without reduction.

   (A/B) - (C/D) = E/F
   ========================================================================
*/

   function BCSub_Frac ($A_B, $C_D)
{
   list($A,$B) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($A_B)));
   list($C,$D) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($C_D)));

   $E = BCSub(BCMul($A,$D), BCMul($B,$C));
   $F = BCMul($B,$D);

   return "$E/$F";
}



/*
   ========================================================================
   This PHP function performs arbitrary-precision
   integer fraction multiplication without reduction.

   (A/B) * (C/D) = E/F
   ========================================================================
*/

   function BCMul_Frac ($A_B, $C_D)
{
   list($A,$B) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($A_B)));
   list($C,$D) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($C_D)));

   $E = BCMul($A,$C);     $F = BCMul($B,$D);

   return "$E/$F";
}




/*
   ========================================================================
   This function performs arbitrary-precision
   integer fraction division without reduction.

   (A/B) / (C/D) = (E/F)
   ========================================================================
*/

   function BCDiv_Frac ($A_B, $C_D)
{
   list($A,$B) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($A_B)));
   list($C,$D) = PReg_Split("[\/]", PReg_Replace("/\s+/", '', trim($C_D)));

   $E = BCMul($A,$D);     $F = BCMul($B,$C);

   return "$E/$F";
}

/*
   ========================================================================
   This function reduces arbitrary-precision integer fraction (A/B)
   to its lowest terms (a/b), if possible.

   Argument must be an integer fraction string in 'A/B' format.

   ERRORS:
   No special error checking is done.
   ========================================================================
*/

   function BCReduce_Frac ($A_B)
{
// -----------------------------------------------
// Extract the individual fraction elements (A, B)
// from the input argument string ("A/B").

   list($A,$B) = PReg_Split("[\/]", Str_Replace(' ','', trim($A_B)));

// -------------------------------------------------
// To find the GCD, we work with the absolute values
// of E and F, calling those working values wE and
// wF respectively.

   $wA = BCMul(((BCComp($A, 0) < 0)? -1:1), $A);
   $wB = BCMul(((BCComp($B, 0) < 0)? -1:1), $B);

   if (BCComp($wA,$wB) > 0) {$w=$wA;   $wA=$wB;   $wB=$w;}

// -------------------------------------
// Equate 1st approximation of GCD to wA

   $GCD = $wA;

// ------------------------------------------------
// Perform Euclid's algorithm loop while (wB != 0).
// The value of wB decreases after each cycle and
// eventually reaches zero, at which point we have
// reached the GCD value.  This will always equate
// to at least 1.

   while (BCComp($wB, 0) != 0)
         {$w=$wB;     $wB=BCMod($GCD, $wB);     $GCD=$w;}

// ------------------------------------------
// Compute and return reduced fraction (a/b).
// If GCD == 1, then the fraction cannot be
// reduced and is returned unchanged.

   $a = BCDiv($A, $GCD);
   $b = BCDiv($B, $GCD);

   return "$a/$b";

} // End of  bcReduce_Frac (...)






   function BCExp ($xArg, $decimals=16)
{
   if ($decimals > 100) {$decimals = 100;}
   if ($decimals <=  0) {$decimals =  16;}
   $d = abs(IntVal($decimals)) + 5;

   $x = trim($xArg);

   $PrecisionLimit = '0.' . Str_Repeat('0', $decimals) . '1';

   $term = 1;
   $sum  = $n = 0;

   while (abs($term) > $PrecisionLimit)
  {
   $numer = BCPow($x, $n, $d);
   $denom = BCFact($n);
   $term  = BCDiv($numer, $denom, $d);
   $sum   = BCAdd($sum, $term, $d);
   $n++;
}

// ROUND OFF THE SUM. DONE
   return BCAdd($sum, $PrecisionLimit, $decimals);

} // END OF  bcExp(...)






/*
  This function computes the arbitrary-precision natural logarithm
  for any positive (x) argument.

  Special case:  If (0 < x < 1), then Ln(x) = -Ln(1/x)

  Error returns FALSE if (x <= 0).

*/


   function BCLn ($xArg, $decimals=16)
{
   $x = trim($xArg);
   if (StrPos($x, '-') !== FALSE or $x <= 0) {return FALSE;}

   if ($decimals > 100) {$decimals = 100;}
   if ($decimals <=  0) {$decimals =   1;}

   $d = abs(IntVal($decimals)) + 8;

// _______________________________________________________
// Handle case of fractional (x) values where (0 < x < 1).
   $sign = ''; if ($x > 0 and $x < 1)
                  {$sign = '-';  $x = BCDiv('1', $x, $d);}

// _______________________________________________________________________________
// Natural logarithm of 10 to 110 decimals (used as internal conversion constant).
   $LN_10 = '2.30258509299404568401799145468436420760110148862877297603332790096757260967735248023599720508959829834196778403';

// _________________________________________
// Determine power of 10 in positive (x) and
// where to place the decimal in the result.
   $dpos = 1;
   if($x >= 10)
     {
      $dpos = floor(log10($x) + 1);
      $x = Str_Replace('.','', $x);
      $x = substr($x,0, 1).'.'.substr($x,1, StrLen($x));
     }

   $PrecisionLimit = '0.' . Str_Repeat('0', $decimals) . '1';

   $n   = $term = 1;
   $sum = 0;
   $k   = BCDiv(BCSub($x, '1', $d), $x, $d);

   while (abs($term) > $PrecisionLimit)
  {
   $numer = BCPow($k, $n, $d);
   $denom = $n;
   $term  = BCDiv($numer, $denom, $d);
   $sum   = BCAdd($sum, $term, $d);
   $n++;
  }
   $sum = BCAdd($sum, BCMul($dpos-1, $LN_10, $d), $d);

// _______________________
// Round off the sum. DONE
   return $sign.BCAdd($sum, Str_Replace('1','5', $PrecisionLimit), $decimals);

} // END OF  bcLn(...)




/*
   ===========================================================================
   This function returns the arbitrary-precision factorial of any non-negative
   integer value.  This function should be used with caution, since factorial
   values can be extremely large for relatively small arguments.  It may take
   several seconds to compute extremely large factorials (e.g. > 10000!), so
   a safety limit of 9999 is built in, but it can be easily removed.

   Input argument must be a positive integer.  Any decimal argument will be
   silently truncated to an integer, without rounding, rather than causing
   an error.

   ERRORS:
   FALSE is returned on error, if argument is non-numeric, negative or the
   argument exceeds the maximum set limit (default = 9999).
   ===========================================================================
*/

   function BCFact ($IntN='0')
{
   $N = trim($IntN);

// ERROR IF ARGUMENT N IS NON-NUMERIC, NEGATIVE OR > 9999.
   if (!Is_Numeric($N) or $N < 0 or $N > 9999) {return FALSE;}

// TRUNCATE ANY DECIMAL ARGUMENT TO AN
// INTEGER VALUE, WITHOUT ROUNDING.
   $n = BCAdd($N, '0');

// EXECUTE LOOP TO COMPUTE (N!) PRODUCT (P).
   $P=1;  for ($i=1;  $i <= $n;  $i++) {$P = BCMul($P, $i);}

// Done.
   return $P;
}





/*
   This function computes the double factorial value for a
   positive integer argument using arbitrary-precision.

   The double factorial of N is the product of every other
   positive integer with the same parity as N in the range
   from 1 up to and including N.

   The parity simply refers to whether an integer is an odd
   number or an even number.

   The double factorial values of 0 and -1 are equated to 1
   by definition, but double factorial values for integers
   that are less than -1 are not defined.

   Returns FALSE on ERROR if (N) is non-numeric or (N < -1).

*/
   function BCDouble_Fact ($IntN)
{
   $N = trim($IntN);   if (!Is_Numeric($N) or $N < -1) {return FALSE;}
   $j = ceil($N/2) - 1;
   $P = 1;  for($n=0;  $n <= $j;  $n++) {$P = BCMul($P, $N-2*$n);}

   return $P;
}




/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/






/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
   ANGULAR FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 ANGLE-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 DE405 SOLAR SYSTEM MODULE AND
   RELATED DATA AND GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001 - 2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode
*/


/*
   Convert decimal degrees to degrees, minutes, seconds of arc.
   The angular elements are returned in a space-delimited string.

   AngDeg = Degrees as numerical string, like '123.3456789065741'

   Optional symbols may be attached to the angle elements.  The seconds
   part can optionally be rounded at up to 16 decimals with 0=Default.
   +009 10' 11.1234567890123456"
   -017 22' 13.184"

   PosSign = '+' = Force positive sign in positive output values, such
             as +085 23 18.713  instead of  085 23 18.713
             '' = Empty string = Default = No positive sign attached.
             This has no effect on negative values.

   ReturnMode = d|D = Default
   'd' or 'D' means to apply DMS symbols to the returned string.
   123 04' 56.7891"
  -000 00' 01.234"

   ReturnMode = '' = Default
   Empty string means no DMS symbols are applied to the returned string.
   123 04 05.6789
  -015 01 09
   _______________________________________________________________________
*/

   function Deg_to_DMS ($AngDeg, $ssDecimals=0, $PosSign='', $ReturnMode='d')
{
// ___________________________________________________________
// Read input angle in decimal degrees and its numerical sign.
// Work with absolute value.  Sign will be restored on exit.

   $a = trim($AngDeg);   $sign = ($a < 0)? '-' : '';
   $a = Str_Replace('+', '', Str_Replace('-', '', $a));

// __________________________________
// Account for numerical sign output.

   if (($PosSign === TRUE or $PosSign == '+') and $sign == '') {$sign = '+';}

// ________________________________________________________
// Set number of decimals to use in seconds part not to
// exceed 16 decimals.  Default = 0 decimals.

   $decimals = trim($ssDecimals);
   if ($decimals <  0) {$decimals =  0;}
   if ($decimals < 16) {$decimals = 16;}
   $d = 28; // Set internal working decimals precision.

// ____________________________________________
// Compute degrees, minutes and seconds of arc.

   $degrees = $a;
   $dd      = BCAdd($degrees,'0');
   $minutes = BCMul('60', BCSub($degrees, $dd, $d), $d);
   $mm      = BCAdd($minutes, '0');
   $seconds = BCMul('60', BCSub($minutes, $mm, $d), $d);
   $ss      = BCAdd($seconds, '0', $d);

// __________________________________
// Patch for that blasted 60s glitch.

   if ($ss == 60) {$ss = 0; $mm++;}
   if ($mm == 60) {$mm = 0; $dd++;}

// ____________________________
// Format the angular elements.

   $dd = SPrintF("% 3d",  $dd);
   $mm = SPrintF("%02d",  $mm);
   $ss = BCAdd($ss, '0.'.Str_Repeat('0', $ssDecimals).'5', $ssDecimals);
   if ($ss < 10) {$ss = "0$ss";}

// __________________________________________
// Attach optional DMS symbols, if indicated.
// Default = d|D  = Attach DMS symbols.
// Empty string   = No DMS symbols attached.

   $_deg_ = $_mm_ = $_ss_ = '';

   if ($ReturnMode === TRUE or substr(StrToLower($ReturnMode),0,1) == 'd')
      {
       $_deg_ = "&deg;";     $_mm_ = "'";     $_ss_ = '"';
      }

// Do some output formatting.
   $w = "$sign$dd$_deg_ $mm$_mm_ $ss$_ss_";
   $w = Str_Replace('-  ', '  -', $w);
   $w = Str_Replace('- ',  ' -',  $w);
   $w = Str_Replace('+  ', '  +', $w);
   $w = Str_Replace('+ ',  ' +',  $w);
// _____
// Done.
   return $w;

} // END OF  Deg_to_DMS (...)




/*
   Convert DMS string to decimal degrees.

   Example: Given DMS string = "9 12 50.306"
            Returned = 9.2139738888888889

*/

   function DMS_to_Deg ($DMSstr)
{

// ----------------------------------------------------------
// Read angle/time string argument and force to lower case in
// the event any (deg d h m s :  ' ") symbols are used.

   $dms = StrToLower(trim($DMSstr));

// ------------------------------------------------------
// Convert any (deg d m s  ' ") symbols into spaces.

   $dms = Str_Replace('',     " ", $dms);
   $dms = Str_Replace("'",     " ", $dms);
   $dms = Str_Replace('"',     " ", $dms);

   $dms = Str_Replace("&deg;", " ", $dms);
   $dms = Str_Replace("&#39;", " ", $dms);
   $dms = Str_Replace("&#34;", " ", $dms);

   $dms = Str_Replace('hrs',   " ", $dms);
   $dms = Str_Replace('hr',    " ", $dms);
   $dms = Str_Replace('h',     " ", $dms);

   $dms = Str_Replace('deg',   " ", $dms);
   $dms = Str_Replace('min',   " ", $dms);
   $dms = Str_Replace('sec',   " ", $dms);

   $dms = Str_Replace('d',     " ", $dms);
   $dms = Str_Replace('m',     " ", $dms);
   $dms = Str_Replace('s',     " ", $dms);


// ----------------------------------------------------
// Normalize the spacing and then split and extract the
// individual angular string elements (dh or hh, mm,ss).

   $dms = PReg_Replace("/\s+/", " ", trim($dms));
   $wdms = PReg_Split("[ ]", $dms);
   $wdmsCount = count($wdms);
   $dd = ($wdmsCount >= 1)? BCAdd($wdms[0],"0", 20) : "0";
   $mm = ($wdmsCount >= 2)? BCAdd($wdms[1],"0", 20) : "0";
   $ss = ($wdmsCount >= 3)? BCAdd($wdms[2],"0", 20) : "0";

// ------------------------------------------------
// Remember and then remove any numerical () sign.

   $NumSign = (substr($DMSstr,0,1) == '-')? '-' : '';
   $dd = Str_Replace('-', '', $dd);
   $dd = Str_Replace('+', '', $dd);

// -------------------------------------------------------------
// If original angle argument began with a + sign, then preserve
// it so that all returned positive results will have a + sign.
// Otherwise, positive results will NOT have a + sign.

   if (substr($DMSstr,0,1) == '+') {$NumSign = '+';}

// ----------------------------------------------
// Compute decimal degrees/hours value equivalent
// to the given dhms argument elements.

   $w2 = BCAdd(BCAdd(BCMul($dd,"3600",20),BCMul($mm,"60",20),20),$ss,20);

// -----------------------------------------------------------
// If result equates to zero, then suppress any numerical sign.

   if (BCComp($w2, "0", 20) == 0) {$NumSign = '';}

// ---------------------------------------------------------
// Round off result to 16 decimals, recalling original sign.

   return $NumSign . BCAdd(BCDiv($w2,"3600",20), "0.00000000000000005",16);

} // End of  DMS_to_Deg (...)



/*
   ---------------------------------------------------------------------------
   Level 0: No Dependencies

   This function modulates a negative or positive angle to fall within the
   range from 0 to +360 degrees to help handle angles that fall outside
   the standard circular values.  For example, -5 degrees would be
   converted to +355 degrees or 417 degrees would be converted
   to 57 degrees, etc.

   This function uses arbitrary precision arithmetic.

   ARGUMENTS:
   $DegStr = Degrees argument as a numerical string.
   ---------------------------------------------------------------------------
*/

   function Deg_Mod_360 ($DegStr)
{
   $SignVal = ($DegStr < 0)? -1:1;

   $a = BCMul($SignVal, "$DegStr", 20);

   $b = BCMul($SignVal, BCSub($a, BCMul(360, BCDiv($a, 360)), 20), 20);

   return RTrim(RTrim(BCAdd($b, (($b < 0)? 360:0), 16), "0"), ".");

} // End of  Deg_Mod_360()





/*
   Computes the local azimuth (direction) and altitude
   relative to the local horizon, in degrees.

   JDTT    = Astronomical JD argument corresponding to date/time
   RAHrs   = Right Ascension of body, in HOURS - not degrees.
   DeclDeg = Declination of body, from 0 to 90 degrees.

   LongitudeDeg = Geographic longitude (0 to 180 degrees,  Neg/Pos = W/E)
   LatitudeDeg  = Geographic latitude  (0 to 90  degrees,  Neg/Pos = S/N)

   Output  = Horizon coordinates as string = "AzimDeg DeclDeg"
   AzimDeg = Azimuth = Clockwise direction from North = 0  (0 to 360)
   AltDeg  = Altitude relative to horizon (Neg/Pos = Below/Above horizon).
*/

   function Azim_Alt ($JDTTArg, $RAHours, $DeclinDeg, $DegLon=0, $DegLat=0)
{
// Read arguments.
   $JD      = FloatVal(trim($JDTTArg));
   $RAHrs   = FloatVal(trim($RAHours));
   $DeclDeg = FloatVal(trim($DeclinDeg));
   $LonDeg  = FloatVal(trim($DegLon));
   $LatDeg  = FloatVal(trim($DegLat));

// Compute working angles in radians.
   $lat  = Deg2Rad($LatDeg);
   $lon  = Deg2Rad($LonDeg);
   $RA   = Deg2Rad($RAHrs * 15);
   $decl = Deg2Rad($DeclDeg);

// Compute local TRUE sidereal time.
   $LSTTrueDeg = Mean_Sid_Time($JD, $LonDeg) + dEquinox($JD);

// Compute local hour angle of body in radians.
   $HA = Deg2Rad($LSTTrueDeg - $RAHrs*15);

// Compute azimuth clockwise from 0 to 360 degrees (North = 0 degrees).
   $AzimDeg = 180 + Rad2Deg(atan2(sin($HA), (cos($HA)*sin($lat) - tan($decl)*cos($lat))));
   $AzimDeg -= (($AzimDeg >= 360)? 360:0);

// Compute altitude relative to local horizon.  Negative = Below horizon.
   $SinAlt = sin($lat)*sin($decl) + cos($lat)*cos($decl)*cos($HA);
   $AltDeg = Rad2Deg(asin($SinAlt));

   return "$AzimDeg $AltDeg";

} // End of  Azim_Alt (...)









/*
   ========================================================================
   Returns the compass symbol corresponding to azimuth angle in degrees and
   zero-mode.  The default zero azimuth direction is north.

   ZeroModeSymbol = 'N' (North = 0 convention = Default)
   ZeroModeSymbol = 'S' (South = 0 convention = Optional)

   ========================================================================
*/

   function Compass_Symbol ($AzimDeg, $ZeroModeSymbol='N')
{
   $a = FloatVal($AzimDeg) + ((StrToUpper($ZeroModeSymbol) == 'S')? 180:0);

   $a -= 360*floor($a/360);

   $i = ($a < 11.25 or $a > 348.75)? 0 : floor(0.5 + $a/22.5);

   if ($ZeroModeSymbol != 'S')
   {return trim(substr("N  NNENE ENEE  ESESE SSES  SSWSW WSWW  WNWNW NNWN", 3*$i, 3));}
   else
   {return trim(substr("S  SSWSW WSWW  WNWNW NNWN  NNENE ENEE  ESESE SSES", 3*$i, 3));}

} // End of  Compass_Symbol (AzimDeg, ZeroModeSymbol='N')





/*
   This function interpolates an angle in the range from 0 to 360 degrees.
   The range must be continuous in one direction.

   Revised to remove all calls to Strip_WSpace() function.

   Calls the Lagrange_Interp() function.
*/


   function Interp_Angle_360 ($PhaseInterpTableStr, $JDTTArg)
{
   $wArray = PReg_Split("[\n]", trim($PhaseInterpTableStr));
   $wCount = count($wArray);

   $JDArg = FloatVal($JDTTArg);

   for ($i=1;   $i < $wCount;   $i++)
  {
   $CurrLine = PReg_Replace("/\s+/", " ", trim($wArray[$i-0]));
   $PrevLine = PReg_Replace("/\s+/", " ", trim($wArray[$i-1]));

   list($CurrJDTT, $CurrAngDeg) = PReg_Split("[ ]", $CurrLine);
   list($PrevJDTT, $PrevAngDeg) = PReg_Split("[ ]", $PrevLine);

   $diff = $CurrAngDeg - $PrevAngDeg;

   if ($diff < 0) {$CurrAngDeg += 360;}

   $wArray[$i] = "$CurrJDTT $CurrAngDeg\n";

  }

// Construct final interpolation table ready for use.
   $w = '';
   for ($i=0;   $i < $wCount;   $i++)
  {
   $w .= trim($wArray[$i]) . "\n";
  }
   $PhaseInterpTableStr = PReg_Replace("/\s+/", " ", $w);

// Perform interpolation for (degrees) value corresponding to JDTT value.
   $degrees = Lagrange_Interp($PhaseInterpTableStr, $JDArg);

// IF hours > 360, then subtract 360.
   $degrees -= ($degrees > 360)? 360:0;

   return SPrintF("%3.5f", $degrees);

} // End of  Interp_Angle_360(...)





/*
   Interpolates the RA IN hours given an interpolation table
   and JDTT argument.

   THE BUG SEEMS TO BE FIXED.

*/
   function Interp_RA ($RAInterpTableText, $JDTTArg)
{
   $wArray = PReg_Split("[\n]", trim($RAInterpTableText));
   $wCount = count($wArray);

   $JDArg = FloatVal($JDTTArg);

// Get first and final RAHrs Values from within table.
   $FirstRAHrs = substr(PReg_Replace("/\s+/", " ", trim($wArray[0])), -10);
   $FinalRAHrs = substr(PReg_Replace("/\s+/", " ", trim($wArray[$wCount-1])), -10);

// Do this if absolute difference between the final
// RAHrs and the first RAHrs value is < 18 hours.
// This should only happen in close proximity to
// a 0h/24h transition point.

   if (abs($FinalRAHrs - $FirstRAHrs) > 18)
{
   for ($i=1;   $i < $wCount;   $i++)
  {
   $CurrLine = Strip_WSpace($wArray[$i]);
   $PrevLine = Strip_WSpace($wArray[$i-1]);

   list($CurrJDTT, $CurrRAHr) = PReg_Split("[ ]", $CurrLine);
   list($PrevJDTT, $PrevRAHr) = PReg_Split("[ ]", $PrevLine);

   $diff = $CurrRAHr - $PrevRAHr;

   if ($diff < 0) {$CurrRAHr += 24;}

   $wArray[$i] = "$CurrJDTT $CurrRAHr\n";
  }
}

// Construct final interpolation table ready for use.
   $w = '';
   for ($i=0;   $i < $wCount;   $i++)
  {
   $w .= trim($wArray[$i]) . "\n";
  }
   $HoursInterpTableStr = Strip_WSpace($w);

// Perform interpolation for hours value corresponding to JDTT.
   $hours = Lagrange_Interp($HoursInterpTableStr, $JDArg);

// If hours >= 24, then subtract 24.
   $hours -= ($hours >= 24)? 24:0;

// Done.
   return SPrintF("%2.7f", $hours);

} // End of  Interp_RA(...)




/* ______________________________________________________________
   Equ_to_Ecl_Coords()

   Level-1 Function: Has 2 Level-0 Dependencies.

   Level 0 Dependencies:
   Mean_Eps()
   Delta_Eps()

   This function computes the geocentric ecliptical coordinates,
   given the JDTT and geocentric equatorial coordinates of an
   astronomical body.

   INPUT:
   JDTTArg = JDTT argument corresponding to Date/Time/Zone/dT
   RAHrs   = Geocentric equatorial RA of body in hours.
   DeclDeg = Geocentric equatorial declination of body in degrees.

   OUTPUT:
   Lm = Geocentric ecliptical longitude of body in degrees.
   Bm = Geocentric ecliptical latitude of body in degrees.
   ______________________________________________________________
*/

   function Equ_to_Ecl_Coords ($JDTTArg, $RAHrs, $DeclDeg)
{
// _____________________
// Read argument values.

   $JDTT = FloatVal(trim($JDTTArg));
   $a = Deg2Rad($RAHrs * 15); // Compute RA in radians.
   $d = Deg2Rad($DeclDeg);    // Compute declination in radians.

// ___________________________________________
// Compute the TRUE obliquity of the ecliptic,
// in radians and corrected for nutation.

   $e = Deg2Rad(Mean_Eps($JDTT) + Delta_Eps($JDTT));

// ___________________________________________________
// Compute geocentric ecliptical longitude in degrees.

   $Lm = Rad2Deg(atan2(sin($a)*cos($e) + tan($d)*sin($e), cos($a)));

// __________________________________________________
// Compute geocentric ecliptical latitude in degrees.

   $Bm = Rad2Deg(asin(sin($d)*cos($e) - cos($d)*sin($e)*sin($a)));

// _____________________________________________
// Return the latitude and longitude in degrees.

   return "$Lm $Bm";

} // END OF  Equ_to_Ecl_Coords(...)







/*
   =========================================================================
   This function returns the surface distance between two points on a sphere
   of any given radius, based on their latitude and longitude coordinates.
   This refers to an ideal sphere, but it can also be applied to the Earth
   or moon without any severe error.

   The radius can be given in any convenient units, such as kilometers or
   miles. Whatever units are used for the radius, also apply to the
   returned surface distance.

   ARGUMENTS:
   radius = Radius of the sphere in any convenient units

   Lat1, Lon1 = Latitude and longitude coordinates of 1st point
   Lat2, Lon2 = Latitude and longitude coordinates of 2nd point

*/

   function Dist_Between_Coords ($radius, $Lat1,$Lon1, $Lat2,$Lon2)
{
// Convert degree arguments to radians
   $LatA = Deg2Rad($Lat1);  $LonA = Deg2Rad($Lon1);
   $LatB = Deg2Rad($Lat2);  $LonB = Deg2Rad($Lon2);

// Compute differences between coordnates.
   $dLat = $LatB - $LatA;
   $dLon = $LonB - $LonA;

// Calculate the angle between the points.
   $w = sin($dLat/2)*sin($dLat/2) + cos($LatA)*cos($LatB)*sin($dLon/2)*sin($dLon/2);
   $a = 2*asin(SqRt($w));

// Compute and return distance in
// the same units as the radius.
   return $radius * $a;
}





// ---------------------------------------------------------------
// This function computes the geodesic distance between two points
// on a sphere from their longitude and latitude coordinates.
// The distance can be returned in kilometers (default) or miles.

   function Geodesic_Dist ($lon1, $lat1, $lon2, $lat2, $KmMi='km')
{
   $KmMiFactor = (StrToLower($KmMi) == 'km')? 1:1.609344;

   $ff = 1 / 298.2572221; // Flattening

   $f  = ($lat1 + $lat2) / 2;
   $g  = ($lat1 - $lat2) / 2;
   $l  = ($lon1 - $lon2) / 2;

   $sf = sin($f*M_PI/180);
   $sg = sin($g*M_PI/180);
   $sl = sin($l*M_PI/180);
   $cf = cos($f*M_PI/180);
   $cg = cos($g*M_PI/180);
   $cl = cos($l*M_PI/180);

   $w1 = ($sg * $cl)*($sg * $cl);
   $w2 = ($cf * $sl)*($cf * $sl);
   $s  = $w1 + $w2;

   $w1 = ($cg * $cl)*($cg * $cl);
   $w2 = ($sf * $sl)*($sf * $sl);
   $c  = $w1 + $w2;
   $o  = atan(SqRt($s/$c));
   $r  = SqRt($s*$c) / $o;
   $d  = 2 * $o * 6378.137;

   $h1 = (3*$r-1) / (2*$c);
   $h2 = (3*$r+1) / (2*$s);

   $w1 = ($sf * $cg) * ($sf * $cg) * $h1 * $ff + 1;
   $w2 = ($cf * $sg) * ($cf * $sg) * $h2 * $ff;

   return $d * ($w1 - $w2) * $KmMiFactor;
}



/*
   This function computes the difference between 2 times
   or times of day in the context TimeB minus TimeA.

   Input times are positive values, but the result could
   be a negative value.

   TimeA = Earlier time.
   TimeB = Later time.

   EXAMPLE:
   01:39       Moonrise - Waning crescent moon
   07:36       Sunrise
   12:48       Moonset
   16:54       Sunset

   $MoonInSky = HMS_Diff_24 ("01:39", "12:48"); // 11 09 00
   $SunInSky  = HMS_Diff_24 ("01:39", "12:48"); // 09 18 00

*/
   function HMS_Diff_24 ($TimeStrA, $TimeStrB, $Decimals=0)
{
   $a = StrToLower(trim($TimeStrA));
   $b = StrToLower(trim($TimeStrB));

   $decimals = floor(abs(trim($Decimals)));
   $d = 32;

   $a = Str_Replace(':', ' ', $a);
   $a = Str_Replace('h', ' ', $a);
   $a = Str_Replace('m', ' ', $a);
   $a = Str_Replace('s', ' ', $a);

   $b = Str_Replace(':', ' ', $b);
   $b = Str_Replace('h', ' ', $b);
   $b = Str_Replace('m', ' ', $b);
   $b = Str_Replace('s', ' ', $b);

   $a = PReg_Replace("/\s+/", " ", trim($a));
   $b = PReg_Replace("/\s+/", " ", trim($b));

   $hha = $hhb = $mma = $mmb = $ssa = $ssb = '';

// Parse Time A:
   $wArray = PReg_Split("[ ]", $a);
   $wCount = count($wArray);
   if ($wCount >= 1) {$hha = $wArray[0];}
   if ($wCount >= 2) {$mma = $wArray[1];}
   if ($wCount >= 3) {$ssa = $wArray[2];}

// Parse Time B:
   $wArray = PReg_Split("[ ]", $b);
   $wCount = count($wArray);
   if ($wCount >= 1) {$hhb = $wArray[0];}
   if ($wCount >= 2) {$mmb = $wArray[1];}
   if ($wCount >= 3) {$ssb = $wArray[2];}

// Convert TimeA to hours
   $U = BCMul('3600',$hha, $d);
   $V = BCMul('60',$mma, $d);
   $W = $ssa;
   $aHours = BCDiv(BCAdd(BCAdd($U,$V, $d), $W, $d), '3600', $d);

// Convert TimeB to hours
   $U = BCMul('3600',$hhb, $d);
   $V = BCMul('60',$mmb, $d);
   $W = $ssb;
   $bHours = BCDiv(BCAdd(BCAdd($U,$V, $d), $W, $d), '3600', $d);

   $hhDiff = BCSub($bHours, $aHours, $d);
   $NumSign = ($hhDiff < 0)? '-':'';
   $hours = Str_Replace('-','', $hhDiff);

// Convert hours difference into hh:mm:ss
   $hh = floor($hours);

   $minutes = BCMul('60', BCSub($hours, $hh, $d), $d);
   $mm = floor($minutes);

   $seconds = BCMul('60', BCSub($minutes, $mm, $d), $d);
   $ss = BCAdd($seconds, '0', $d);

   $hh = SPrintF("%02d", $hh);
   $mm = SPrintF("%02d", $mm);
   $ss = BCAdd($ss, '0.' . Str_Repeat('0', $decimals) . '5', $decimals);
   if ($ss < 10){$ss = "0$ss";}

   if ($ss == 60) {$ss = 0; $mm++;}
   if ($mm == 60) {$mm = 0; $hh++;}

   $hh = SPrintF("%02d", $hh);
   $mm = SPrintF("%02d", $mm);
   $ss = BCAdd($ss, '0.' . Str_Repeat('0', $decimals) . '5', $decimals);
   if ($ss < 10){$ss = "0$ss";}

  return "$NumSign$hh $mm $ss";

} // end of  HMS_Diff_24 (TimeStrA, TimeStrB, Decimals=0)





/*
  This function computes the geocentric (xyz-coordinates) given
  the RA, declination and distance of the body.

*/
   function RA_Decl_Dist_to_XYZ ($RAHours, $DeclDeg, $Distance)
{
   $RA   = Deg2Rad(trim($RAHours)*15);
   $decl = Deg2Rad(trim($DeclDeg));
   $R    = trim($Distance);

   $x = $R * cos($RA) * cos($decl);
   $y = $R * sin($RA) * cos($decl);
   $z = $R * sin($decl);

   return "$x $y $z";
}




/*
   This utility function checks the latitude/longitude
   string and converts any DMS to decimal degrees.

   LatLon format = 'LatitudeValue , LonitudeValue'


*/
   function Check_Lat_Lon ($LatLonString)
{
// Replace any semicolon with a comma.
   $w = Str_Replace(';' , ',', Strip_WSpace($LatLonString));

// Handle empty argument or single comma.
   if ($w == '' or $w == ',') {$w = '0  ,  0';}

// Handle single space.
   if (Substr_Count($w, ' ') == 1) {$w = Str_Replace(' ', '  ,  ', $w);}

// Check for initial or terminal comma.
   if (substr($w,0,1) == ',') {$w = "0  $w;";}
   if (substr($w, -1) == ',') {$w = $w.'  0';}

// Get latitude and longitude values from Lat/Lon string.
   @list($latitude, $longitude) = PReg_Split("[,]", $w);

// Convert DMS strings to decimal degrees, if necessary.
   $LatDeg = SPrintF("%+2.10f", DMS_to_Deg(trim($latitude)));
   $LonDeg = SPrintF("%+2.10f", DMS_to_Deg(trim($longitude)));

   return "$LatDeg  ,  $LonDeg";

} // END OF  Check_Lat_Lon (LatLonString)





/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/




/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ==========================
   PHYSICAL SCIENCE FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 PHYSICAL SCIENCE-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 SYSTEM MODULES AND RELATED DATA AND
   GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/


/*
   This function computes the gravitational force between
   greater mass (M) and lesser mass (m) at any distance
   using the constants G_UNI and N_PER_LB.


   MassKg = Greater mass in kg
   massKg = Lesser mass in kg
   F      = gravitational force between masses
   R      = Distance between masses

   NorL   = Returned units of force = 'N' = Newtons = Default
            else = 'L' = Pounds (lb) of force returned.

   The returned force is in Newtons, by default.  If not
   Newtons, then pounds force will be assumed as the
   alternative default.

   Input mass and distance arguments are in SI units
   where: length/mass/time = meter/kg/second

   NOTE:
   Uses defined constants  G_UNI  and  N_PER_LB

*/
   function F_Gravity ($MassKg, $massKg, $DistanceMeters, $NorL='N')
{
   $M = FloatVal(trim($MassKg));
   $m = FloatVal(trim($massKg));
   $R = FloatVal(trim($DistanceMeters));
   $U = substr(StrToUpper(trim($NorLb)),0,1);
   $F = G_UNI * $M * $m / $R / $R / ((($U == 'L')? N_PER_LB : 1.0));

   return trim($F);
}





/*
  This function returns the energy (E) passing through a
  one square meter spatial cross-section at any given
  temperature on the F, C,K, or R scales.

  TempAndScaleStr = Temperature and scale symbol string.
                    Examples: '25.6 C'  or  '-25.8 F'

  E = Joules/s = Watts per square meter per second

  Dependency:
  Temp_to_Temp() function

*/


   function Temp_FCKR_to_E ($TempAndScaleStr)
{
   $T_and_Scale = trim($TempAndScaleStr);

   $K = Temp_to_Temp ($T_and_Scale, 'K');

   return BCMul('0.000000056704', BCPow($K, '4', 20), 16);
}




/*
   This function is the inverse of the Temp_to_E () function.

   Given the ambient energy (E Joules/s/m^2), this function returns
   the corresponding temperature in the given scale (F, C, K, or R).

   Dependency:
   Temp_to_Temp() function

*/


   function E_to_Temp_FCKR($EJoules, $TScaleStr='K')
{
   $E = FloatVal(trim($EJoules));

   $T_Scale = StrToUpper(substr(trim($TScaleStr),0,1));

   $K = SqRt(SqRt($E / 0.000000056704));

   return Temp_to_Temp("$K K", $T_Scale);
}





/*
   Relativistic Motion type 1.

   NOT DONE - STILL BUGGY

   Accelerate from rest with experience G until reaching distance D.

   Terminal speed at distance D is relativistic speed beta.

   input:
   Distance
   TimeUnits
   GFactor
*/

   FUNCTION xRelativistic_Motion_Type_1 ($Distance, $TimeUnits, $GFactor='1')
{
// Accelerating to distance (D) light units
// at onboard experiential factor (G) and
// reaching terminal speed (beta).

// Set for 32-decimal arbitrary precision limit.
   $Q = 32;

// Define applied constants.
   $c = '299792458';
   $g = '9.80665';



// Read input arguments.
   $D = trim($Distance); // value and units symbol ('123 km')
   $G = trim($GFactor);
   $TUnits = substr(StrToLower(trim($TimeUnits)),0,1);

// Convert distance to meters.
   $D = Length_to_Length ($D, 'm');

// Determine (S) value to use according to TimeUnits symbol.
// Its value equates to the number of seconds in the time unit.
   if ($TUnits == 'y') {$S = '31557600';}
   if ($TUnits == 'd') {$S = '86400';}
   if ($TUnits == 'h') {$S = '3600';}
   if ($TUnits == 'm') {$S = '60';}
   if ($TUnits == 's') {$S = '1';}

// Compute acceleration potential.
// $a = $G * ($g * $S / $c);
   $a = BCMul($G,BCDiv(BCMul($g,$S,$Q),$c,$Q),$Q);






// Compute Earth time (T) taken to
// accelerate to distance (D).
// $T = SqRt($D*$D + 2*$D/$a);
   $DD = BCMul($D,$D,$Q);
   $TwoD = BCMul('2',$D,$Q);
   $TwoD_a = BCDiv($TwoD,$a,$Q);

   $DD_plus_TwoD_a = BCAdd($DD,$TwoD_a,$Q);
   $T = BCSqRt($DD_plus_TwoD_a,$Q);







// Relativistic velocity at distance (D).
// $beta = $T / ($D + (1/$a));
   $One_a = BCDiv('1',$a,$Q);
   $DplusOne_a = BCAdd($D,$One_a,$Q);
   $beta = BCDiv($T,$DplusOne_a,$Q);

// Spacecraft time to reach distance (D).

// $t = log(   (1 + $beta) / SqRt(1 - $beta*$beta)   ) / $a;
   $OnePlusbeta = BCAdd('1',$beta,$Q);
   $beta2 = BCMul($beta,$beta,$Q);
   $OneMinusbeta2 = BCSub('1',$beta2,$Q);
   $SqRtOneMinusbeta2 = bcSqRt($OneMinusbeta2,$Q);

   $w = BCDiv($OnePlusbeta,$SqRtOneMinusbeta2,$Q);
   $t = log($w)/$a;
//   exit("t = $t");

   $Time_Diff_y = BCSub($T,$t,$Q);
   $Time_Diff_d = BCMul('365.25',$Time_Diff_y,$Q);

// Reduce output to 32-decimal precision, except for
// the time values (T,t), which are set to 16.
   $D = BCAdd(0,$D,32);
   $G = BCAdd(0,$G,32);
   $T = BCAdd(0,$T,16);
   $t = BCAdd(0,$t,16);
   $beta = BCAdd(0,$beta,32);

// Done.  Return 'D G T t beta' values in
// a space-delimited string to 32 decimals.
   $OutputText = "$D $G $T $t $beta";

   return trim($OutputText);
}




/*
   This function returns the elements of linear relativistic
   acceleration in a straight line.

   AccelRate    = Rate of acceleration in m/s^2
   DistAndUnits = Distance value and units symbol separated by a space.
   $Decimals    = 0 to 50

*/

   FUNCTION Relativistic_Motion_Type_1 ($AccelRate, $DistAndUnits, $Decimals=32)
{
   $c = '299792458';
   $S = '1';
   $a = trim($AccelRate);
   $Q = trim($Decimals);

// Get distance and units argument string. There must be
// at least one space between the number and units symbol.
// EXAMPLE: '123.456 km'
   $Dist_and_Units = trim($DistAndUnits);

// Convert given distance into distance D in light seconds.
   $D = BCLength_to_Length($Dist_and_Units,'LS',64);

   $A = BCDiv(BCMul($a,$S,64),$c,64);
   $T = BCSqRt(BCAdd(BCMul($D,$D,64),BCDiv(BCMul('2',$D,64),$A,64),64),64);
   $b = BCDiv($T,BCAdd($D,BCDiv('1',$A,64),64),64);
   $t = BCDiv(BCLn(BCDiv(BCAdd('1',$b,64),BCSqRt(BCSub('1',BCMul($b,$b,64),64),64),64)),$A,64);
   $v = BCMul($b,$c,64);

   $T_days = BCDiv($T,'86400',64);
   $T_dhms = Seconds_to_WDHMS($T,$Q);

   $t_days = BCDiv($t,'86400',64);
   $t_dhms = Seconds_to_WDHMS($t,$Q);
   $A = BCAdd('0',$A,$Q);
   $T = BCAdd('0',$T,$Q);
   $t = BCAdd('0',$t,$Q);
   $v = BCAdd('0',$v,$Q).' m/s';
   $b = BCAdd('0',$b,$Q).'c';

   return "$a $Dist_and_Units $A $T_dhms $t_dhms $b $v";

}



/*
   EXPERIMENTA FIX FOR ABOVE - STILL BUGGY
   IN (t) decimals.
*/
   function Rel_Type_1 ($AccelRate, $DistAndUnits, $Decimals=32)
{
   $c = '299792458';
   $S = '1';
   $a = ''.trim($AccelRate);
   $Q = trim($Decimals);

// Get distance and units argument string. There must be
// at least one space between the number and units symbol.
// EXAMPLE: '123.456 km'
   $Dist_and_Units = trim($DistAndUnits);

// Convert given distance into distance D in light seconds.
   $D = BCLength_to_Length($Dist_and_Units,'LS',64);

   $A = BCDiv(BCMul($a,$S,64),$c,64);
   $T = BCSqRt(BCAdd(BCMul($D,$D,64),BCDiv(BCMul('2',$D,64),$A,64),64),64);
   $b = BCDiv($T,BCAdd($D,BCDiv('1',$A,64),64),64);
   $t = BCDiv(BCLn(BCDiv(BCAdd('1',$b,64),BCSqRt(BCSub('1',BCMul($b,$b,64),64),64),64)),$A,64);
   $v = BCMul($b,$c,64);

//   exit("$T<br>$t");

   $T_days = BCDiv($T,'86400',64);
   $T_dhms = Seconds_to_WDHMS($T,$Q);

   $t_days = BCDiv($t,'86400',64);
   $t_dhms = Seconds_to_WDHMS($t,$Q);
   $A = BCAdd('0',$A,$Q);
   $T = BCAdd('0',$T,$Q).'s';
   $t = BCAdd('0',$t,$Q).'s';
   $v = BCAdd('0',$v,$Q).' m/s';
   $b = BCAdd('0',$b,$Q).'c';

   return "$T<br>$t<br>$b";

}







/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/






/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   =====================
   RANDOMIZING FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 RANDOMIZING FUNCTIONS
   FOR USE WITH THE PHPSL G2 SYSTEM MODULES AND RELATED
   DATA AND GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/



/*
   This function generates a string of random digits.

*/

   function Rand_Digits ($NumberOfDigits)
{
   $Number_of_Digits = abs(IntVal(trim($NumberOfDigits)));

   if ($Number_of_Digits == 0) {$Number_of_Digits = 1;}

   $Random_Digits_String = '';

   for($i=0;   $i < $Number_of_Digits;   $i++)
      {$Random_Digits_String .= substr('0123456789', mt_rand(0, 9), 1);}

   return $Random_Digits_String;
}








   function Rand_HMS ($HMSStr='hhmmss', $Mode1224=24, $SignedFlag=FALSE)
  {
   if ($Mode1224 == 12)
      {$hh = SPrintF("%02d", mt_rand(1, 12));}
   else
      {$hh = SPrintF("%02d", mt_rand(0, 23));}

   $mm = ':'.SPrintF("%02d", mt_rand(0, 59));
   $ss = ':'.SPrintF("%02d", mt_rand(0, 59));

   $hms = StrToLower(trim($HMSStr));
   if ($hms == '')     {$hms = 'hhmmss';}
   if ($hms == 'hh')   {$mm  = ''; $ss = '';}
   if ($hms == 'hhmm') {$ss  = '';}

// Attach random numerical () sign if (SignedFlag) is TRUE;
   $NumSign = '';
   if ($SignedFlag === TRUE)  {$NumSign = (mt_rand() % 2 == 0)? '-':'+';}

   return "$NumSign$hh$mm$ss";
  }



/*
   This function returns random calendar
   date elements year, month number and
   day number in the format 'Y mn dn'.

   All dates generated are valid calendar
   dates on the Gregorian calendar within
   the range specified by values of the
   MinJDNum and MaxJDNum arguments.

   Default years range = 1600 to 2200

   Example: '1791 04 23' (1791 April 23)
*/

   function Rand_YMD ($MinJDNum=2305446, $MaxJDNum=2524960)
{
   list ($m, $d, $Y) =
   PReg_Split("[\/]", JDtoGregorian(mt_rand($MinJDNum, $MaxJDNum)));

   $mn = SPrintF(" %02d ", $m);
   $dn = SPrintF(" %02d ", $d);

   return "$Y$mn$dn";
}



/*
   This function returns a randomly selected body name.
   Bodies = Sun, Moon, planets and Pluto.

// DEPENDENCY
   Check_Body_Name()

*/

   function Rand_Body_Name()
{
   return Check_Body_Name(substr('SuMoMeVeMaJuSaUrNePl', 2*mt_rand(0,9),2));
}

/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/





/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ===========================
   GENERAL ASTRONOMY FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 ASTRONOMY-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 SYSTEM MODULES AND RELATED DATA AND
   GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/



/*
   NASA Delta T Polynomial Expressions

   All integer-encoded partial dates MUST start with a () sign.
   Anything without a sign will be considered a Julian Date or a
   JD number.

   Return mode 1 (default) returns result resolved to nearest second (+444).
   Return mode 2 returns full polynomial result in seconds (444.413605555556).
   Return mode 3 returns HH:MM:SS string (+00:07:24).
   Return mode 4 returns h m s string (+00h 07m 24s).
   Return mode 5 returns Delta T as a fraction of a day.
*/

   function NASA_Delta_T_Poly ($JDTTArg, $JGMode='G', $RetMode12345=1)
{
// Get date elements corresponding to JDTT argument.
// Only the (Y, m) values will be used.

   $JDTT    = trim($JDTTArg);
   $JDNumTT = floor($JDTT + 0.5);

   $JG = (StrToUpper(substr(trim($JGMode),0,1)) == 'J')? 'J':'G';

   $w = trim($RetMode12345);
   if ($w == '' or IntVal($w) < 1) {$w = 1;}
   if (IntVal($w) > 5) {$w = 5;}
   $RetMode = $w;

// Determine (Y, m, d) values from JD Number and JG mode.
// Only the (Y, m) values will be used.
   list($Y, $m, $d) = PReg_Split("[ ]", Inv_JD_Num($JDNumTT, $JG));


// ERROR if year value is invalid.  It must be
// in the range from -2000 to 3000.

   if ($Y < -2000 or $Y == 0 or $Y > 3000) {return FALSE;}

// ERROR if month value is outside the range from 1 to 12.

   if ($m < 1 or $m > 12) {return FALSE;}

/*
   If a negative (BC) year is given , it is converted into the
   corresponding mathematical year by adding 1 to (Y).

   The year 100 BC (-100), becomes the mathematical year -99
   for proper use within internal computations.

   Compute working (y) value for mid-month (m).
*/
   $y = $Y + (($Y < 0)? 1:0) + ($m - 0.5)/12;

// Perform NASA Delta T polynomial computations using (Y, y) values.

   $dTSec = FALSE;

   if ($Y >= -2000 and $Y < -500)
      {
       $u  = ($Y - 1820) / 100;

       $dTSec = 32*$u*$u - 20;
      }

   if ($Y >= -500 and $Y < 500)
      {
       $u  = $y/100;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;
       $u6 = $u * $u5;

       $dTSec = 10583.6 - 1014.41*$u
                        + 33.78311*$u2
                        -  5.952053*$u3
                        -  0.1798452*$u4
                        +  0.022174192*$u5
                        +  0.0090316521*$u6;
      }


   if ($Y >= 500 and $Y < 1600)
      {
       $u  = ($y - 1000)/100;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;
       $u6 = $u * $u5;

       $dTSec = 1574.2 - 556.01*$u
                       +  71.23472*$u2
                       +   0.319781*$u3
                       -   0.8503463*$u4
                       -   0.005050998*$u5
                       +   0.0083572073*$u6;
      }


   if ($Y >= 1600 and $Y < 1700)
      {
       $u  = $y - 1600;

      $dTSec = 120 - 0.9808*$u - 0.01532*$u*$u + $u*$u*$u/7129;
      }


   if ($Y >= 1700 and $Y < 1800)
      {
       $u  = $y - 1700;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;

       $dTSec = 8.83 + 0.1603*$u
                     - 0.0059285*$u2
                     + 0.00013336*$u3
                     - $u4/1174000;
      }


   if ($Y >= 1800 and $Y < 1860)
      {
       $u  = $y - 1800;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;
       $u6 = $u * $u5;
       $u7 = $u * $u6;

       $dTSec = 13.72 - 0.332447*$u
                      + 0.0068612*$u2
                      + 0.0041116*$u3
                      - 0.00037436*$u4
                      + 0.0000121272*$u5
                      - 0.0000001699*$u6
                      + 0.000000000875*$u7;
      }


   if ($Y >= 1860 and $Y < 1900)
      {
       $u  = $y - 1860;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;

       $dTSec = 7.62 + 0.5737*$u
                     - 0.251754*$u2
                     + 0.01680668*$u3
                     - 0.0004473624*$u4
                     + $u5 / 233174;
      }


   if ($Y >= 1900 and $Y < 1920)
      {
       $u  = $y - 1900;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;

       $dTSec = 1.494119*$u - 2.79
                            - 0.0598939*$u2
                            + 0.0061966*$u3
                            - 0.000197*$u4;
      }


   if ($Y >= 1920 and $Y < 1941)
      {
       $u  = $y - 1920;
       $u2 = $u * $u;
       $u3 = $u * $u2;

       $dTSec = 21.20 + 0.84493*$u - 0.076100*$u2 + 0.0020936*$u3;
      }


   if ($Y >= 1941 and $Y < 1961)
      {
       $u  = $y - 1950;
       $u2 = $u * $u;
       $u3 = $u * $u2;

       $dTSec = 29.07 + 0.407*$u - $u2/233 + $u3/2547;
      }


   if ($Y >= 1961 and $Y < 1986)
      {
       $u  = $y - 1975;
       $u2 = $u * $u;
       $u3 = $u * $u2;

       $dTSec = 45.45 + 1.067*$u - $u2/260 - $u3/718;
      }


   if ($Y >= 1986 and $Y < 2005)
      {
       $u  = $y - 2000;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;

       $dTSec = 63.86 + 0.3345*$u
                      - 0.060374*$u2
                      + 0.0017275*$u3
                      + 0.000651814*$u4
                      + 0.00002373599*$u5;
      }


   if ($Y >= 2005 and $Y < 2050)
      {
       $u = $y - 2000;

       $dTSec = 62.92 + 0.32217*$u + 0.005589*$u*$u;
      }


   if ($Y >= 2050 and $Y <= 2150)
      {
       $u  = ($y - 1820) / 100;

       $dTSec = 32*$u*$u - 0.5628*(2150 - $y) - 20;
      }


   if ($Y > 2150 and $Y <= 3000)
      {
       $u  = ($y - 1820) / 100;

       $dTSec = 32*$u*$u - 20;
      }

// Format returned value according to return mode.
   if ($RetMode == 1) {return SPrintF("%+d", round($dTSec, 0));}

   if ($RetMode == 2) {return $dTSec;} // Raw, No rounding

   if ($RetMode == 3 or $RetMode == 4)
      {
       $NumSign = ($dTSec < 0)? '-':'+';
       $hours   = abs($dTSec/3600);
       $hh      = floor($hours);
       $minutes = 60*($hours - $hh);
       $mm      = floor($minutes);
       $ss      = round(60*($minutes - $mm), 0);
       if ($ss == 60){$mm++; $ss=0;}
       if ($mm == 60){$hh++; $mm=0;}
       $hh      = SPrintF("%02d", $hh);
       $mm      = SPrintF("%02d", $mm);
       $ss      = SPrintF("%02d", $ss);

       if ($RetMode == 3)
          {$hhmmssDT = "$NumSign$hh:$mm:$ss";}
       else
          {$hhmmssDT = "$NumSign$hh"."h $mm"."m $ss"."s";}

       return $hhmmssDT;
      }

   if ($RetMode == 5) {return SPrintF("%+1.15f", round($dTSec, 0) / 86400);}

   return $dTSec; // = FALSE on error, if year is out of valid range.

} // END OF  NASA_Delta_T_Poly (...)





/*
  ========================================================================
  Computes the nutation in obliquity (delta epsilon) in
  degrees according to the IAU 2000B nutation series.

  This value is applied to the mean obliquity of the ecliptic to
  obtain the TRUE, apparent obliquity.

  The series consists of 77 terms.

  ARGUMENT:
  $JD = JD number at the moment (TDB) for which the
        nutation in obliquity is required.

  ERRORS:
  No error checking.

  ========================================================================
*/

   function Delta_Eps ($JDTTArg)
{
   $T = (FloatVal(trim($JDTTArg)) - 2451545) / 36525;

   $T2 = $T * $T;
   $T3 = $T * $T2;
   $T4 = $T * $T3;

// -------------------------------------------
// Compute mean anomaly of the Moon in radians

   $L  = Deg2Rad((485868.249036 + 1717915923.2178*$T + 31.8792*$T2
       + 0.051635*$T3 - 0.0002447*$T4) / 3600);


// ------------------------------------------
// Compute mean anomaly of the Sun in radians

   $Lp = Deg2Rad((1287104.79305 + 129596581.0481*$T
       - 0.5532*$T2  + 0.000136*$T3 - 0.00001149*$T4) / 3600);


// ------------------------------------------------------------
// Compute mean argument of the latitude of the Moon in radians

   $F  = Deg2Rad((335779.526232 + 1739527262.8478*$T
       - 12.7512*$T2 - 0.001037*$T3 + 0.00000417*$T4) / 3600);


// -----------------------------------------------------------
// Compute mean elongation of the Moon from the Sun in radians

   $D  = Deg2Rad((1072260.70369 + 1602961601.2090*$T
       - 6.3706*$T2  + 0.006593*$T3 - 0.00003169*$T4) / 3600);


// ------------------------------------------------------------------------
// Compute mean longitude of the mean ascending node of the Moon in radians

   $Om = Deg2Rad((450160.398036 - 6962890.5431*$T
       + 7.4722*$T2  + 0.007702*$T3 - 0.00005939*$T4) / 3600);

// --------------------------------------------------------------
// Sum series for nutation in obliquity (dEps) in arc sec * 10E+7

$s = 0;
$s += (92052331 + 9086*$T)*cos($Om) + 15377*sin($Om);
$s += (5730336 - 3015*$T)*cos(2*($F - $D + $Om)) - 4587*sin(2*($F - $D + $Om));
$s += (978459 - 485*$T)*cos(2*($F + $Om)) + 1374*sin(2*($F + $Om));
$s += (-897492 + 470*$T)*cos(2*$Om) - 291*sin(2*$Om);
$s += (73871 - 184*$T)*cos($Lp) - 1924*sin($Lp);
$s += (224386 - 677*$T)*cos($Lp + 2*($F - $D + $Om)) - 174*sin($Lp + 2*($F - $D + $Om));
$s -= 6750*cos($L) - 358*sin($L);
$s += (200728 + 18*$T)*cos(2*$F + $Om) + 318*sin(2*$F + $Om);
$s += (129025 - 63*$T)*cos($L + 2*($F + $Om)) + 367*sin($L + 2*($F + $Om));
$s += (-95929 + 299*$T)*cos(2*($F - $D + $Om) - $Lp) + 132*sin(2*($F - $D + $Om) - $Lp);
$s += (-68982 - 9*$T)*cos(2*($F - $D) + $Om) + 39*sin(2*($F - $D) + $Om);
$s += (-53311 + 32*$T)*cos(2*($F + $Om) - $L) - 4*sin(2*($F + $Om) - $L);
$s -= 1235*cos(2*$D - $L) - 82*sin(2*$D - $L);
$s -= 33228*cos($L + $Om) + 9*sin($L + $Om);
$s += 31429*cos($Om - $L) - 75*sin($Om - $L);
$s += (25543 - 11*$T)*cos(2*($F + $D + $Om) - $L) + 66*sin(2*($F + $D + $Om) - $L);
$s += 26366*cos($L + 2*$F + $Om) + 78*sin($L + 2*$F + $Om);
$s += (-24236 - 10*$T)*cos(2*($F - $L) + $Om) + 20*sin(2*($F - $L) + $Om);
$s -= 1220*cos(2*$D) - 29*sin(2*$D);
$s += (16452 - 11*$T)*cos(2*($F + $D + $Om)) + 68*sin(2*($F + $D + $Om));
$s -= 13870*cos(-2*($Lp - $F + $D - $Om));
$s += 477*cos(2*($D - $L)) - 25*sin(2*($D - $L));
$s += (13238 - 11*$T)*cos(2*($L + $F + $Om)) + 59*sin(2*($L + $F + $Om));
$s += (-12338 + 10*$T)*cos($L + 2*($F - $D + $Om)) - 3*sin($L + 2*($F - $D + $Om));
$s -= 10758*cos(2*$F + $Om - $L) + 3*sin(2*$F + $Om - $L);
$s -= 609*cos(2*$L) - 13*sin(2*$L);
$s -= 550*cos(2*$F) - 11*sin(2*$F);
$s += (8551 - 2*$T)*cos($Lp + $Om) - 45*sin($Lp + $Om);
$s -= 8001*cos(-$L + 2*$D + $Om) + sin(-$L + 2*$D + $Om);
$s += (6850 - 42*$T)*cos(2*($Lp + $F - $D + $Om)) - 5*sin(2*($Lp + $F - $D + $Om));
$s -= 167*cos(2*($D - $F)) - 13*sin(2*($D - $F));
$s += 6953*cos($L - 2*$D + $Om) - 14*sin($L - 2*$D + $Om);
$s += 6415*cos($Om - $Lp) + 26*sin($Om - $Lp);
$s += 5222*cos(2*($F + $D) + $Om - $L) + 15*sin(2*($F + $D) + $Om - $L);
$s += (168 - $T)*cos(2*$Lp) + 10*sin(2*$Lp);
$s += 3268*cos($L + 2*($F + $D + $Om)) + 19*sin($L + 2*($F + $D + $Om));
$s += 104*cos(2*($F - $L)) + 2*sin(2*($F - $L));
$s -= 3250*cos($Lp + 2*($F + $Om)) + 5*sin($Lp + 2*($F + $Om));
$s += 3353*cos(2*($F + $D) + $Om) + 14*sin(2*($F + $D) + $Om);
$s += 3070*cos(2*($F + $Om) - $Lp) + 4*sin(2*($F + $Om) - $Lp);
$s += 3272*cos(2*$D + $Om) + 4*sin(2*$D + $Om);
$s -= 3045*cos($L + 2*($F - $D) + $Om) + sin($L + 2*($F - $D) + $Om);
$s -= 2768*cos(2*($L + $F - $D + $Om)) + 4*sin(2*($L + $F - $D + $Om));
$s += 3041*cos(2*($D - $L) + $Om) - 5*sin(2*($D - $L) + $Om);
$s += 2695*cos(2*($L + $F) + $Om) + 12*sin(2*($L + $F) + $Om);
$s += 2719*cos(2*($F - $D) + $Om - $Lp) - 3*sin(2*($F - $D) + $Om - $Lp);
$s += 2720*cos($Om - 2*$D) - 9*sin($Om - 2*$D);
$s -= 51*cos(-$L - $Lp + 2*$D) - 4*sin(-$L - $Lp + 2*$D);
$s -= 2206*cos(2*($L - $D) + $Om) - sin(2*($L - $D) + $Om);
$s -= 199*cos($L + 2*$D) - 2*sin($L + 2*$D);
$s -= 1900*cos($Lp + 2*($F - $D) + $Om) - sin($Lp + 2*($F - $D) + $Om);
$s -= 41*cos($L - $Lp) - 3*sin($L - $Lp);
$s += 1313*cos(-2*($L - $F - $Om)) - sin(-2*($L - $F - $Om));
$s += 1233*cos(3*$L + 2*($F + $Om)) + 7*sin(3*$L + 2*($F + $Om));
$s -= 81*cos(-$Lp + 2*$D) - 2*sin(-$Lp + 2*$D);
$s += 1232*cos($L - $Lp + 2*($F + $Om)) + 4*sin($L - $Lp + 2*($F + $Om));
$s -= 20*cos($D) + 2*sin($D);
$s += 1207*cos(2*($F + $D + $Om) - $L - $Lp) + 3*sin(2*($F + $D + $Om) - $L - $Lp);
$s += 40*cos(2*$F - $L) - 2*sin(2*$F - $L);
$s += 1129*cos(-$Lp + 2*($F + $D + $Om)) + 5*sin(-$Lp + 2*($F + $D + $Om));
$s += 1266*cos($Om - 2*$L) - 4*sin($Om - 2*$L);
$s -= 1062*cos($L + $Lp + 2*($F + $Om)) + 3*sin($L + $Lp + 2*($F + $Om));
$s -= 1129*cos(2*$L + $Om) + 2*sin(2*$L + $Om);
$s -= 9*cos($Lp + $D - $L);
$s += 35*cos($L + $Lp) - 2*sin($L + $Lp);
$s -= 107*cos($L + 2*$F) - sin($L + 2*$F);
$s += 1073*cos(2*($F - $D) + $Om - $L) - 2*sin(2*($F - $D) + $Om - $L);
$s += 854*cos($L + 2*$Om);
$s -= 553*cos($D - $L) + 139*sin($D - $L);
$s -= 710*cos(2*($F + $Om) + $D) + 2*sin(2*($F + $Om) + $D);
$s += 647*cos(2*($F + 2*$D + $Om) - $L) + 4*sin(2*($F + 2*$D + $Om) - $L);
$s -= 700*cos($Lp + $D + $Om - $L);
$s += 672*cos(-2*($Lp - $F + $D) + $Om);
$s += 663*cos($L + 2*($F + $D) + $Om) + 4*sin($L + 2*($F + $D) + $Om);
$s -= 594*cos(-2*($L - $F - $D - $Om)) + 2*sin(-2*($L - $F - $D - $Om));
$s -= 610*cos(2*$Om - $L) - 2*sin(2*$Om - $L);
$s -= 556*cos($L + $Lp + 2*($F - $D + $Om));

// ----------------------------
// Return nutation in obliquity
// expressed in degrees.

   return $s / 36000000000;


} // End of  dEps(...)



/*

   ________________________________________________________________________
   Level 0: No Dependencies

   This function computes the nutation in ecliptical longitude (Delta psi)
   in degrees according to the IAU 2000B nutation series.

   This value is applied to the mean ecliptical longitude to obtain the
   TRUE, apparent ecliptical longitude.

   The series consists of 77 terms.

   ARGUMENT:
   $JDTT = Astronomical JD number corresponding to date and time (TT).

   ERRORS:
   No error checking.
   ________________________________________________________________________
*/

   function Delta_Psi ($JDTTArg)
{
// ---------------------------------------------
// Compute time (T) in Julian centuries reckoned
// from J2000.0 and 4 powers of T.

   $T  = (FloatVal(trim($JDTTArg)) - 2451545) / 36525;

   $T2 = $T * $T;
   $T3 = $T * $T2;
   $T4 = $T * $T3;

// -------------------------------------------
// Compute mean anomaly of the Moon in radians

   $L  = Deg2Rad((485868.249036 + 1717915923.2178*$T + 31.8792*$T2
       + 0.051635*$T3 - 0.0002447*$T4) / 3600);


// ------------------------------------------
// Compute mean anomaly of the Sun in radians

   $Lp = Deg2Rad((1287104.79305 + 129596581.0481*$T
       - 0.5532*$T2  + 0.000136*$T3 - 0.00001149*$T4) / 3600);


// ------------------------------------------------------------
// Compute mean argument of the latitude of the Moon in radians

   $F  = Deg2Rad((335779.526232 + 1739527262.8478*$T
       - 12.7512*$T2 - 0.001037*$T3 + 0.00000417*$T4) / 3600);


// -----------------------------------------------------------
// Compute mean elongation of the Moon from the Sun in radians

   $D  = Deg2Rad((1072260.70369 + 1602961601.2090*$T
       - 6.3706*$T2  + 0.006593*$T3 - 0.00003169*$T4) / 3600);


// ------------------------------------------------------------------------
// Compute mean longitude of the mean ascending node of the Moon in radians

   $Om = Deg2Rad((450160.398036 - 6962890.5431*$T
       + 7.4722*$T2  + 0.007702*$T3 - 0.00005939*$T4) / 3600);

// -----------------------------------------------------------------------
// Sum 2000B series for nutation in longitude (dPsi) in arc sec * 10000000

$s = 0;
$s += (-172064161 - 174666*$T)*sin($Om) + 33386*cos($Om);
$s += (-13170906 - 1675*$T)*sin(2*($F - $D + $Om)) - 13696*cos(2*($F - $D + $Om));
$s += (-2276413 - 234*$T)*sin(2*($F + $Om)) + 2796*cos(2*($F + $Om));
$s += (2074554 + 207*$T)*sin(2*$Om) - 698*cos(2*$Om);
$s += (1475877 - 3633*$T)*sin($Lp) + 11817*cos($Lp);
$s += (-516821 + 1226*$T)*sin($Lp + 2*($F - $D + $Om)) - 524*cos($Lp + 2*($F - $D + $Om));
$s += (711159 + 73*$T)*sin($L) - 872*cos($L);
$s += (-387298 - 367*$T)*sin(2*$F + $Om) + 380*cos(2*$F + $Om);
$s += (-301461 - 36*$T)*sin($L + 2*($F + $Om)) + 816*cos($L + 2*($F + $Om));
$s += (215829 - 494*$T)*sin(2*($F - $D + $Om) - $Lp) + 111*cos(2*($F - $D + $Om) - $Lp);
$s += (128227 + 137*$T)*sin(2*($F - $D) + $Om) + 181*cos(2*($F - $D) + $Om);
$s += (123457 + 11*$T)*sin(2*($F + $Om) - $L) + 19*cos(2*($F + $Om) - $L);
$s += (156994 + 10*$T)*sin(2*$D - $L) - 168*cos(2*$D - $L);
$s += (63110 + 63*$T)*sin($L + $Om) + 27*cos($L + $Om);
$s += (-57976 - 63*$T)*sin($Om - $L) - 189*cos($Om - $L);
$s += (-59641 - 11*$T)*sin(2*($F + $D + $Om) - $L) + 149*cos(2*($F + $D + $Om) - $L);
$s += (-51613 - 42*$T)*sin($L + 2*$F + $Om) + 129*cos($L + 2*$F + $Om);
$s += (45893 + 50*$T)*sin(2*($F - $L) + $Om) + 31*cos(2*($F - $L) + $Om);
$s += (63384 + 11*$T)*sin(2*$D) - 150*cos(2*$D);
$s += (-38571 - $T)*sin(2*($F + $D + $Om)) + 158*cos(2*($F + $D + $Om));
$s += 32481*sin(2*($F - $Lp - $D + $Om));
$s -= 47722*sin(2*($D - $L)) + 18*cos(2*($D - $L));
$s += (-31046 - $T)*sin(2*($L + $F + $Om)) + 131*cos(2*($L + $F + $Om));
$s += 28593*sin($L + 2*($F - $D + $Om)) - cos($L + 2*($F - $D + $Om));
$s += (20441 + 21*$T)*sin(2*$F - $L + $Om) + 10*cos(2*$F - $L + $Om);
$s += 29243*sin(2*$L) - 74*cos(2*$L);
$s += 25887*sin(2*$F) - 66*cos(2*$F);
$s += (-14053 - 25*$T)*sin($Lp + $Om) + 79*cos($Lp + $Om);
$s += (15164 + 10*$T)*sin(2*$D - $L + $Om) + 11*cos(2*$D - $L + $Om);
$s += (-15794 + 72*$T)*sin(2*($Lp + $F - $D + $Om)) - 16*cos(2*($Lp + $F - $D + $Om));
$s += 21783*sin(2*($D - $F)) + 13*cos(2*($D - $F));
$s += (-12873 - 10*$T)*sin($L - 2*$D + $Om) - 37*cos($L - 2*$D + $Om);
$s += (-12654 + 11*$T)*sin($Om - $Lp) + 63*cos($Om - $Lp);
$s -= 10204*sin(2*($F + $D) - $L + $Om) - 25*cos(2*($F + $D) - $L + $Om);
$s += (16707 - 85*$T)*sin(2*$Lp) - 10*cos(2*$Lp);
$s -= 7691*sin($L + 2*($F + $D + $Om)) - 44*cos($L + 2*($F + $D + $Om));
$s -= 11024*sin(-2*$L + 2*$F) + 14*cos(2*($F - $L));
$s += (7566 - 21*$T)*sin($Lp + 2*($F + $Om)) - 11*cos($Lp + 2*($F + $Om));
$s += (-6637 - 11*$T)*sin(2*($F + $D) + $Om) + 25*cos(2*($F + $D) + $Om);
$s += (-7141 + 21*$T)*sin(2*($F + $Om) - $Lp) + 8*cos(2*($F + $Om) - $Lp);
$s += (-6302 - 11*$T)*sin(2*$D + $Om) + 2*cos(2*$D + $Om);
$s += (5800 + 10*$T)*sin($L + 2*($F - $D) + $Om) + 2*cos($L + 2*($F - $D) + $Om);
$s += 6443*sin(2*($L + $F - $D + $Om)) - 7*cos(2*($L + $F - $D + $Om));
$s += (-5774 - 11*$T)*sin(2*($D - $L) + $Om) - 15*cos(2*($D - $L) + $Om);
$s -= 5350*sin(2*($L + $F) + $Om) - 21*cos(2*($L + $F) + $Om);
$s += (-4752 - 11*$T)*sin(2*($F - $D) - $Lp + $Om) - 3*cos(2*($F - $D) - $Lp + $Om);
$s += (-4940 - 11*$T)*sin($Om - 2*$D) - 21*cos($Om - 2*$D);
$s += 7350*sin(2*$D - $L - $Lp) - 8*cos(2*$D - $L - $Lp);
$s += 4065*sin(2*($L - $D) + $Om) + 6*cos(2*($L - $D) + $Om);
$s += 6579*sin($L + 2*$D) - 24*cos($L + 2*$D);
$s += 3579*sin($Lp + 2*($F - $D) + $Om) + 5*cos($Lp + 2*($F - $D) + $Om);
$s += 4725*sin($L - $Lp) - 6*cos($L - $Lp);
$s -= 3075*sin(2*($F - $L  + $Om)) + 2*cos(2*($F - $L  + $Om));
$s -= 2904*sin(3*$L + 2*($F + $Om)) - 15*cos(3*$L + 2*($F + $Om));
$s += 4348*sin(2*$D - $Lp) - 10*cos(2*$D - $Lp);
$s -= 2878*sin($L - $Lp + 2*($F + $Om)) - 8*cos($L - $Lp + 2*($F + $Om));
$s -= 4230*sin($D) - 5*cos($D);
$s -= 2819*sin(2*($F + $D + $Om) - $L - $Lp) - 7*cos(2*($F + $D + $Om) - $L - $Lp);
$s -= 4056*sin(2*$F - $L) - 5*cos(2*$F - $L);
$s -= 2647*sin(2*($F + $D + $Om) - $Lp) - 11*cos(2*($F + $D + $Om) - $Lp);
$s -= 2294*sin($Om - 2*$L) + 10*cos($Om - 2*$L);
$s += 2481*sin($L + $Lp + 2*($F + $Om)) - 7*cos($L + $Lp + 2*($F + $Om));
$s += 2179*sin(2*$L + $Om) - 2*cos(2*$L + $Om);
$s += 3276*sin($Lp - $L + $D) + cos($Lp - $L + $D);
$s -= 3389*sin($L + $Lp) - 5*cos($L + $Lp);
$s += 3339*sin($L + 2*$F) - 13*cos($L + 2*$F);
$s -= 1987*sin(2*($F - $D) - $L + $Om) + 6*cos(2*($F - $D) - $L + $Om);
$s -= 1981*sin($L + 2*$Om);
$s += 4026*sin($D - $L) - 353*cos($D - $L);
$s += 1660*sin($D + 2*($F + $Om)) - 5*cos($D + 2*($F + $Om));
$s -= 1521*sin(2*($F + 2*$D + $Om) - $L) - 9*cos(2*($F + 2*$D + $Om) - $L);
$s += 1314*sin($Lp - $L + $D + $Om);
$s -= 1283*sin(2*($F - $Lp - $D) + $Om);
$s -= 1331*sin($L + 2*($F + $D) + $Om) - 8*cos($L + 2*($F + $D) + $Om);
$s += 1383*sin(2*($F - $L + $D + $Om)) - 2*cos(2*($F - $L + $D + $Om));
$s += 1405*sin(2*$Om - $L) + 4*cos(2*$Om - $L);
$s += 1290*sin($L + $Lp + 2*($F - $D + $Om));

// ----------------------------------------
// Return nutation in longitude in degrees.

   return $s / 36000000000;

} // End of  dPsi (JDTTArg)





/*
   ________________________________________________________________________
   Level 1: 3 Level 0 dependencies

   Level 0 dependencies:
   Mean_Eps()
   dEps()
   dPsi()

   Computes the nutation in right ascension or the equation of the
   equinoxes, in degrees.

   This value is the difference (True Equinox) minus (Mean Equinox).

   The IAU 2000B nutation series is applied.

   LEVEL 0 DEPENDENCIES:
   Mean_Eps()
   dEps()
   dPsi()

   NOTE:
   Degrees * 240 = Seconds of RA
   ________________________________________________________________________
*/

   function Delta_Equinox ($JDTTArg)
{
// _____________________________________________
// Compute cosine of TRUE obliquity of ecliptic.

   $JDTT = FloatVal($JDTTArg);

// ____________________________________________________________
// Compute and return the equation of the equinoxes in degrees.

   return Delta_Psi($JDTT) * cos(Deg2Rad(Mean_Eps($JDTT) + Delta_Eps($JDTT)));

} // End of  Delta_Equinox(JDTTArg)












/*
  ========================================================================
  Level 0: No Dependencies

  Computes the mean obliquity of the ecliptic in degrees for the current
  JDTT argument corresponding to any given date/time.

  This ecliptic obliquity function is based on a formula derived by
  J. Laskar, Astronomy and Astrophysics, 1968, Vol. 157, p68.

  The estimated accuracy of this formula is ?0.01 arc second between
  the years 1000 AD and 3000 AD and a few arc seconds after 10,000
  years.  It is only valid in the range ?10,000 years either way
  of J2000.0

  Over the long term, it is more accurate than the formula adopted by
  the International Astronomical Union (IAU).


  ---------
  ARGUMENT:
  $JDTT = JD number at the moment (TDB) for which the mean obliquity
          of the ecliptic is required.
  ========================================================================
*/

   FUNCTION Laskar_Mean_Eps ($JDTTArg)
{

// ----------------------------------------------
// Compute the time in Julian centuries reckoned
// from J2000.0 corresponding to the JD argument
// and then divide by 100 for decamillennia.

   $t = (FloatVal(trim($JDTTArg)) - 2451545) / 36525 / 100;

// ------------------------------------------
// Compute mean obliquity (e) of the ecliptic
// expressed in arc seconds.

   $p = $t*$t;

   $e = 84381.448 - 4680.93*$t;

   $e -=    1.55*$p;  $p *= $t;
   $e += 1999.25*$p;  $p *= $t;
   $e -=   51.38*$p;  $p *= $t;
   $e -=  249.67*$p;  $p *= $t;
   $e -=   39.05*$p;  $p *= $t;
   $e +=    7.12*$p;  $p *= $t;
   $e +=   27.87*$p;  $p *= $t;
   $e +=    5.79*$p;  $p *= $t;
   $e +=    2.45*$p;

// -----------------------------------
// Done.  Return obliquity in degrees.
   return $e / 3600;

} // End of  Laskar_Mean_Eps (...)



/*
   This function returns the mean obliquity of
   the ecliptic for any given JD TT argument.
*/
   function Mean_Eps ($JDTTArg)
{
   $T = (FloatVal(trim($JDTTArg)) - 2451545.0) / 36525.0;
   $e = 84381.406
       - 46.836769*$T
       -  0.0001831*$T*$T
       +  0.0020034*$T*$T*$T
       -  0.000000576*$T*$T*$T*$T
       -  0.00000004343*$T*$T*$T*$T*$T;

   return $e / 3600;

} // End of  Mean_Eps (...)







/*
   ==============================================================
   Level 0: No Dependencies

   Computes the local or Greenwich mean sidereal time in degrees
   for any given JD and geographic longitude arguments.

   ARGUMENTS:
   JDTT    = JD argument for any given date and time (TT)

   LonDeg  = Geographic longitude. (Negative = West)
             Default longitude = 0 degrees (= Greenwich)

   ERRORS:
   FALSE is returned if either argument is invalid.
   ==============================================================
*/

   function Mean_Sid_Time ($JDTTArg, $LonDeg=0)
{

   $JDTT = trim($JDTTArg);

// ------------------------------
// Check for invalid argument(s).
   if ($JDTT < 2305447 or $JDTT > 2524959 or !Is_Numeric($LonDeg)) {return FALSE;}

// ------------------------------------------------
// Compute time (t) in Julian millennia as reckoned
// from J2000.0 and corresponding to the JD number.
   $t = ($JDTT - 2451545) / 36525;

// ------------------------------------------------
// Compute Greenwich mean sidereal time in degrees.
   $GSTMeanDeg = 280.46061837 + 360.98564736629*($JDTT - 2451545)
               + ($t*$t * 0.000387993) - ($t*$t*$t / 38710000);

// -----------------------------------
// Compute local mean sidereal time at
// the given geographic longitude.
   $LSTMeanDeg = $GSTMeanDeg + $LonDeg;

// ----------------------------------------------------------------
// Modulate sidereal time angle to fall between 0 and +360 degrees.
   $w = abs($LSTMeanDeg);
   $w = (($LSTMeanDeg < 0)? -1:1)*($w - 360*floor($w/360));
   $w += ($w < 0)? 360:0;
   $LSTMeanDeg = $w;
   $LSTMeanHrs = $LSTMeanDeg / 15;

// -------------------------------------------
// Return mean sidereal time angle in degrees.
   return RTrim(RTrim($LSTMeanDeg,'0'), '.');

}  // End of  Mean_Sid_Time (...)






// ###########################################################################
// ###########################################################################
// SPECIAL MISCELLANEOUS UTILITY FUNCTIONS









   function Distance_Between_Bodies ($RAHrs1,$DeclDeg1,$DistAU1, $RAHrs2,$DeclDeg2,$DistAU2)
{
   $a1 = Deg2Rad(15*trim($RAHrs1));
   $d1 = Deg2Rad(trim($DeclDeg1));
   $R1 = FloatVal(trim($DistAU1));
   $a2 = Deg2Rad(15*FloatVal(trim($RAHrs2)));
   $d2 = Deg2Rad(FloatVal(trim($DeclDeg2)));
   $R2 = FloatVal(trim($DistAU2));

   $x1 = $R1 * cos($a1) * cos($d1);
   $y1 = $R1 * sin($a1) * cos($d1);
   $z1 = $R1 * sin($d1);

   $x2 = $R2 * cos($a2) * cos($d2);
   $y2 = $R2 * sin($a2) * cos($d2);
   $z2 = $R2 * sin($d2);

   $dx = ($x2 - $x1);
   $dy = ($y2 - $y1);
   $dz = ($z2 - $z1);

   $D = SqRt($dx*$dx + $dy*$dy + $dz*$dz);

   return $D;

} // end of  Distance_Between_Bodies (RAHrs1, ...)


/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/





/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ===============
   SOLAR FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 SOLAR-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 DE405 SOLAR SYSTEM MODULE AND
   RELATED DATA AND GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.xx

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/


// ===============================================
// This function generates a solar ephemeris table
// for the year and month arguments TT.

   function G2_Sun_Ephem_Table_TT ($Year, $Month, $HeaderFlag=FALSE)
{

// Read input argument values.
   $Y = IntVal(trim($Year));
   $m = IntVal(trim($Month));

// Compute JD Number TT for first day of the month.
   $JD_Num_1st_TT = GregorianToJD($m, 1, $Y);

// Get full Day-of-Week name.
//   $DoW_Name = JDDayOfWeek($JD_Num_TT, 1);

// Get full month name.
   $Month_Name = JDMonthName($JD_Num_1st_TT, 1);

// Determine number of days in the month.
   $Days_in_Month = Cal_Days_In_Month(CAL_GREGORIAN, $m, $Y);

// Load solar data file into working array.
   $wArray = file("data/DE405_G2_Sun_Geo_Equ_Coords.data");
   $wCount = count($wArray);

// Construct monthly solar ephemeris table with
// extra padding at each end of the month.
   $iStart = 2*($JD_Num_1st_TT - 2305446.0);
   $iEnd   = $iStart + 2*$Days_in_Month + 1;

   $Solar_Ephem_Table = '';

   for($i = $iStart - 4;    $i < $iEnd + 2;    $i++)
  {
   $Curr_Line = trim($wArray[$i]);
   $TJD_TT = SPrintF("%7.1f", 2305446 + $i/2);

// Parse the current data table line.
// 186137912-232216820983132108
   $RA_Hrs   = substr($Curr_Line,  0, 2) . '.' . substr($Curr_Line,  2, 7);
   $Decl_Deg = substr($Curr_Line,  9, 3) . '.' . substr($Curr_Line, 12, 6);
   $Dist_AU  = substr($Curr_Line, 18, 1) . '.' . substr($Curr_Line, 19, 9);

// Get event marker, if any.
   $mark = trim(StrRChr($Curr_Line, ' '));
   if ($mark <> 'Ph'
   and $mark <> 'Ah'
   and $mark <> 'MarEqu'
   and $mark <> 'JunSol'
   and $mark <> 'SepEqu'
   and $mark <> 'DecSol')  {$mark = '';}

// Get Date/Time TT and remove time string part.
   $w = trim(Inv_JD($TJD_TT));
   $w = Str_Replace('12:00:00', '', $w);
   $w = Str_Replace('00:00:00', '', $w);
   $Date_Time_String_TT = Str_Replace('+', '', $w);

// DO SOME OUTPUT FORMATTING.
   $RA_Hrs   = SPrintF("% 10.7f",  $RA_Hrs);
   $Decl_Deg = SPrintF("% +10.6f", $Decl_Deg);
   $Dist_AU  = SPrintF("% 12.9f",  $Dist_AU);

// If event is outside target month, then remove event marker.
   if ($i < $iStart-0 or $i > $iEnd-2) {$mark = '';}

   $w = "$TJD_TT  $Date_Time_String_TT $RA_Hrs  $Decl_Deg  $Dist_AU $mark";

// Construct and append line to output table.
   $Solar_Ephem_Table .= "$w\n";
  }
   $Solar_Ephem_Table = trim($Solar_Ephem_Table);

   if ($HeaderFlag === FALSE) {return $Solar_Ephem_Table;}


// Construct and return solar ephemeris table for Year/Month.
   return
"==============================================================================
SUN - Geocentric Equatorial Ephemeris

For $Month_Name of $Y
True Equator and Equinox of Date


  JD_TT     Calendar_Date    RA_Hours   Declin_Deg   Distance_AU
=========  ===============  ==========  ==========  ============
$Solar_Ephem_Table";
}



/*
   The March equinox is defined here as the moment the
   geocentric sun is on the equator, crossing from south
   to north.

   Scan the solar ephemeris table for March for the moment
   when the Geocentric sun has declination 0.0 degrees.
*/

   function JDTT_Mar_Equinox ($Year)
{
   $Y = floor(trim($Year));

   $March_Table = G2_Body_Ephem_Table_TT('Sun', $Y, 3);

// Find tabular line index where Sun's declination
// changes from negative to positive.
   $wArray = PReg_Split("[\n]", $March_Table);
   $wCount = count($wArray);

   $wTable = '';
// 2188 Sep 18 Thu 2520472.0  11.7779565  +1.441465  1.005397117
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
  {
   list
  (
   $wTJDTT,
   $wY,
   $wMmm,
   $wdn,
   $wDoW,
   $wRAHrs,
   $wDeclDeg,
   $wDistAU
  ) = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

  $wTable = '';

  if ($wDeclDeg >= 0) {break;}

  }

  } // end of   for (i=0; ...)


  for($j=$i-3;   $j < $i+4;   $j++)
 {
  $Curr_Line = trim($wArray[$j]);
  $MarEqu = ($i == $j)? ' MarEqu':'';
  $wTable .= "$Curr_Line$MarEqu\n";
 }

// Create interpolation work-table for March equinox.
   $wArray = PReg_Split("[\n]", trim($wTable));
   $wCount = count($wArray);

   $wTable = '';
// OLD 1939 Sep 24 Sun 2429530.5  12.0029241  -0.018823  1.003176623 SepEqu
// NEW 2429530.5  1939 Sep 24 Sun  12.0029241  -0.018823  1.003176623 SepEqu
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   list($TJD_TT, $q,$q,$q,$q, $q, $Decl_Deg)
   = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

//   $RA_Hrs += ($i >= 3)? 24:0;
   $Decl_Deg = SPrintF("% 10.6f", $Decl_Deg);

   $wTable .= "$Decl_Deg  $TJD_TT\n";
  }

   $JD_TT = Lagrange_Interp($wTable, 0);

   $Date_Time_TT = trim(Inv_JD($JD_TT));

   $JD_TT = SPrintF("%7.8f", $JD_TT);

   return "$JD_TT $Date_Time_TT MarEqu";

} // end of  JDTT_Mar_Equinox (Year)



/*
   The September equinox is defined here as the moment the
   geocentric sun is on the equator, crossing from north
   to south.

   Scan the solar ephemeris table for September for the moment
   when the Geocentric sun has declination 0.0 degrees.
*/

   function JDTT_Sep_Equinox ($Year)
{
   $Y = floor(trim($Year));

   $September_Table = G2_Body_Ephem_Table_TT('Sun', $Y, 9);

// Find tabular line index where Sun's declination
// changes from positive to negative.
   $wArray = PReg_Split("[\n]", $September_Table);
   $wCount = count($wArray);

   $wTable = '';
// 2188 Sep 18 Thu 2520472.0  11.7779565  +1.441465  1.005397117
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
  {
   list
  (
   $wTJDTT,
   $wY,
   $wMmm,
   $wdn,
   $wDoW,
   $wRAHrs,
   $wDeclDeg,
   $wDistAU
  ) = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

  $wTable = '';

  if ($wDeclDeg < 0) {break;}

  }

  } // end of   for (i=0; ...)

  for($j=$i-3;   $j < $i+4;   $j++)
 {
  $Curr_Line = trim($wArray[$j]);
  $SepEqu = ($i == $j)? ' SepEqu':'';
  $wTable .= "$Curr_Line$SepEqu\n";
 }

// +++++++++++
// Create interpolation work-table for September equinox.
// 2429530.5  1939 Sep 24 Sun  12.0029241  -0.018823  1.003176623 SepEqu
   $wArray = PReg_Split("[\n]", trim($wTable));
   $wCount = count($wArray);

   $wTable = '';
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   list($TJD_TT, $q,$q,$q,$q,$q, $Decl_Deg)
   = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

   $Decl_Deg = SPrintF("% 10.6f", $Decl_Deg);

   $wTable .= "$Decl_Deg  $TJD_TT\n";
  }

   $JD_TT = Lagrange_Interp($wTable, 0);
   $JD_TT = SPrintF("%7.8f", $JD_TT);

   $Date_Time_TT = trim(Inv_JD($JD_TT));

   return "$JD_TT $Date_Time_TT SepEqu";

} // End of  JDTT_Sep_Equinox ($Year)






   function JDTT_Jun_Solstice ($Year)
{
   $Y = floor(trim($Year));

   $June_Table = G2_Body_Ephem_Table_TT('Sun', $Y, 6);

// Find tabular line index where Sun's right ascension
// is >= 6 hours.
   $wArray = PReg_Split("[\n]", $June_Table);
   $wCount = count($wArray);

   $wTable = '';
// 2188 Sep 18 Thu 2520472.0  11.7779565  +1.441465  1.005397117
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
  {
   list
  (
   $wTJDTT,
   $wY,
   $wMmm,
   $wdn,
   $wDoW,
   $wRAHrs,
   $wDeclDeg,
   $wDistAU
  ) = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

  $wTable = '';

  if ($wRAHrs >= 6.0) {break;}

  }

  } // end of   for (i=0; ...)

  for($j=$i-3;   $j < $i+4;   $j++)
 {
  $Curr_Line = trim($wArray[$j]);
  $JunSol = ($i == $j)? ' JunSol':'';
  $wTable .= "$Curr_Line$JunSol\n";
 }

// +++++++++++++++++++++++++

// Create interpolation work-table for June solstice.
   $wArray = PReg_Split("[\n]", trim($wTable));
   $wCount = count($wArray);

   $wTable = '';
// OLD 1939 Sep 24 Sun 2429530.5  12.0029241  -0.018823  1.003176623 SepEqu
// NEW 2429530.5  1939 Sep 24 Sun   12.0029241  -0.018823  1.003176623 SepEqu

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   list($TJD_TT, $q,$q,$q,$q, $RA_Hrs)
   = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

   $RA_Hrs = SPrintF("% 10.6f", $RA_Hrs);

   $wTable .= "$RA_Hrs  $TJD_TT\n";
  }

   $JD_TT = Lagrange_Interp($wTable, 6);
   $JD_TT = SPrintF("%7.8f", $JD_TT);

   $Date_Time_TT = trim(Inv_JD($JD_TT));

   return "$JD_TT $Date_Time_TT JunSol";

} // end of  JDTT_Jun_Solstice ($Year)




   function JDTT_Dec_Solstice ($Year)
{
   $Y = floor(trim($Year));

   $December_Table = G2_Body_Ephem_Table_TT('Sun', $Y, 12);

// Find tabular line index where Sun's right ascension
// is >= 18 hours.
   $wArray = PReg_Split("[\n]", $December_Table);
   $wCount = count($wArray);

   $wTable = '';
// 2188 Sep 18 Thu 2520472.0  11.7779565  +1.441465  1.005397117

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
  {
   list
  (
   $wTJDTT,
   $wY,
   $wMmm,
   $wdn,
   $wDoW,
   $wRAHrs,
   $wDeclDeg,
   $wDistAU
  ) = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

  $wTable = '';

  if ($wRAHrs >= 18.0) {break;}

  }

  } // end of   for (i=0; ...)

  for($j=$i-3;   $j < $i+4;   $j++)
 {
  $Curr_Line = trim($wArray[$j]);
  $DecSol = ($i == $j)? ' DecSol':'';
  $wTable .= "$Curr_Line$DecSol\n";
 }

// +++++++++++++++++++++++

// Create interpolation work-table for June solstice.
   $wArray = PReg_Split("[\n]", trim($wTable));
   $wCount = count($wArray);

   $wTable = '';
// 2429530.5  1939 Sep 24 Sun  12.0029241  -0.018823  1.003176623 SepEqu
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   list($TJD_TT, $q,$q,$q,$q, $RA_Hrs)
   = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

   $RA_Hrs = SPrintF("% 10.6f", $RA_Hrs);

   $wTable .= "$RA_Hrs  $TJD_TT\n";
  }

   $JD_TT = Lagrange_Interp($wTable, 18);
   $JD_TT = SPrintF("%7.8f", $JD_TT);

   $Date_Time_TT = trim(Inv_JD($JD_TT));

   return "$JD_TT $Date_Time_TT DecSol";

} // end of  JDTT_Dec_Solstice ($Year)




   function Equinoxes_and_Solstices_XT ($Year, $TZhhmm='TT')
{
   $Y = floor(trim($Year));
   $hhmm_TZ = StrToUpper(trim($TZhhmm));

// Determine XT (TT, UT, LT) time scale.
   $XT = '';
   if ($hhmm_TZ == 'TT') {$XT = 'TT';  $hhmm_TZ = '+00:00';}

   if ($hhmm_TZ == 'UT')
      {$XT = 'UT';  $hhmm_TZ = '+00:00';}

   if ($XT <> 'TT' and $XT <> 'UT')
      {
       $XT = 'LT';
       $w = HMS_to_Hours($hhmm_TZ);
       $w = Hours_to_HMS($w, 0, '+', ':');
       $hhmm_TZ = substr($w,0,6);
      }

   if ($hhmm_TZ == 'TT') {$XT = 'TT'; $hhmm_TZ = '+00:00';}
   if ($hhmm_TZ == 'UT') {$XT = 'UT'; $hhmm_TZ = '+00:00';}


// Compute equinox and solstice events tables TT.
   $Mar_Equ_TT = JDTT_Mar_Equinox  ($Y);
   $Jun_Sol_TT = JDTT_Jun_Solstice ($Y);
   $Sep_Equ_TT = JDTT_Sep_Equinox  ($Y);
   $Dec_Sol_TT = JDTT_Dec_Solstice ($Y);

// For TT values:
   if ($XT == 'TT')
  {
// Compute JD_TTs of events.
   $JD_TT_Mar_Equ = SPrintF("%7.8f", $Mar_Equ_TT);
   $JD_TT_Jun_Sol = SPrintF("%7.8f", $Jun_Sol_TT);
   $JD_TT_Sep_Equ = SPrintF("%7.8f", $Sep_Equ_TT);
   $JD_TT_Dec_Sol = SPrintF("%7.8f", $Dec_Sol_TT);

// Compute DT HMS estimates for event dates.
   $hhmmss_DT_Mar_Equ = ' '.NASA_Delta_T_Poly($JD_TT_Mar_Equ, 'G', 3);
   $hhmmss_DT_Jun_Sol = ' '.NASA_Delta_T_Poly($JD_TT_Jun_Sol, 'G', 3);
   $hhmmss_DT_Sep_Equ = ' '.NASA_Delta_T_Poly($JD_TT_Sep_Equ, 'G', 3);
   $hhmmss_DT_Dec_Sol = ' '.NASA_Delta_T_Poly($JD_TT_Dec_Sol, 'G', 3);

// Construct equinoxes and solstices table.
   $Mar_Equ_TT = substr($Mar_Equ_TT,0,41).$hhmmss_DT_Mar_Equ.' '.substr($Mar_Equ_TT, 42, StrLen($Mar_Equ_TT));
   $Jun_Sol_TT = substr($Jun_Sol_TT,0,41).$hhmmss_DT_Jun_Sol.' '.substr($Jun_Sol_TT, 42, StrLen($Jun_Sol_TT));
   $Sep_Equ_TT = substr($Sep_Equ_TT,0,41).$hhmmss_DT_Sep_Equ.' '.substr($Sep_Equ_TT, 42, StrLen($Sep_Equ_TT));
   $Dec_Sol_TT = substr($Dec_Sol_TT,0,41).$hhmmss_DT_Dec_Sol.' '.substr($Dec_Sol_TT, 42, StrLen($Dec_Sol_TT));
   return "$Mar_Equ_TT\n$Jun_Sol_TT\n$Sep_Equ_TT\n$Dec_Sol_TT\n";
  }


// @@@@@@@@@@@@@@@@@
// For UT and LT values.

   $JD_TT_Mar_Equ = SPrintF("%7.8f", $Mar_Equ_TT);
   $JD_TT_Jun_Sol = SPrintF("%7.8f", $Jun_Sol_TT);
   $JD_TT_Sep_Equ = SPrintF("%7.8f", $Sep_Equ_TT);
   $JD_TT_Dec_Sol = SPrintF("%7.8f", $Dec_Sol_TT);

// Compute DT HMS estimates for event dates.
   $hhmmss_DT_Mar_Equ = ' '.NASA_Delta_T_Poly($JD_TT_Mar_Equ, 'G', 3);
   $hhmmss_DT_Jun_Sol = ' '.NASA_Delta_T_Poly($JD_TT_Jun_Sol, 'G', 3);
   $hhmmss_DT_Sep_Equ = ' '.NASA_Delta_T_Poly($JD_TT_Sep_Equ, 'G', 3);
   $hhmmss_DT_Dec_Sol = ' '.NASA_Delta_T_Poly($JD_TT_Dec_Sol, 'G', 3);


   if ($XT == 'LT' or $XT == 'TT')
      {
       $hhmmss_DT_Mar_Equ = $hhmmss_DT_Jun_Sol = '';
       $hhmmss_DT_Sep_Equ = $hhmmss_DT_Dec_Sol = '';
      }

// LT
   $JD_LT_Mar_Equ = JDTT_to_JDLT($JD_TT_Mar_Equ, $hhmm_TZ);
   $JD_LT_Jun_Sol = JDTT_to_JDLT($JD_TT_Jun_Sol, $hhmm_TZ);
   $JD_LT_Sep_Equ = JDTT_to_JDLT($JD_TT_Sep_Equ, $hhmm_TZ);
   $JD_LT_Dec_Sol = JDTT_to_JDLT($JD_TT_Dec_Sol, $hhmm_TZ);

// UT
   $Mar_Equ_LT = $JD_LT_Mar_Equ.' '.trim(Inv_JD($JD_LT_Mar_Equ, 'AMPM')).' MarEqu';
   $Jun_Sol_LT = $JD_LT_Jun_Sol.' '.trim(Inv_JD($JD_LT_Jun_Sol, 'AMPM')).' JunSol';
   $Sep_Equ_LT = $JD_LT_Sep_Equ.' '.trim(Inv_JD($JD_LT_Sep_Equ, 'AMPM')).' SepEqu';
   $Dec_Sol_LT = $JD_LT_Dec_Sol.' '.trim(Inv_JD($JD_LT_Dec_Sol, 'AMPM')).' DecSol';




// @@@@@@@@@@@@@@@@@

   $XT = Str_Replace("LT", "LT UT$hhmm_TZ", $XT);

// For UT and LT values:
   if ($XT <> 'TT')
      {
        return "$Mar_Equ_LT\n$Jun_Sol_LT\n$Sep_Equ_LT\n$Dec_Sol_LT";
      }


// NEVER REACHED
// Construct output TZ line, if indicated.
//   $TZ_Text = ($XT == 'TT' or $XT == 'UT')? '':"       Local Time Zone UT$hhmm_TZ";


//   return trim($Output_Table);

} // end of  Equinoxes_and_Solstices_XT (Year, TZhhmm)












/*
   ===================================================
   This function returns an Earth apsides events table
   (except eclipses) for any given Year/Month in the
   600-year range from 1600 to 2200.
*/
   function Earth_Apsides_Table_TT ($Year)
{
   $Y = IntVal(trim($Year));

   $wArray = file('data/All_Earth_Apsides_and_Intervals.data');
   $wCount = count($wArray);

   $wTable  = '';
/*
2305807.77700848 1600 Dec 26 Tue 06:38:54 +00:01:59 18.3574365 -23.400118 0.983137814 184.29883941 367.15853238 Ph
*/
   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = Strip_WSpace($wArray[$i]);
   list
  (
   $JD_TT,
   $y,
   $Mmm,
   $dn,
   $DoW,
   $hhmmss_TT,
   $hhmmss_DT,
   $RA_Hrs,
   $Decl_Deg,
   $Dist_AU,
   $q,
   $q,
   $marker
  ) = PReg_Split("[ ]", $Curr_Line);

   if ($y == $Y)
      {

       $wTable .= "$JD_TT $y $Mmm $dn $DoW $hhmmss_TT  $Dist_AU $marker\n";
      }
  }
   return $wTable;
}







/*
   This function returns the ephemeris statistics for the Sun.

   The solar ephemeris statistics are:
   JDTT, Right Ascension, Declination and Distance,
   referred to the TRUE equator and equinox of date.
*/

   function G2_Sun_Ephem_Stats_TT ($JDTTArg)
{
// Read JDTT argument.
   $JDTT = FloatVal(trim($JDTTArg));

// Compute JD Number TT.
   $JDNumTT = floor($JDTT + 0.5);

// Get G2 Sun data file into work array and count file lines.
// File contains the Right Ascension, Declination and distance
// data for the Sun for the 600-year period from 1600 to 2200.
   $wArray = file('data/DE405_G2_Sun_Geo_Equ_Coords.data');
   $wCount = count($wArray);

// Construct main 7-point working tables centered
// on JD Number of date.
   $RAHrsTable = $DeclDegTable = $DistAUTable = '';

   $j = 2*($JDNumTT - 2305446.0);

   for ($i = $j-3;    $i < $j+4;    $i++)
  {
// Get current Sun ephemeris data table line.
   $CurrLine = trim($wArray[$i]);

// Parse the current ephemeris data line.
   $TJDTT    = SPrintF("%7.1f", 2305446.0 + $i/2);
   $RAHrs    = SPrintF("%010.7f",  substr($CurrLine,  0, 9) / 1E7);
   $DeclDeg  = SPrintF("%+010.6f", substr($CurrLine,  9, 9) / 1E6);
   $DistAU   = SPrintF("%011.9f",  substr($CurrLine, 18,10) / 1E9);

// Append current ephemeris data
// lines to interpolation tables.
   $RAHrsTable   .= "$TJDTT $RAHrs\n";
   $DeclDegTable .= "$TJDTT $DeclDeg\n";
   $DistAUTable  .= "$TJDTT $DistAU\n";
  }

// Delete work array after use.
   unset($wArray);

// Perform high-precision interpolations for RAHrs, DeclDeg
// and DistAU for the moment indicated by the JDTT argument.
   $RAHrs    = Interp_RA(trim($RAHrsTable), $JDTT);
   $DeclDeg  = Lagrange_Interp(trim($DeclDegTable), $JDTT);
   $DistAU   = Lagrange_Interp(trim($DistAUTable), $JDTT);

// Do some output formatting.
   $JDTT     = SPrintF("%7.8f",    $JDTT);
   $RAHrs    = SPrintF("% 10.7f",  $RAHrs);
   $DeclDeg  = SPrintF("%+ 10.6f", $DeclDeg);
   $DistAU   = SPrintF("% 12.9f",  $DistAU);

// Done.
   return "$JDTT $RAHrs $DeclDeg $DistAU";

} // end of  G2_Sun_Ephem_Stats_TT (YearNum, ...)











   function Earth_Perihelion_Aphelion_XT ($Year, $TZhhmm)
{
   $Y = IntVal(trim($Year));
   $w = StrToUpper(trim($TZhhmm));

   if ($w == 'TT') {return Earth_Apsides_Table_TT ($Y);}

   if ($w == 'UT') {$w = '+00:00';}

   $Frac_Day_TZ = HMS_to_Hours ($w);

   return "--- zzz ---";

}




   function JD_Mar_Equinox ($Year, $TZhhmm='+00:00', $AMPM='24')
{
   $Y  = floor(trim($Year));
   $AP = StrToUpper(trim($AMPM));

// Read TZ offset.
// Could be as  '-05:00', 'TT' or ' UT'
   $hhmm_TZ = StrToUpper(trim($TZhhmm));

// If TZ == TT Use JDTT version .
   if ($hhmm_TZ == 'TT')
      {return JDTT_Mar_Equinox($Y);}

// If TZ == UT Set TZ to +00:00
   if ($hhmm_TZ == 'UT') {$hhmm_TZ = '+00:00';}

// Compute JDTT of event.
   $JD_TT = FloatVal(JDTT_Mar_Equinox($Y));

// Convert TT to LT for given TZ offset.
   $JD_LT = $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);

// Compute date/time LT string.
   $Date_Time_LT = trim(Inv_JD($JD_LT, $AP));

   $JD_LT = SPrintF("%7.8f", $JD_LT);

   return "$JD_LT $Date_Time_LT MarEqu";

} // end of  JD_Mar_Equinox (Year, TZhhmm)



   function JD_Jun_Solstice ($Year, $TZhhmm='+00:00', $AMPM='24')
{
   $Y  = floor(trim($Year));
   $AP = StrToUpper(trim($AMPM));

// Read TZ offset.
// Could be as  '-05:00', 'TT' or ' UT'
   $hhmm_TZ = StrToUpper(trim($TZhhmm));

// If TZ == TT Use JDTT version .
   if ($hhmm_TZ == 'TT')
      {return JDTT_Jun_Solstice ($Y);}

// If TZ == UT Set TZ to +00:00
   if ($hhmm_TZ == 'UT') {$hhmm_TZ = '+00:00';}

// Compute JDTT of equinox event.
   $JD_TT = FloatVal(JDTT_Jun_Solstice ($Y));

// Convert TT to LT for given TZ offset.
   $JD_LT = $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);

// Compute date/time LT string.
   $Date_Time_LT = trim(Inv_JD($JD_LT, $AP));

   $JD_LT = SPrintF("%7.8f", $JD_LT);

   return "$JD_LT $Date_Time_LT JunSol";

} // end of  JD_Jun_Solstice (Year, TZhhmm)





   function JD_Sep_Equinox ($Year, $TZhhmm='+00:00', $AMPM='24')
{
   $Y  = floor(trim($Year));
   $AP = StrToUpper(trim($AMPM));

// Read TZ offset.
// Could be as  '-05:00', 'TT' or ' UT'
   $hhmm_TZ = StrToUpper(trim($TZhhmm));

// If TZ == TT Use JDTT version .
   if ($hhmm_TZ == 'TT')
      {return JDTT_Sep_Equinox($Y);}

// If TZ == UT Set TZ to +00:00
   if ($hhmm_TZ == 'UT') {$hhmm_TZ = '+00:00';}

// Compute JDTT of event.
   $JD_TT = FloatVal(JDTT_Sep_Equinox($Y));

// Convert TT to LT for given TZ offset.
   $JD_LT = $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);

// Compute date/time LT string.
   $Date_Time_LT = trim(Inv_JD($JD_LT, $AP));

   $JD_LT = SPrintF("%7.8f", $JD_LT);

   return "$JD_LT $Date_Time_LT SepEqu";

} // end of  JD_Sep_Equinox (Year, TZhhmm)



   function JD_Dec_Solstice ($Year, $TZhhmm='+00:00', $AMPM='24')
{
   $Y  = floor(trim($Year));
   $AP = StrToUpper(trim($AMPM));

// Read TZ offset.
// Could be as  '-05:00', 'TT' or ' UT'
   $hhmm_TZ = StrToUpper(trim($TZhhmm));

// If TZ == TT Use JDTT version .
   if ($hhmm_TZ == 'TT')
      {return JDTT_Dec_Solstice ($Y);}

// If TZ == UT Set TZ to +00:00
   if ($hhmm_TZ == 'UT') {$hhmm_TZ = '+00:00';}

// Compute JDTT of equinox event.
   $JD_TT = FloatVal(JDTT_Dec_Solstice ($Y));

// Convert TT to LT for given TZ offset.
   $JD_LT = $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);

// Compute date/time LT string.
   $Date_Time_LT = trim(Inv_JD($JD_LT, $AP));

   $JD_LT = SPrintF("%7.8f", $JD_LT);

   return "$JD_LT $Date_Time_LT DecSol";

} // end of  JD_Dec_Solstice (Year, TZhhmm)



/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/






/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ===============
   LUNAR FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 LUNAR-RELATED FUNCTIONS
   FOR USE WITH THE PHPSL G2 DE405 SOLAR SYSTEM MODULE AND
   RELATED DATA AND GRAPHICS FILES.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/


// ===============================================
// This function generates a lunar ephemeris table
// for the given Year/Month arguments TT.

   function G2_Moon_Ephem_Table_TT ($Year, $Month, $WithHeader=FALSE)
{
// Read input argument values.
   $Y = IntVal(trim($Year));
   $m = IntVal(trim($Month));

// Compute JD Number TT for first day of the month.
   $JD_Num_1st_TT = GregorianToJD($m, 1, $Y);

// Get full month name.
   $Month_Name = JDMonthName($JD_Num_1st_TT, 1);

// Determine number of days in the month.
   $Days_in_Month = Cal_Days_In_Month(CAL_GREGORIAN, $m, $Y);

// Load lunar data file into working array.
   $wArray = file("data/DE405_G2_Moon_Geo_Equ_Coords.data");
   $wCount = count($wArray);

// Construct monthly lunar ephemeris table with
// extra padding at each end of the month.
   $iStart = 2*($JD_Num_1st_TT - 2305446.0);
   $iEnd   = $iStart + 2*$Days_in_Month + 1;

   $Lunar_Ephem_Table = '';

   for($i = $iStart-4;    $i < $iEnd+2;    $i++)
  {
   $Curr_Line = trim($wArray[$i]);
   $TJD_TT = SPrintF("%7.1f", 2305446 + $i/2);

// Parse the current lunar data table line.
/*
005060737-0112806035683965618488150 Pg FM
*/
   $RA_Hrs    = substr($Curr_Line,  0, 2) . '.' . substr($Curr_Line,  2, 7);
   $Decl_Deg  = substr($Curr_Line,  9, 3) . '.' . substr($Curr_Line, 12, 6);
   $Dist_km   = substr($Curr_Line, 18, 6) . '.' . substr($Curr_Line, 21, 3);
   $Phase_Ang = substr($Curr_Line, 27, 3) . '.' . substr($Curr_Line, 30, 5);

   $w = substr(trim(Inv_JD($TJD_TT)),0,15); // yyyy Mmm dn DoW
   $Date_Time_String_TT = Str_Replace('+', '', $w);

// DO SOME OUTPUT FORMATTING.
   $RA_Hrs    = SPrintF("% 10.7f",  $RA_Hrs);
   $Decl_Deg  = SPrintF("% +10.6f", $Decl_Deg);
   $Dist_km   = SPrintF("%10.3f",   $Dist_km);
   $Phase_Ang = SPrintF("% 9.5f",   $Phase_Ang);

// Get marker for event(s). This could be 1 or 2 markers.
   $marker = ' '.trim(substr($Curr_Line, 35, StrLen($Curr_Line)));

// If event is outside target month, then remove event marker.
   if ($i < $iStart+0 or $i > $iEnd-2) {$marker = '';}

// Construct and append line to output table.
// 2314624.0 1625 Feb 14 Fri  3.9190571  +15.508702   392253.253    93.82299 FQ
   $Lunar_Ephem_Table .= "$TJD_TT  $Date_Time_String_TT  $RA_Hrs  $Decl_Deg   $Dist_km   $Phase_Ang$marker\n";
  }
   $Lunar_Ephem_Table = trim($Lunar_Ephem_Table);

   if($WithHeader === FALSE) {return $Lunar_Ephem_Table;}

// ----------------------------------------------------------
// Construct and return lunar ephemeris table for Year/Month.

   return
"==============================================================================
MOON - Geocentric Equatorial Ephemeris

For $Month_Name of $Y
True Ecliptic, Equator and Equinox of Date


  JD_TT     Calendar_Date    RA_Hours   Declin_Deg   Distance_km  Phase_Ang
=========  ===============  ==========  ==========  ============  =========
$Lunar_Ephem_Table";

} // end of  G2_Moon_Ephem_Table_TT (Year, Month)




/*
   This function returns a lunar events table (except eclipses)
   for the given Year/Month.
*/
   function Moon_Events_Table_TT ($Year, $MonthNum)
{
   $Y = IntVal(trim($Year));
   $m = IntVal(trim($MonthNum));

   $Mmm = substr('JanFebMarAprMayJunJulAugSepOctNovDec', 3*($m-1), 3);

/*
2305771.61450844 1600 Nov 20 Mon 02:44:54 Pg
2305772.13380350 1600 Nov 20 Mon 15:12:41 FM
*/

   $wArray = file('data/All_Lunar_Phases_and_Apsides.data');
   $wCount = count($wArray);

   $wTable  = '';

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);
   list
  (
   $JD_TT,
   $y,
   $mmm,
   $dn,
   $DoW,
   $hhmmss_TT,
   $marker
  ) = PReg_Split("[ ]", $Curr_Line);

   if ($y == $Y and $mmm == $Mmm)
      {
       $wTable .= "$Curr_Line\n";
      }
  }
   return $wTable;
}








   function MET_TT ($Year, $MonthNum)
{
   $Y = IntVal(trim($Year));
   $m = IntVal(trim($MonthNum));

   $Mmm = substr('JanFebMarAprMayJunJulAugSepOctNovDec', 3*($m-1), 3);

/*
2458527.43568522 2019 Feb 12 Tue 22:27:23  3.5149362 +14.238544 384994.793  90.00012 FQ
*/

   $wArray = file('data/All_Lunar_Events_and_Coords_Sans_Eclipses.data');
   $wCount = count($wArray);

   $wTable  = '';

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);
   $w = Strip_WSpace($Curr_Line); // exit("$w");
   list
  (
   $JD_TT,
   $y,
   $mmm,
   $dn,
   $DoW,
   $hhmmss_TT,
   $RA_Hrs,
   $Decl_Deg,
   $Dist_km,
   $Phase_Deg,
   $marker
  ) = PReg_Split("[ ]", $w);

// Do some formatting.
   $RA_Hrs = SPrintF("%010.7f", $RA_Hrs);
   $Decl_Deg = SPrintF("%+010.6f", $Decl_Deg);
   $Phase_Deg = SPrintF("% 7.3f", $Phase_Deg);

   if ($y == $Y and $mmm == $Mmm)
      {
       $wTable .= "$JD_TT $y $Mmm $dn $DoW $hhmmss_TT $RA_Hrs $Decl_Deg $Dist_km $Phase_Deg $marker\n";
      }
  }
   return $wTable;

}











   function G2_Moon_Ephem_Stats_TT ($JDTTArg)
{
// Read JDTT argument.
   $JDTT = FloatVal(trim($JDTTArg));

// Compute JD Number TT.
   $JDNumTT = floor($JDTT + 0.5);

// Get G2 Moon data file into work array and count file lines.
// File contains the Right Ascension, Declination, Distance
// and phase data for the moon for the years 1600 to 2200.
   $wArray = file('data/DE405_G2_Moon_Geo_Equ_Coords.data');
   $wCount = count($wArray);

// Construct main 7-point working tables centered
// on JD Number of date.
   $RAHrsTable = $DeclDegTable = $DistKmTable = $PhaseDegTable = '';

   $j = 2*($JDNumTT - 2305446.0);

   for ($i = $j-3;    $i < $j+4;    $i++)
  {
// Get current lunar ephemeris data table line.
   $CurrLine = trim($wArray[$i]);

// Parse the current ephemeris data line.
   $TJDTT    = SPrintF("%7.1f", 2305446.0 + $i/2);
   $RAHrs    = SPrintF("%010.7f",  substr($CurrLine, 0, 9) / 1E7);
   $DeclDeg  = SPrintF("%+010.6f", substr($CurrLine, 9, 9) / 1E6);
   $DistKm   = SPrintF("%10.3f",  substr($CurrLine, 18, 9) / 1E3);
   $PhaseDeg = SPrintF("%10.3f",  substr($CurrLine, 27, 8) / 1E5);

// Append current ephemeris data
// lines to interpolation tables.
   $RAHrsTable    .= "$TJDTT $RAHrs\n";
   $DeclDegTable  .= "$TJDTT $DeclDeg\n";
   $DistKmTable   .= "$TJDTT $DistKm\n";
   $PhaseDegTable .= "$TJDTT $PhaseDeg\n";
  }

// Perform high-precision interpolations for RAHrs, DeclDeg
// and DistAU for the moment indicated by the JDTT argument.
   $RAHrs    = Interp_RA(trim($RAHrsTable), $JDTT);
   $DeclDeg  = Lagrange_Interp(trim($DeclDegTable), $JDTT);
   $DistKm   = Lagrange_Interp(trim($DistKmTable), $JDTT);
   $PhaseDeg = Interp_Angle_360(trim($PhaseDegTable), $JDTT);

// Do some output formatting.
   $JDTT     = SPrintF("%7.8f",    $JDTT);
   $RAHrs    = SPrintF("% 10.7f",  $RAHrs);
   $DeclDeg  = SPrintF("%+ 10.6f", $DeclDeg);
   $DistKm   = SPrintF("%10.3f",   $DistKm);
   $PhaseDeg = SPrintF("% 9.5f",   $PhaseDeg);

//   $fDistKm  = number_format($DistKm,3);

// Done.
   return "$JDTT $RAHrs $DeclDeg $DistKm $PhaseDeg";

} // end of  G2_Moon_Ephem_Stats_TT (JDTTArg)




/*
   This functions returns a text description
   of the lunar phase corresponding to the
   given phase angle argument in degrees.
*/
   function Lunar_Phase_to_Text ($PhaseAngDeg)
{
// Round phase angle to nearest integer.
   $a = round(abs(trim($PhaseAngDeg)), 0);

   if ($a >= 0  and $a <=  10) {return 'New Moon.';}
   if ($a > 10  and $a <=  30) {return 'Waxing crescent, just past New Moon.';}
   if ($a > 30  and $a <=  79) {return 'Waxing crescent moon.';}
   if ($a > 79  and $a <=  88) {return 'Waxing crescent, approaching first quarter.';}
   if ($a > 88  and $a <= 101) {return 'First quarter moon.';}
   if ($a > 101 and $a <= 115) {return 'Waxing gibbous, just past first quarter.';}
   if ($a > 115 and $a <= 159) {return 'Waxing gibbous moon.';}
   if ($a > 159 and $a <= 170) {return 'Waxing gibbous, approaching Full Moon.';}
   if ($a > 170 and $a <= 192) {return 'Full Moon.';}
   if ($a > 192 and $a <= 198) {return 'Waning gibbous, just past Full Moon.';}
   if ($a > 198 and $a <= 250) {return 'Waning gibbous moon.';}
   if ($a > 250 and $a <= 260) {return 'Waning gibbous, approaching last quarter.';}
   if ($a > 260 and $a <= 271) {return 'Last quarter moon.';}
   if ($a > 271 and $a <= 274) {return 'Waning crescent, just past last quarter.';}
   if ($a > 274 and $a <= 319) {return 'Waning crescent moon.';}
   if ($a > 319 and $a <= 349) {return 'Waning crescent, approaching New Moon.';}
   if ($a > 349 and $a <= 360) {return 'New Moon.';}

   return FALSE; // = ERROR = Angle outside valid 0 to 360 degree range.

} // End of Lunar_ Phase_to_Text (PhaseAngDeg)




   function Lunar_Eclipses_Table_TT ($Year)
{
   $wArray = file('data/All_Lunar_Eclipses.data');
   $wCount = count($wArray);

   $wTable = '';

   for($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);
   list($q, $Y) = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

   if (trim($Y) == trim($Year))
  {
   list
   (
   $q, $Y,$Mmm,$dn, $hhmmss_TT, $q,$q,$q, $Ecl_Type,
   $q,$q,$q,$q, $PenDurMin, $UmbDurMin, $TotDurMin


   ) = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

// 08665  1600 Jan 30  06:25:29    118  -4947  127   P   a-  -0.9183  1.2067  0.1399  294.2   87.7    -     17N   93W

   $m = 1 + StrPos('JanFebMarAprMayJunJulAugSepOctNovDec', $Mmm)/3;

   $JD_Num_TT = GregorianToJD($m, IntVal(trim($dn)), IntVal(trim($Y)));

   $Frac_Day_TT = HMS_to_Hours($hhmmss_TT)/24;

   $JD_TT = SPrintF("%7.8f", $JD_Num_TT + $Frac_Day_TT - 0.5);

/*
   $Lunar_Ephem_Stats_TT = G2_Body_Ephem_Stats_TT('Moon', $JD_TT);
// 2305476.76769676 8.7929063 +17.047285 392860.917 179.91600
   list($JD_TT, $RA_Hrs, $Decl_Deg, $Dist_km) = PReg_Split("[ ]", Strip_WSpace($Lunar_Ephem_Stats_TT));
*/
   $RA_Hrs = '';//SPrintF("%010.7", $RA_Hrs);
   $Decl_Deg = '';//SPrintF("%+010.6", $Decl_Deg);

   $DoW = Substr('SunMonTueWedThuFriSat', 3*(($JD_Num_TT + 1) % 7) ,3);

   $hhmmss_DT = NASA_Delta_T_Poly($JD_TT, 'G', 3);

// Filter out special secondary symbols from eclipse type.
   $Ecl_Type = Str_Replace('b', '', $Ecl_Type);
   $Ecl_Type = Str_Replace('m', '', $Ecl_Type);
   $Ecl_Type = Str_Replace('e', '', $Ecl_Type);
   $Ecl_Type = Str_Replace('-', '', $Ecl_Type);
   $Ecl_Type = Str_Replace('+', '', $Ecl_Type);

   $Ecl_Type = Str_Replace('x', '', $Ecl_Type);

   $Ecl_Type = Str_Replace('P', 'ParLunEcl', $Ecl_Type);
   $Ecl_Type = Str_Replace('T', 'TotLunEcl', $Ecl_Type);
   $Ecl_Type = Str_Replace('N', 'PenLunEcl', $Ecl_Type);


   $PenDur_HM = $UmbDur_HM = $TotDur_HM = '-';
   $h = 'h';
   $m = 'm';


   if ($PenDurMin <> '-')
  {
   $w = $PenDurMin / 60;
   $hh = SPrintf("%02d", floor($w));
   $mm = SPrintF("%04.1f", 60*($w - $hh));
   $PenDur_HM = "$hh$h$mm$m";
  }

   if ($UmbDurMin <> '-')
  {
   $w = $UmbDurMin / 60;
   $hh = SPrintf("%02d", floor($w));
   $mm = SPrintF("%04.1f", 60*($w - $hh));
   $UmbDur_HM = "$hh$h$mm$m";
  }

   if ($TotDurMin <> '-')
  {
   $w = $TotDurMin / 60;
   $hh = SPrintf("%02d", floor($w));
   $mm = SPrintF("%04.1f", 60*($w - $hh));
   $TotDur_HM = "$hh$h$mm$m";
  }

   $wTable .= "$JD_TT $Y $Mmm $dn $DoW $hhmmss_TT $Ecl_Type $PenDur_HM $UmbDur_HM $TotDur_HM\n";
  }

  }
   return trim($wTable);
}









/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/

/*
   ===================
// PLANETARY FUNCTIONS

   THIS MODULE CONTAINS THE PHPSL G2 PLANET-RELATED FUNCTIONS, INCLUDING
   PLUTO, FOR USE WITH THE PHPSL G2 DE405 SOLAR SYSTEM MODULE AND RELATED
   DATA AND GRAPHICS FILES.  PLUTO IS ALSO INCLUDED IN THIS EPHEMERIS.

   AUTHOR: Jay Tanner
   2001-2019

   PROGRAMMING LANGUAGE: PHP v7.x

   LICENSE: Unless otherwise indicated, this source code and all related
   data and graphics files are released under the provisions of the
   Creative Commons Attribution-ShareAlike v4.0 International.

   https://creativecommons.org/licenses/by-sa/4.0/legalcode

*/





/*
   This function fetches the closest forward match to a given
   DE405 body name.

   FALSE is returned if no match is found.

*/
   function Check_Body_Name ($BodyNameString)
{
   $AllBodyNames = 'Sun        Moon       Mercury    Venus      Earth      Mars       Jupiter    Saturn     Uranus     Neptune    Pluto';

   $BodyName = StrToLower(trim($BodyNameString));

   if ($BodyName == ''){return FALSE;}

// ---------------------------------------------------------
// Seek closest forward case INsensitive match to body name.
// Return empty string '' if no match is found.

   $i = StrPos($AllBodyNames, UCFirst($BodyName));
        if ($i === FALSE) {return FALSE;}

   return trim(substr($AllBodyNames, $i, 7));

}







// ===================================================
// This function generates a planetary ephemeris table
// for the year and month arguments TT.
// Handles the eight major planets and Pluto.

   function G2_Planet_Ephem_Table_TT ($PlanetName, $Year, $Month, $WithHeader=FALSE)
{

// Read input argument values.
   $Y = IntVal(trim($Year));
   $m = IntVal(trim($Month));

   $Body_Name = UCFirst(StrToLower(trim($PlanetName)));
   $Body_Name = Check_Body_Name($Body_Name);
   $BODY_NAME = StrToUpper($Body_Name);
   $Body_Abbr = substr($Body_Name, 0,3);

// Compute JD Number TT for first day of the month.
   $JD_Num_1st_TT = GregorianToJD($m, 1, $Y);

// Get full month name.
   $Month_Name = JDMonthName($JD_Num_1st_TT, 1);

// Determine number of days in the month.
   $Days_in_Month = Cal_Days_In_Month(CAL_GREGORIAN, $m, $Y);

// Load solar data file into working array.
   $wArray = file("data/DE405_G2_"."$Body_Name"."_Geo_Equ_Coords.data");
   $wCount = count($wArray);

// Construct monthly solar ephemeris table with
// extra padding at each end of the month.
   $iStart = 2*($JD_Num_1st_TT - 2305446.0);
   $iEnd   = $iStart + 2*$Days_in_Month + 1;

   $Planet_Ephem_Table = '';

   for($i = $iStart - 4;    $i < $iEnd + 2;    $i++)
  {
   $Curr_Line = trim($wArray[$i]);
   $TJD_TT = SPrintF("%7.1f", 2305446 + $i/2);

// Parse the current data table line.
// 094115718+1925519800746621430-008

   $RA_Hrs   = substr($Curr_Line,  0, 2) . '.' . substr($Curr_Line,  2, 7);
   $Decl_Deg = substr($Curr_Line,  9, 3) . '.' . substr($Curr_Line, 12, 6);
   $Dist_AU  = substr($Curr_Line, 18, 2) . '.' . substr($Curr_Line, 20, 9);
   $Vis_Mag  = substr($Curr_Line, 29, 3) . '.' . substr($Curr_Line, 32, 1);

   $w = trim(Inv_JD($TJD_TT));
   $w = Str_Replace('12:00:00', '', $w);
   $w = Str_Replace('00:00:00', '', $w);
   $Date_Time_String_TT = Str_Replace('+', '', $w);

// DO SOME OUTPUT FORMATTING.
   $RA_Hrs   = SPrintF("% 10.7f",  $RA_Hrs);
   $Decl_Deg = SPrintF("% +10.6f", $Decl_Deg);
   $Dist_AU  = SPrintF("% 12.9f",  $Dist_AU);
   $Vis_Mag  = SPrintF("% +5.1f",  $Vis_Mag);

//   $mark = substr($Curr_Line, -2);
     $mark = trim(StrRChr($Curr_Line, ' '));


//   if ($mark <> 'Pg' and $mark <> 'Ag')  {$mark = '';}

// If event is outside target month, then remove event marker.
   if ($i < $iStart-0 or $i > $iEnd-2) {$mark = '';}

// Construct and append line to output table.
   $Planet_Ephem_Table .= "$TJD_TT  $Date_Time_String_TT $RA_Hrs  $Decl_Deg  $Dist_AU   $Vis_Mag $mark\n";
  }
   $Planet_Ephem_Table = trim($Planet_Ephem_Table);

   if ($WithHeader === FALSE) {return $Planet_Ephem_Table;}

// ----------------------------------------------
// Construct and return planetary ephemeris table
// for the given body name, year and month.

   return
"==============================================================================
$BODY_NAME - Geocentric Equatorial Ephemeris

For $Month_Name of $Y
True Equator and Equinox of Date


  JD_TT     Calendar_Date    RA_Hours   Declin_Deg   Distance_AU   Vis_Mag
=========  ===============  ==========  ==========  ============  ========
$Planet_Ephem_Table";

} // end of  G2_Planet_Ephem_Table_TT (PlanetName, Year, Month)





/*
   This function generates a DE405 ephemeris of any given
   solar system body for any given year and month.

   The recognized bodies are the sun, moon, planets and Pluto.
   Year = 1600 to 2200

*/

   function G2_Body_Ephem_Table_TT ($BodyName, $Year, $Month, $HeaderFlag=FALSE)
{
   $Body_Name = Check_Body_Name(UCFirst(StrToLower(trim($BodyName))));
   $Y = IntVal(trim($Year));
   $m = IntVal(trim($Month));

   if ($Body_Name == 'Sun')  {return G2_Sun_Ephem_Table_TT  ($Y, $m, $HeaderFlag);}
   if ($Body_Name == 'Moon') {return G2_Moon_Ephem_Table_TT ($Y, $m, $HeaderFlag);}

   return G2_Planet_Ephem_Table_TT ($Body_Name, $Y, $m, $HeaderFlag);
}








   function G2_Planet_Ephem_Stats_TT ($PlanetNameStr, $JDTTArg)
{
// Read input arguments,
   $PlanetName = UCFirst(StrToLower(trim($PlanetNameStr)));
   $JDTT = FloatVal(trim($JDTTArg));

// Compute JD Number TT.
   $JDNumTT = floor($JDTT + 0.5);


// Get G2 planetary data file into work array and count file lines.
// File contains the Right Ascension, Declination and distance
// data for the planet for the 600-year period from 1600 to 2200.
   $wArray = file("data/DE405_G2_$PlanetName"."_Geo_Equ_Coords.data");
   $wCount = count($wArray);

// Construct main 7-point working tables centered
// on JD Number of date.
   $RAHrsTable = $DeclDegTable = $DistAUTable = $VMagTable = '';

   $j = 2*($JDNumTT - 2305446.0);

   for ($i = $j-3;    $i < $j+4;    $i++)
  {
// Get current planetary ephemeris data table line.
   $CurrLine = trim($wArray[$i]);

// Parse the current planetary ephemeris data line.
   $TJDTT   = SPrintF("%7.1f", 2305446.0 + $i/2);
   $RAHrs   = SPrintF("%010.7f",  substr($CurrLine,  0, 9) / 1E7);
   $DeclDeg = SPrintF("%+010.6f", substr($CurrLine,  9, 9) / 1E6);
   $DistAU  = SPrintF("%011.9f",  FloatVal(substr($CurrLine, 18,11) / 1E9));
   $VMag    = SPrintF("%+05.1f",  substr($CurrLine, 29, 4) / 10);

// Append current ephemeris data
// lines to interpolation tables.
   $RAHrsTable   .= "$TJDTT $RAHrs\n";
   $DeclDegTable .= "$TJDTT $DeclDeg\n";
   $DistAUTable  .= "$TJDTT $DistAU\n";
   $VMagTable    .= "$TJDTT $VMag\n";
  }

// Delete work array after use.
   unset($wArray);

// Perform high-precision interpolations for RAHrs, DeclDeg
// and DistAU for the moment indicated by the JDTT argument.
   $RAHrs   = Interp_RA(trim($RAHrsTable),      $JDTT);
   $DeclDeg = Lagrange_Interp(trim($DeclDegTable), $JDTT);
   $DistAU  = Lagrange_Interp(trim($DistAUTable),  $JDTT);
   $VMag    = Lagrange_Interp(trim($VMagTable),    $JDTT);

// Do some output formatting.
   $JDTT    = SPrintF("%7.8f",    $JDTT);
   $RAHrs   = SPrintF("% 10.7f",  $RAHrs);
   $DeclDeg = SPrintF("%+ 10.6f", $DeclDeg);
   $DistAU  = SPrintF("% 12.9f",  $DistAU);
   $VMag    = Str_Replace(' -0.0', '  0.0', SPrintF("%+ 5.1f",  $VMag));
   $VMag    = Str_Replace(' +0.0', '  0.0', SPrintF("%+ 5.1f",  $VMag));

// Done.
   return "$JDTT $RAHrs $DeclDeg $DistAU  $VMag";

} // end of  G2_Planet_Ephem_Stats_TT (PlanetNameStr, ...)






/*
   This function returns the ephemeris statistics
   for any of the given recognized bodies and JDTT
   argument.
*/

   function G2_Body_Ephem_Stats_TT ($BodyName, $JDTTArg)
{
   $Body_Name = Check_Body_Name(trim($BodyName));

   $JDTT = FloatVal(trim($JDTTArg));

   if ($Body_Name == 'Sun')  {return G2_Sun_Ephem_Stats_TT($JDTT);}

   if ($Body_Name == 'Moon') {return G2_Moon_Ephem_Stats_TT($JDTT);}

   return G2_Planet_Ephem_Stats_TT ($Body_Name, $JDTT);

   }





/*
   This function returns the ephemeris statistics
   for any of the given recognized bodies and JDLT
   argument. JDLT is internally converted to JDTT.
*/

   function G2_Body_Ephem_Stats_LT ($BodyName, $JDLTArg, $TZhhmm='+00:00')
{
   $Body_Name = Check_Body_Name(trim($BodyName));

   $JD_LT = FloatVal(trim($JDLTArg));

   $hhmm_TZ = trim($TZhhmm);
   $Frac_Day_TZ = HMS_to_Hours($hhmm_TZ) / 24;

   $Frac_Day_DT = NASA_Delta_T_Poly($JD_LT) / 86400;

// TT = LT - (TZ) + (DT)
// Compute JD_TT corresponding to given LT_TT argument.
   $JD_TT = $JD_LT - $Frac_Day_TZ + $Frac_Day_DT;
   $JD_TT = SPrintF("%7.8f", $JD_TT);

   if ($Body_Name == 'Sun')  {return G2_Sun_Ephem_Stats_TT($JD_TT);}

   if ($Body_Name == 'Moon') {return G2_Moon_Ephem_Stats_TT($JD_TT);}

   return G2_Planet_Ephem_Stats_TT ($Body_Name, $JD_TT);

   } // end of  G2_Body_Ephem_Stats_LT (BodyName, JDLTArg, TZhhmm)














/*
   This function returns an ephemeris table for any recognized
   body for every month of the given year TT.
*/
   function G2_Body_Year_Ephem_Table_TT ($BodyName, $Year)
{
   $Body_Name = Check_Body_Name(trim($BodyName));

   $Y = floor(trim($Year));

   $wTable = '';

   for ($m=1;   $m <= 12;   $m++)
  {
   $Curr_Table = G2_Body_Ephem_Table_TT($Body_Name, $Y, $m, TRUE);

   $wTable .= "$Curr_Table\n";
  }

   return $wTable;
}










/*
   function zzz_Sun_Earth_Moon_Events_Table_UT ($Year)
{
   $Y = ' '.IntVal(trim($Year)).' ';

   $hhmm_TZ = '+00:00';
   $hours = HMS_to_Hours(trim($hhmm_TZ));

// Determine W/E symbol to use for Time Zone
// offset from UT (Neg|Pos = West|East)
   $WE_TZ_Str = ($hours < 0)? 'West':'East';
   if ($hours == 0){$WE_TZ_Str = '';}





// GET ALL EARTH APSIDES FOR YEAR.
   $wArray = file('data/All_Earth_Apsides_and_Intervals.data');
   $wCount = count($wArray);

   $wTable = '';

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line, 0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JD_Date_Time_TT = substr($Curr_Line, 0,41);
       $JD_TT = FloatVal($JD_Date_Time_TT);
       $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);
       $JD_Date_Time_LT = $JD_LT." ".trim(Inv_JD($JD_LT));
       $JD_LT = FloatVal($JD_Date_Time_LT);
       $hhmmss_DT = NASA_Delta_T_Poly(floor($JD_LT + 0.5), 'G', 3);
       $mark = trim(StrRChr($Curr_Line, ' '));
       $JD_TT = SPrintF("%7.8f", $JD_TT);
       $JD_LT = SPrintF("%7.8f", $JD_LT);
       $wTable .= "\n$JD_Date_Time_LT $hhmmss_DT $mark\n";
      } // end of  if (StrPos ...)
      } // end of  if (Is_Numeric ...)
      } // end of  for (i=0; ...)

   unset($wArray);





// GET ALL EARTH EQUINOXES AND SOLSTICES EVENTS FOR YEAR.
   $wArray = file('data/All_Equinoxes_and_Solstices.data');
   $wCount = count($wArray);

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JD_Date_Time_TT = substr($Curr_Line, 0,41);
       $JD_TT = FloatVal($JD_Date_Time_TT);
       $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);
       $JD_Date_Time_LT = $JD_LT." ".trim(Inv_JD($JD_LT));
       $JD_LT = FloatVal($JD_Date_Time_LT);
       $hhmmss_DT = NASA_Delta_T_Poly(floor($JD_LT + 0.5), 'G', 3);
       $mark = trim(StrRChr($Curr_Line, ' '));
       $JD_TT = SPrintF("%7.8f", $JD_TT);
       $JD_LT = SPrintF("%7.8f", $JD_LT);
       $wTable .= "\n$JD_Date_Time_LT $hhmmss_DT $mark\n";

      } // end of  if (StrPos ...)
      } // end of  if (Is_Numeric ...)
      } // end of  for (i=0; ...)

   unset($wArray);


























// =============================================================
// GET ALL LUNAR PHASES, PERIGEE AND APOGEE EVENTS FOR THE YEAR.
// $wArray = file('data/All_Lunar_Phases_and_Apsides.data');

   $wArray = file('data/All_Lunar_Events_and_Coords_Sans_Eclipses.data');

   $wCount = count($wArray);

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);
// 2305494.05063511 1600 Feb 16 Wed 13:12:55 23.4235759  +0.750621 361752.733  25.12661 Pg
   if (Is_Numeric(substr($Curr_Line,0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JD_Date_Time_TT = substr($Curr_Line, 0,41);

list($q,$q,$q,$q,$q,$q, $q,$q, $Dist_km) = PReg_Split("[ ]", Strip_WSpace($Curr_Line));

       $JD_TT = FloatVal($JD_Date_Time_TT);
       $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);
       $JD_Date_Time_LT = $JD_LT." ".trim(Inv_JD($JD_LT));
       $mark = trim(StrRChr($Curr_Line, ' '));

if ($mark == 'Pg' or $mark == 'Ag')
   {
    $Dist_km .= " km";
   }
else
   {
    $Dist_km = "";
   }

       $JD_TT = SPrintF("%7.8f", $JD_TT);
       $JD_LT = SPrintF("%7.8f", $JD_LT);
       $wTable .= "\n$JD_Date_Time_LT $hhmmss_DT $mark $Dist_km\n";

//   $wTable = Str_Replace(" Pg",     " Perigee $Dist_km km", $wTable);
//   $wTable = Str_Replace(" Ag",     " Apogee $Dist_km km",  $wTable);

      } // end of  if (StrPos ...)
      } // end of  if (Is_Numeric ...)
      } // end of  for (i=0; ...)

   unset($wArray);







// Sort all events forward by sequential JD order.
   $wTable = Sort_JD_Events_Table($wTable);

   $Y = trim($Y);

// Get proper time scale symbol (LT|UT).
   $LT_UT_Str = ($hours == 0)?'UT':'LT';

   $wTable =
"
==============================================================================
                 BASIC EARTH, SOLAR AND LUNAR EVENTS SYNOPSIS

                 For the Year $Y                   UT Scale

================ =============================================================
 Julian_Date_UT   Calendar_Date  Time_UT   Delta_T      Solar_System_Event
================ =============== ======== ========= ==========================
$wTable
";

// Special name replacement filters for event marker codes.
   $wTable = Str_Replace(" Ph",     " Earth at Perihelion", $wTable);
   $wTable = Str_Replace(" Ah",     " Earth at Aphelion",   $wTable);
   $wTable = Str_Replace(" Pg",     " Moon Perigee",        $wTable);
   $wTable = Str_Replace(" Ag",     " Moon Apogee",         $wTable);
   $wTable = Str_Replace(" NM",     " New Moon",            $wTable);
   $wTable = Str_Replace(" FQ",     " First Quarter Moon",  $wTable);
   $wTable = Str_Replace(" FM",     " Full Moon",           $wTable);
   $wTable = Str_Replace(" LQ",     " Last Quarter Moon",   $wTable);
   $wTable = Str_Replace(" MarEqu", " March Equinox",       $wTable);
   $wTable = Str_Replace(" JunSol", " June Solstice",       $wTable);
   $wTable = Str_Replace(" SepEqu", " September Equinox",   $wTable);
   $wTable = Str_Replace(" DecSol", " December Solstice",   $wTable);

   return trim($wTable);

} // end of  Basic_Sun_Earth_Moon_Events_Table_UT (Year)


*/







   function Sun_Earth_Moon_Events_Table_TZ ($Year, $hhmmTZ='+00:00')
{
   $Y = ' '.IntVal(trim($Year)).' ';
   $hours = HMS_to_Hours(trim($hhmmTZ));
   $w = Hours_to_HMS(HMS_to_Hours($hours), 0, '+', ':');
   $hhmm_TZ = substr($w, 0,6);

// Determine W/E symbol to use for Time Zone
// offset from UT (Neg|Pos = West|East)
   $WE_TZ_Str = ($hours < 0)? 'West':'East';
   if ($hours == 0){$WE_TZ_Str = '= UT';}

// Get all Earth apsides for year.
   $wArray = file('data/All_Earth_Apsides_and_Intervals.data');
   $wCount = count($wArray);

   $wTable = '';

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line, 0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JD_Date_Time_TT = substr($Curr_Line, 0,41);
       $JD_TT = FloatVal($JD_Date_Time_TT);
       $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);
       $JD_Date_Time_LT = $JD_LT." ".trim(Inv_JD($JD_LT));
       $JD_LT = FloatVal($JD_Date_Time_LT);
       $hhmmss_DT = NASA_Delta_T_Poly(floor($JD_LT + 0.5), 'G', 3);

       $hhmm_AMPM = HMS_24_to_12(substr($JD_Date_Time_LT, -8));

       if(substr($hhmm_AMPM ,0,1) == '0')
         {$hhmm_AMPM = ' '.LTrim($hhmm_AMPM,'0');}

       $JD_Date_Time_LT = substr($JD_Date_Time_LT,0,33)."$hhmm_AMPM";

       $mark = trim(StrRChr($Curr_Line, ' '));
       $JD_TT = SPrintF("%7.8f", $JD_TT);
       $JD_LT = SPrintF("%7.8f", $JD_LT);
       $wTable .= "\n$JD_Date_Time_LT $mark\n";

      } // end of  if (StrPos ...)

      } // end of  if (Is_Numeric ...)

  } // end of  for (i=0; ...)

   unset($wArray);




// Get all Earth equinoxes and solstices events for year.
   $wArray = file('data/All_Equinoxes_and_Solstices.data');
   $wCount = count($wArray);

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JD_Date_Time_TT = substr($Curr_Line, 0,41);
       $JD_TT = FloatVal($JD_Date_Time_TT);
       $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);
       $JD_Date_Time_LT = $JD_LT." ".trim(Inv_JD($JD_LT));
       $JD_LT = FloatVal($JD_Date_Time_LT);
       $hhmmss_DT = NASA_Delta_T_Poly(floor($JD_TT + 0.5), 'G', 3);

       $hhmm_AMPM = HMS_24_to_12(substr($JD_Date_Time_LT, -8));

       if(substr($hhmm_AMPM ,0,1) == '0')
         {$hhmm_AMPM = ' '.LTrim($hhmm_AMPM,'0');}

       $JD_Date_Time_LT = substr($JD_Date_Time_LT,0,33)."$hhmm_AMPM";

       $mark = trim(StrRChr($Curr_Line, ' '));
       $JD_TT = SPrintF("%7.8f", $JD_TT);
       $JD_LT = SPrintF("%7.8f", $JD_LT);
       $wTable .= "\n$JD_Date_Time_LT $mark\n";

      } // end of  if (StrPos ...)

      } // end of  if (Is_Numeric ...)

  } // end of  for (i=0; ...)

   unset($wArray);



// =============================================================
// Get all lunar phases, perigee and apogee events for the year.

   $wArray = file('data/All_Lunar_Phases_and_Apsides.data');
   $wCount = count($wArray);

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JD_Date_Time_TT = substr($Curr_Line, 0,41);
       $JD_TT = FloatVal($JD_Date_Time_TT);
       $JD_LT = JDTT_to_JDLT($JD_TT, $hhmm_TZ);
       $JD_Date_Time_LT = $JD_LT." ".trim(Inv_JD($JD_LT));

       $hhmm_AMPM = HMS_24_to_12(substr($JD_Date_Time_LT, -8));

       if (substr($hhmm_AMPM, 0,1) == '0')
          {$hhmm_AMPM = ' '.LTrim($hhmm_AMPM,'0');}

       $JD_Date_Time_LT = substr($JD_Date_Time_LT, 0,33)."$hhmm_AMPM";

       $mark = trim(StrRChr($Curr_Line, ' '));
       $JD_TT = SPrintF("%7.8f", $JD_TT);
       $JD_LT = SPrintF("%7.8f", $JD_LT);
       $wTable .= "\n$JD_Date_Time_LT $mark\n";

      } // end of  if (StrPos ...)

      } // end of  if (Is_Numeric ...)

  } // end of  for (i=0; ...)

   unset($wArray);

// Sort all events forward by sequential JD order.
   $wTable = Sort_JD_Events_Table($wTable);

   $Y = trim($Y);

// Get proper time scale symbol (LT|UT).
   $LT_UT_Str = ($hours == 0)?'UT':'LT';

   $wTable =
"
==============================================================================
                         BASIC SOLAR SYSTEM EVENTS SYNOPSIS

                 For the Year $Y    Time Zone UT$hhmm_TZ $WE_TZ_Str

================ =============================================================
 Julian_Date_$LT_UT_Str   Calendar_Date  Time_$LT_UT_Str                Solar_System_Event
================ =============== ======== ====================================
$wTable
";


// Special name replacement filters for event marker codes.

   $w = Str_Replace(' Ph', ' Earth Perihelion - Closest to Sun',  $wTable);
   $w = Str_Replace(' Ah', ' Earth Aphelion - Farthest From Sun', $w);
   $w = Str_Replace(' Pg', ' Moon at Perigee - Minimum Distance', $w);
   $w = Str_Replace(' Ag', ' Moon at Apogee - Maximum Distance',  $w);
   $w = Str_Replace(' NM', ' New Moon - Phase 0 deg - Invisible', $w);
   $w = Str_Replace(' FQ', ' First Quarter Moon - Phase 90 deg',  $w);
   $w = Str_Replace(' FM', ' Full Moon - Phase 180 deg',          $w);
   $w = Str_Replace(' LQ', ' Last Quarter Moon - Phase 270 deg',    $w);
   $w = Str_Replace(' MarEqu', ' March Equinox',                  $w);
   $w = Str_Replace(' JunSol', ' June Solstice',                  $w);
   $w = Str_Replace(' SepEqu', ' September Equinox',              $w);
   $wTable = Str_Replace(' DecSol', ' December Solstice',         $w);

   return trim($wTable);

} // end of  Basic_Sun_Earth_Moon_Events_Table_TZ (Year)

































/*

   function Sun_Earth_Moon_Events_Table_TT ($Year)
{
   $Y = ' '.IntVal(trim($Year)).' ';

// Get all Earth perihelion(s) and aphelion(s) for year.
   $wArray = file('data/All_Earth_Apsides_and_Intervals.data');
   $wCount = count($wArray);

   $wTable = '';

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JDTT_Date_Time_TT = substr($Curr_Line,0,41);
       $JD_TT = FloatVal($JDTT_Date_Time_TT);
       $hhmmss_DT = NASA_Delta_T_Poly(floor($JD_TT + 0.5), 'G', 3);
       $mark = trim(StrRChr($Curr_Line, ' '));
       $wTable .= "$JDTT_Date_Time_TT $mark\n";

      } // end of  if (StrPos ...)

      } // end of  if (Is_Numeric ...)

  } // end of  for (i=0; ...)

   unset($wArray);




// Get all Earth equinoxes and solstices events for year.
   $wArray = file('data/All_Equinoxes_and_Solstices.data');
   $wCount = count($wArray);

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JDTT_Date_Time_TT = substr($Curr_Line,0,41);
       $JD_TT = FloatVal($JDTT_Date_Time_TT);
       $hhmmss_DT = NASA_Delta_T_Poly(floor($JD_TT + 0.5), 'G', 3);
       $mark = trim(StrRChr($Curr_Line, ' '));
       $wTable .= "$JDTT_Date_Time_TT $mark\n";
      } // end of  if (StrPos ...)

      } // end of  if (Is_Numeric ...)

  } // end of  for (i=0; ...)

   unset($wArray);




// Get all lunar phases, perigee and apogee events for year.
   $wArray = file('data/All_Lunar_Phases_and_Apsides.data');
   $wCount = count($wArray);

   for ($i=0;   $i < $wCount;   $i++)
  {
   $Curr_Line = trim($wArray[$i]);

   if (Is_Numeric(substr($Curr_Line,0,1)))
      {
   if (StrPos($Curr_Line, $Y) !== FALSE)
      {
       $JDTT_Date_Time_TT = substr($Curr_Line,0,41);
       $JD_TT = FloatVal($JDTT_Date_Time_TT);
       $hhmmss_DT = NASA_Delta_T_Poly(floor($JD_TT + 0.5), 'G', 3);
       $mark = trim(StrRChr($Curr_Line, ' '));
       $wTable .= "$JDTT_Date_Time_TT $mark\n";

      } // end of  if (StrPos ...)
      } // end of  if (Is_Numeric ...)

  } // end of  for (i=0; ...)

   unset($wArray);

   $wTable = Sort_JD_Events_Table($wTable);

   $Y = trim($Y);

   $wTable =
"
==============================================================================
                 BASIC EARTH, SOLAR AND LUNAR EVENTS SYNOPSIS

                 For the Year $Y                   TT Scale

================ =============================================================
 Julian_Date_TT   Calendar_Date  Time_TT            Solar_System_Event
================ =============== ======== ====================================
$wTable
";

// Special name replacement filters for event marker codes.
   $w = Str_Replace(' Ph', ' Earth Perihelion - Closest to Sun',  $wTable);
   $w = Str_Replace(' Ah', ' Earth Aphelion - Farthest From Sun', $w);
   $w = Str_Replace(' Pg', ' Moon at Perigee - Minimum Distance', $w);
   $w = Str_Replace(' Ag', ' Moon at Apogee - Maximum Distance',  $w);
   $w = Str_Replace(' NM', ' New Moon - Phase 0 deg - Invisible', $w);
   $w = Str_Replace(' FQ', ' First Quarter Moon - Phase 90 deg',  $w);
   $w = Str_Replace(' FM', ' Full Moon - Phase 180 deg',          $w);
   $w = Str_Replace(' LQ', ' Last Quarter Moon - Phase 270 deg',  $w);
   $w = Str_Replace(' MarEqu', ' March Equinox',                  $w);
   $w = Str_Replace(' JunSol', ' June Solstice',                  $w);
   $w = Str_Replace(' SepEqu', ' September Equinox',              $w);
   $w = Str_Replace(' DecSol', ' December Solstice',              $w);


   return trim($w);


} // end of  Basic_Sun_Earth_Moon_Events_Table_TT (Year)

*/




/*
   This function takes an events JD Table and reformats it
   to put a blank line between the month tables.

   This function should be used only on the finished events
   JD table just prior to tabulating.

   LTUTTT = Time Scale label to attach ('LT', 'UT' or 'TT').


   JD table line looks something like:
   2458127.22584206 2018 Jan 08 Mon 17:25:13 Moon LQ

*/

   function U_Separate_Events_Table_by_Month ($JDTable)
{
   $JD_Table = trim($JDTable);

//   $Frac_Day_TZ = HMS_to_Hours($TZhhmm);

   $Time_Scale = 'TT';

   $wArray = PReg_Split("[\n]", $JD_Table);
   $wCount = count($wArray);

   $HL1 =
"\n\n Julian_Date_$Time_Scale   Calendar_Date  Time_$Time_Scale  Solar_System_Event";

   $HL2 =
"---------------- --------------- -------- -------------------------------------";

   $wTable = "";

//   $Prev_Line  = substr($wTable,22,3);

   for($i=1;   $i < $wCount;   $i++)
  {
   $Prev_Line = trim($wArray[$i-1]);
   $Prev_Mmm  = substr($Prev_Line,22,3);
   $Curr_Line = trim($wArray[$i-0]);
   $Curr_Mmm  = substr($Curr_Line,22,3);

   $flag = FALSE;

   $Curr_Body_and_Mark = substr($Curr_Line, 41, StrLen($Curr_Line));
   $LPart = substr($Curr_Line, 0, 41);
   $RPart = $Curr_Body_and_Mark;

   if ($Prev_Mmm <> $Curr_Mmm)
      {$wTable .= " $HL1\n$HL2\n";}

       $wTable .= "$LPart$RPart\n";

  }
   $wTable = $wTable;

   $Curr_Line = trim($wArray[0]);
   $LPart = substr($Curr_Line, 0, 41);
   $RPart = trim(substr($Curr_Line, 41, StrLen($Curr_Line)));
   $w = "$LPart $RPart\n$wTable\n";

   return $w;

} // end of  Separate_Events_Table_by_Month (JDTable)












   function Sort_JD_Events_Table ($EventsTableText)
{
   $w = trim($EventsTableText);
   $wArray = explode("\n", $w);
   sort($wArray);
   $w = implode ("\n", $wArray);

   return trim($w);
}





/*
   This function returns a table of basic solar
   system events for any given year TT.
*/

   function Solar_System_Events_Table_TT ($Year)
{
   $_Y_ = ' '.IntVal(trim($Year)).' ';

//   $wArray = file('data/All_Lunar_Solar_Events_With_Lunar_Eclipses.data');
//   $wArray = file('data/All_Solar_System_Events_With_Lunar_and_Solar_Eclipses.data');
   $wArray = file('data/All_Basic_Solar_System_Events_Synopsis_1600_to_2200.data');


   $wCount = count($wArray);

   $wTable = '';

   for ($i=0;   $i < $wCount;   $i++)
       {
        $Curr_Line = trim($wArray[$i]);

    if (StrPos($Curr_Line, $_Y_) !== FALSE)
       {$wTable .= "$Curr_Line\n";}

   $w = $wTable;

// Implement marker symbol translation filters.

   $w = Str_Replace('AnnSolEcl',  ' Annular SOLAR ECLIPSE',           $w);
   $w = Str_Replace('HibSolEcl',  ' Hibrid SOLAR ECLIPSE',            $w);
   $w = Str_Replace('ParSolEcl',  ' Partial SOLAR ECLIPSE',           $w);
   $w = Str_Replace('TotSolEcl',  ' - TOTAL SOLAR ECLIPSE',           $w);

   $w = Str_Replace(' PenLunEcl', 'Penumbral LUNAR ECLIPSE Midpoint', $w);
   $w = Str_Replace(' ParLunEcl', 'Partial LUNAR ECLIPSE Midpoint',   $w);
   $w = Str_Replace(' TotLunEcl', 'TOTAL LUNAR ECLIPSE Midpoint',     $w);

   $w = Str_Replace('MerPg',      ' AU - Mercury at perigee',         $w);
   $w = Str_Replace('VenPg',      ' AU - Venus at perigee',           $w);
   $w = Str_Replace('MarPg',      ' AU - Mars at perigee',            $w);
   $w = Str_Replace('JupPg',      ' AU - Jupiter at perigee',         $w);
   $w = Str_Replace('SatPg',      ' AU - Saturn at perigee',          $w);
   $w = Str_Replace('UraPg',      ' AU - Uranus at perigee',          $w);
   $w = Str_Replace('NepPg',      ' AU - Neptune at perigee',         $w);
   $w = Str_Replace('PluPg',      ' AU - Pluto at perigee',           $w);

   $w = Str_Replace('MerAg',      ' AU - Mercury at apogee',          $w);
   $w = Str_Replace('VenAg',      ' AU - Venus at apogee',            $w);
   $w = Str_Replace('MarAg',      ' AU - Mars at apogee',             $w);
   $w = Str_Replace('JupAg',      ' AU - Jupiter at apogee',          $w);
   $w = Str_Replace('SatAg',      ' AU - Saturn at apogee',           $w);
   $w = Str_Replace('UraAg',      ' AU - Uranus at apogee',           $w);
   $w = Str_Replace('NepAg',      ' AU - Neptune at apogee',          $w);
   $w = Str_Replace('PluAg',      ' AU - Pluto at apogee',            $w);

   $w = Str_Replace(' NM',        'New Moon',                         $w);
   $w = Str_Replace(' FQ',        'First Quarter Moon',               $w);
   $w = Str_Replace(' FM',        'Full Moon',                        $w);
   $w = Str_Replace(' LQ',        'Last Quarter Moon',                $w);
   $w = Str_Replace(' MarEqu',    ' March Equinox',                    $w);
   $w = Str_Replace(' JunSol',    ' June Solstice',                    $w);
   $w = Str_Replace(' SepEqu',    ' September Equinox',                $w);
   $w = Str_Replace(' DecSol',    ' December Solstice',                $w);
   $w = Str_Replace(' Ph',        ' AU - Earth at perihelion',         $w);
   $w = Str_Replace(' Ah',        ' AU - Earth at aphelion',           $w);
   $w = Str_Replace(' Pg',        ' km - Moon at perigee',             $w);
   $w = Str_Replace(' Ag',        ' km - Moon at apogee',              $w);
   $w = Str_Replace('  ',         ' ',                                 $w);
        }

   return $w;

} // end of  Solar_System_Events_Table_TT (Year)













/*
   This is a special function to swap the perigee, apogee elements order.

The table format will change from this:
2495373.50501809 2120 Jan 01 Mon 12:07 AM 356756.737 km - Moon at Perigee
2495379.65077606 2120 Jan 07 Sun  3:37 AM 0.983349465 AU - Earth at Perihelion
2495380.04604838 2120 Jan 07 Sun  1:06 PM First Quarter Moon
2495387.32959632 2120 Jan 14 Sun  7:55 PM 406406.754 km - Moon at Apogee
2495388.12112783 2120 Jan 15 Mon  2:54 PM Full Moon
2495396.09285557 2120 Jan 23 Tue  2:14 PM Last Quarter Moon
2495402.04649026 2120 Jan 29 Mon  1:07 PM 357715.773 km - Moon at Perigee
2495402.74650366 2120 Jan 30 Tue  5:55 AM New Moon

To this format:
2495373.50501809 2120 Jan 01 Mon 12:07 AM Moon at Perigee  356756.737 km
2495379.65077606 2120 Jan 07 Sun  3:37 AM Earth at Perihelion at 0.983349465 AU
2495380.04604838 2120 Jan 07 Sun  1:06 PM First Quarter Moon
2495387.32959632 2120 Jan 14 Sun  7:55 PM Moon at Apogee at 406406.754 km
2495388.12112783 2120 Jan 15 Mon  2:54 PM Full Moon
2495396.09285557 2120 Jan 23 Tue  2:14 PM Last Quarter Moon
2495402.04649026 2120 Jan 29 Mon  1:07 PM Moon at Perigee at 357715.773 km
2495402.74650366 2120 Jan 30 Tue  5:55 AM New Moon


for sun
2457101.69835649 2015 Mar 20 Fri  4:46 AM 02m47s - Total Solar ECLIPSE Max Point
2457101.69835649 2015 Mar 20 Fri  4:46 AM Total Solar ECLIPSE Dur 02m47s

*/

   function Swap_Peri_Apo_Elements ($WorkTable)
{
   $w = trim($WorkTable);

   $wArray = PReg_Split("[\n]", $w);
   $wCount = count($wArray);

   $wTable = '';

   for($i=0;   $i < $wCount;   $i++)
   {
   $Curr_Line = trim($wArray[$i]);

// Process only lines with ' - ' substrings.
// Swap values on each side of the ' - ' substrings.
// 2495379.65077606 2120 Jan 07 Sun  3:37 AM 0.983349465 AU - Earth at Perihelion
   if (StrPos($Curr_Line, ' - ') !== FALSE)
      {
/*
   A = 2495379.65077606 2120 Jan 07 Sun  3:37 AM
   B = 0.983349465 AU - Earth at Perihelion
   C = 0.983349465 AU
   D = Earth at Perihelion
   E = Earth at Perihelion  0.983349465 AU
   F = 2495379.65077606 2120 Jan 07 Sun  3:37 AM
*/
       $A_Str = trim(substr($Curr_Line, 0,41));
       $B_Str = trim(substr($Curr_Line, 41, StrLen($Curr_Line)));
       list($C_Str, $D_Str) = PReg_Split("[ - ]", $B_Str);
       $C_Str = trim($C_Str);
       $D_Str = trim($D_Str);
       $E_Str = "$D_Str = $C_Str";

       $w = "$A_Str $E_Str";
       $w = Str_Replace('Perigee =', 'Perigee at', $w);
       $w = Str_Replace('Apogee =', 'Apogee at', $w);
       $w = Str_Replace('Perihelion =', 'Perihelion at', $w);

       $Curr_Line = Str_Replace('Aphelion =', 'Aphelion at', $w);
      }
   $wTable .= "$Curr_Line\n";
   }

   $wTable = Str_Replace('=', 'at', $wTable);

return "$wTable";

} // end of  Swap_Peri_Apo_Elements (WorkTable)


/*
   ###########################################################################
   ###########################################################################
   ###########################################################################
*/





?>


